;
;Super Nintendo Entertainment System(tm) Audio Processing Unit Emulator
;                                                          Copyright (C)2003-04 Alpha-II Productions
;

CPU	386
BITS	32


;
;Header files

%include	"Macro.Inc"
%include	"SPC700.Inc"
%include	"DSP.Inc"
%define	INTERNAL
%include	"APU.Inc"



;
; Variables

SECTION .data ALIGN=4

	pAPURAM	DD	0

	cycLeft	DD	0										;Clock cycles left to emulate in EmuAPU loop
	smpDec	DD	0										;Unused clocks from cycle to sample conversion

	smpRate	DD	0										;Sample rate
	smpRAdj	DD	0										;Sample rate adjustment (16.16)
	smpREmu	DD	0										;Number of emulated samples per second

	songLen	DD	0										;Song and fade length
	fadeLen	DD	0

	apuOpt	DD	(CPU_CYC << 24)  | (DEBUG << 16) | (DSPINTEG << 17) | (VMETERM << 8) | (VMETERV << 9) | (SNESINT << 10) | (STEREO << 11) | (HALFC << 1) | (CNTBK << 2) | (SPEED << 3) | (IPLW << 4) | (DSPBK << 5)



;
; Code

SECTION .text ALIGN=16

;
;Initialize Audio Processing Unit

PROC InitAPU, pInitRAM

	Mov	EAX,[pInitRAM]
	Mov	[pAPURAM],EAX

	Call	InitSPC,[pAPURAM]
	Call	InitDSP,[pAPURAM]

	Mov	dword [smpRate],32000
	Mov	dword [smpRAdj],10000h

	Call	SetAPUSmpClk,[smpRAdj]

ENDP


;
;Reset Audio Processor

PROC ResetAPU, amp

	Call	ResetSPC
	Call	ResetDSP

	Cmp	dword [amp],-1
	JE		short .NoAmp
		Call	SetDSPAmp,[amp]
	.NoAmp:

	Mov	dword [cycLeft],0
	Mov	dword [smpDec],0
	Mov	dword [songLen],7FFFFFFFh				;Set song and fade lengths to max and min values
	Mov	dword [fadeLen],1

ENDP


;
;Fix Audio Processor After Load

PROC FixAPU, pc, a, y, x, psw, s

	Call	FixSPC,[pc],[a],[y],[x],[psw],[s]
	Call	FixDSP

ENDP


;
;Load SPC File

PROC LoadSPCFile, pFile
USES ECX,ESI,EDI

	Call	ResetAPU, -1

	Mov	ESI,[pFile]

	Add	ESI,100h										;memcpy(pAPURAM, &spc[0x100], 0x10000)
	Mov	EDI,[pAPURAM]
	Mov	ECX,10000h / 4
	Rep	MovSD

	Mov	EDI,dsp										;memcpy(&dsp, &spc[0x10100], 128)
	Mov	ECX,128 / 4
	Rep	MovSD

	Add	ESI,40h										;memcpy(xram, &spc[0x101C0], 64)
	Mov	EDI,extraRAM
	Mov	ECX,40h / 4
	Rep	MovSD

	Mov	ESI,[pFile]
	XOr	EAX,EAX
	Mov	AL,[2Bh+ESI]								;SP
	Push	EAX
	Mov	AL,[2Ah+ESI]								;PSW
	Push	EAX
	Mov	AL,[28h+ESI]								;X
	Push	EAX
	Mov	AL,[29h+ESI]								;Y
	Push	EAX
	Mov	AL,[27h+ESI]								;A
	Push	EAX
	Mov	AX,[25h+ESI]								;PC
	Push	EAX
	Call	FixAPU

ENDP

;
;Save Audio Processor State

PROC SaveAPU, pSPC, pDSP

	Call	SaveSPC,[pSPC]
	Call	SaveDSP,[pDSP]

ENDP


;
;Restore Audio Processor State

PROC RestoreAPU, pSPCL, pDSPL

	Call	RestoreSPC,[pSPCL]
	Call	RestoreDSP,[pDSPL]

ENDP


;
;Set Audio Processor Options

PROC SetAPUOpt, mix, chn, bits, rate, inter, opts

	Mov	EAX,[rate]									;Is a rate specified?
	Cmp	EAX,-1										;if (rate!=-1)
	JE		short .KeepRate

		Mov	EAX,[rate]

		Cmp	EAX,8000									;if (rate < 8000) rate = 8000;
		JAE	short .OKL
			Mov	EAX,8000
		.OKL:

		Cmp	EAX,192000								;if (rate > 192000) rate = 192000;
		JBE	short .OKH
			Mov	EAX,192000
		.OKH:

		Mov	[smpRate],EAX
		Call	SetAPUSmpClk,[smpRAdj]				;Calculate the number of clock cycles per sample

	.KeepRate:

	Call	SetDSPOpt,[mix],[chn],[bits],[rate],[inter],[opts]	;Set options in DSP emulator

ENDP


;
;Set Audio Processor Sample Clock

PROC SetAPUSmpClk, speed
USES EDX

	Mov	EAX,[speed]
	Cmp	EAX,32768									;if (speed < 32768) speed = 32768;
	JAE	short .OKL
		Mov	EAX,32768
	.OKL:

	Cmp	EAX,1048576									;if (speed > 1048576) speed = 1048576;
	JBE	short .OKH
		Mov	EAX,1048576
	.OKH:
	Mov	[smpRAdj],EAX

	Mov	EAX,[smpRate]								;smpREmu = (smpRate << 16) / smpRAdj;
	MovZX	EDX,word [2+smpRate]
	ShL	EAX,16
	Div	dword [smpRAdj]
	Mov	[smpREmu],EAX

ENDP


;
;Set Audio Processor Song Length

PROC SetAPULength

	Jmp	SetDSPLength

ENDP


;
;Emulate Audio Processing Unit

PROC EmuAPU, pBuf, len, type
USES ECX,EDX,EBX,EDI

	Mov	ECX,APU_CLK

%if DSPINTEG
	;Fixup samples/cycles --------------------
	Test	byte [type],-1
	JZ		short .Cycles

		Call	SetEmuDSP,[pBuf],[len],[smpREmu]

		Mov	EAX,[len]
		Mul	ECX										;cycles = (APU_CLK * len) / smpREmu
		Div	dword [smpREmu]
		Add	EAX,[cycLeft]
		JLE	short .NoCycles
		Jmp	short .Samples

	.Cycles:
		Mov	EAX,[len]
		Add	EAX,[cycLeft]
		JLE	short .NoCycles

		Push	EAX
		Mul	dword [smpREmu]						;samples = (smpREmu * len) / APU_CLK
		Div	ECX
		Call	SetEmuDSP,[pBuf],EAX,[smpREmu]
		Pop	EAX

	.Samples:

	;Emulate APU -----------------------------
	Call	EmuSPC,EAX

	.NoCycles:
	Mov	[cycLeft],EAX
	Call	SetEmuDSP,0,0,0							;Create any remaining samples

%else

	;If samples were passed, convert to clock cycles
	Mov	EAX,[len]
	Cmp	byte [type],0
	JZ		short .Cycles

		Mul	ECX										;cycles = (APU_CLK * len) / smpREmu
		Div	dword [smpREmu]

	.Cycles:
	Add	[cycLeft],EAX

	;Emulate APU -----------------------------
	XOr	EBX,EBX										;Number of samples generated
	Mov	EDI,[pBuf]
	.Next:
		Mov	EAX,[cycLeft]
		Test	EAX,EAX
		JLE	short .Done

		;SPC700 -------------------------------
		Mov	EDX,EAX
 		Call	EmuSPC,EAX
 		Mov	[cycLeft],EAX

		;DSP ----------------------------------
		Sub	EDX,EAX									;Calculate number of samples to create
		Mov	EAX,EDX									;EAX = number of cycles emulated
		Mul	dword [smpREmu]
		Add	EAX,[smpDec]
		AdC	EDX,0
		Div	ECX										;samples = (((cycles - cycLeft) * smpREmu) + smpDec) / APU_CLK
		Mov	[smpDec],EDX

		Add	EBX,EAX									;size += samples
		Cmp	EBX,[len]								;Sometimes sample count will go over by one
		JBE	short .LenOK
			Add	EAX,[len]
			Sub	EAX,EBX
			Mov	EBX,[len]
		.LenOK:

		Call	EmuDSP,EDI,EAX							;pBuf = EmuDSP(pBuf,samples)
		Mov	EDI,EAX

		;Fade out music -----------------------
;		Mov	EAX,[t64Cnt]							;if (t64Cnt > songLen) SetFade()
;		Cmp	EAX,[songLen]
;		JBE	short .Next
;
;		Call	SetFade

	Jmp	short .Next

	.Done:

	;Make sure enough samples were created to fill buffer
	Cmp	byte [type],0								;if (type && size<len) EmuDSP(pBuf, len-size)
	Mov	EDX,[len]
	SetNZ	AL
	Sub	EDX,EBX
	SetG	AH
	Test	AL,AH
	RetZ	EDI

	Call	EmuDSP,EAX,EDX
%endif

ENDP


;
;Seek to Position

PROC SeekAPU, time, fast
USES ECX,EDX

	XOr	EDX,EDX
	Mov	EAX,[time]									;numSeconds = time / 64000
	Test	EAX,EAX
	RetZF
	Mov	ECX,64000
	Div	ECX
	Mov	ECX,EAX
	IMul	EDX,APU_CLK/64000							;EDX = (time % 64000) * (APU_CLK / 64000)

	Test	byte [fast],-1								;Fast mode completely bypasses the DSP emulation
	JZ		short .Slow

		Call	SetSPCDbg,-1,SPC_NODSP				;Disable writes to the DSP registers

		Add	EDX,[cycLeft]
		Mov	EAX,EDX
		.EmuSPC:

			Add	EAX,APU_CLK							;Emulate SPC for 1 second
			.Next:
				Call	EmuSPC,EAX
			Test	EAX,EAX
			JG		short .Next

		Dec	ECX
		JG		short .EmuSPC
		Mov	[cycLeft],EAX

		Call	SetSPCDbg,-1,0							;Re-enable writes to the DSP registers
		Jmp	short .Done

	.Slow:
		Call EmuAPU,0,APU_CLK,0
		Test	EDX,EDX
		JZ		short .EmuAPU

		Call	EmuAPU,0,EDX,0
		Test	ECX,ECX
		JZ		short .Done

		.EmuAPU:
			Call	EmuAPU,0,APU_CLK,0
		Dec	ECX
		JNZ	short .EmuAPU

	.Done:

	Call	FixSeek,[fast]								;Fixup DSP after seeking

;	Mov	EAX,[t64Cnt]								;if (t64Cnt > songLen) SetFade()
;	Cmp	EAX,[songLen]
;	JBE	short .NoFade
;		Call	SetFade
;	.NoFade:

ENDP


;
;Shutdown Audio Processing Unit

PROC ShutAPU

ENDP
