;
;Sony SPC700 Emulator
;                                                        Copyright (C)1999-2004 Alpha-II Productions
;

CPU	386
BITS	32


;
;Header files

%include	"Macro.Inc"
%include	"APU.Inc"
%include	"DSP.Inc"
%define	INTERNAL
%include	"SPC700.Inc"


;
;Equates

;Clock cycles per operation -----------------
T64_CYC	EQU	384									;64kHz timer clock divisor
T8_CYC	EQU	(8*T64_CYC)							;8kHz timer clock divisor

;Cleanup options ----------------------------
na			EQU	 0										;Option is not applicable to this instruction
WD			EQU	 001b									;Instruction writes to direct page
RD			EQU	 010b									;Instruction reads from direct page
WD16		EQU	 101b									;Instruction writes 16 bits to direct page
RD16		EQU	 110b									;Instruction reads 16 bits from direct page
WA			EQU	1001b									;Instruction writes to absolute address
RA			EQU	1010b									;Instruction reads from absolute address

;Registers ----------------------------------
%define	PC		SI										;Intel equivilants to SPC registers
%define	A		AL
%define	Y		AH
%define	YA		AX
%define	X		CH
%define	PS		DL
%define	S		EDI

;Pointers -----------------------------------
%define	OP1	ESI									;First instruction operand
%define	OP2	ESI+1									;Second instruction operand
%define	DPI 	EBX									;Direct Page Index
%define	ABSL	EBX									;Absolute Location
%define	RAM	EDI									;64K RAM


;
;Structures

;The flags are split up into eight dwords, with the flag being stored in the second byte.  The high
;word of P is loaded with the pointer to RAM, so the 32-bit value of [P-1] is a literal pointer to
;the current Direct Page.

STRUC SPCFlags
		resb	1
	CF	resb	4											;Carry (Called CF because MASM and TASM32 reserved C)
	Z	resb	4											;Zero
	I	resb	4											;Interrupts Enabled (unused in the SNES)
	H	resb	4											;Half-Carry
	B	resb	4											;Software Break
	P	resb	4											;Direct Page selector
	V	resb	4											;Integer Overflow
	N	resb	3											;Negative
ENDSTRUC

STRUC MemMap
	dp0			resb	0F0h							;Direct Page 0
		testReg	resb	1								;	Test register (no use)
		control	resb	1								;	Control register
		dspAddr	resb	1								;	DSP Address
		dspData	resb	1								;	DSP Data
		port0		resb	1								;	Port 0
		port1		resb	1								;	Port 1
		port2		resb	1								;	Port 2
		port3		resb	1								;	Port 3
					resb	1								;	unused
					resb	1								;	unused
		t0			resb	1								;	Timer 0
		t1			resb	1								;	Timer 1
		t2			resb	1								;	Timer 2
		c0			resb	1								;	Counter 0
		c1			resb	1								;	Counter 1
		c2			resb	1								;	Counter 2

	dp1			resb	100h							;Direct Page 1
	gp				resb	0FD00h						;General Pages 2-254
	up				resb	0C0h							;Uppermost Page 255
		ipl		resb	040h							;	Program used to transfer memory
ENDSTRUC



;
; Data and Variables

%ifndef	WIN32
SECTION .data ALIGN=256
%else
SECTION .data ALIGN=32
%endif

															;The 64 bytes of ROM in the IPL region
	iplROM	DB 0CDh,0EFh,0BDh,0E8h,000h,0C6h,01Dh,0D0h,0FCh,08Fh,0AAh,0F4h,08Fh,0BBh,0F5h,078h
				DB 0CCh,0F4h,0D0h,0FBh,02Fh,019h,0EBh,0F4h,0D0h,0FCh,07Eh,0F4h,0D0h,00Bh,0E4h,0F5h
				DB 0CBh,0F4h,0D7h,000h,0FCh,0D0h,0F3h,0ABh,001h,010h,0EFh,07Eh,0F4h,010h,0EBh,0BAh
				DB 0F6h,0DAh,000h,0BAh,0F4h,0C4h,0F4h,0DDh,05Dh,0D0h,0DBh,01Fh,000h,000h,0C0h,0FFh

	extraRAM	TIMES 64 DB 0							;Memory used for writes when ROM reading is enabled

	outPort	DB	0,0,0,0								;Out port values
	inPortCp	DB	0,0,0,0								;In port copies
	portMod	DB	0										;In port was modified (bits 3-0) SPC700 is sleeping (7)
	tControl	DB	0										;Copy of [F1h]
				DB	0,0

	t0Step	DB	0										;Timer pulses left until next counter increase,
	t1Step	DB	0										; aka up counters (even though I use them to count down)
	t2Step	DB	0,0

	clkTotal	DD	0										;Total number of clock cycles left to execute
	clkExec  DD	0										;Number of cycles needed for 64kHz timer to increase
	clkLeft	DD	0										;Number of cycles left to execute until timer increase
	t8kHz		DD	0										;Clock cycles left until 8kHz pulse
	t64kHz	DD	0										;Clock cycles left until 64kHz pulse
	t64Cnt	DD	0										;64kHz counter (increased every 64kHz pulse)
				DD	0,0

	pAPURAM	DD	0										;Pointer to 64K SPC RAM
	pOpFetch	DD	0										;Pointer to opcode fetcher
	pDebug	DD	0										;Pointer to tracing routine
	dbgOpt	DD	0										;Debugging options

	PSW		DD	0,0,0,0,0,0,0,0					;Flags in dword form

	regPC		DD	0										;Storage for registers between calls
	regYA		DD	0
	regSP		DD	0
	regX		DD	0

%if SPEED
	oldT64	DD	0										;Value of 't64Cnt' at the last counter read
%endif



;
; Code

%ifndef	WIN32
SECTION .text ALIGN=256
%else
SECTION .text ALIGN=16
%endif

;
;Convert flags
;   Convert flags between internal 32-byte and external 8-bit format

;Expand PSW into 8 dwords (destroys bit flags)
%macro ExpPSW 0

	Mov	EBX,PSW+1
	Mov	DH,8
	%%Next:
		ShR	PS,1
		SetC	[EBX]
		LEA	EBX,[4+EBX]
	Dec	DH
	JNZ	%%Next

%endmacro

;Compress 8 dwords into PSW (byte flags are unaffected)
%macro CmpPSW 0

	Mov	EBX,PSW+1
	Mov	EDX,80h
	%%Next:
		Mov	DH,[EBX]
		LEA	EBX,[4+EBX]
	ShR	EDX,1
	JNC	%%Next

%endmacro


;
;Initialize SPC700

PROC InitSPC, pAPUBase

	Mov	EAX,[pAPUBase]
	Mov	[pAPURAM],EAX
	Mov	[regPC],EAX
	Mov	AX,1FFh
	Mov	[regSP],EAX

	Call	SetSPCDbg,0,0								;Set fetch pointer to default if debugging is enabled

ENDP


;
;Reset SPC700

PROC ResetSPC
USES ECX,EDX,ESI,EDI

	;Erase 64K SPC RAM -----------------------
	Mov	EDI,[pAPURAM]
	Mov	EAX,-1										;Fill RAM with STOPs
	Mov	ECX,4000h
	Rep	StoSD

	;Reset Function Registers ----------------
	Mov	EDI,[pAPURAM]
	Add	EDI,0F0h
	Inc	EAX											;Set function regs to 0
	Mov	CL,4
	Rep	StoSD
	Mov	byte [EDI-0Fh],80h						;Enable ROM reading

	;Copy IPL ROM into extra RAM -------------
%if IPLW==0
	Mov	ESI,iplROM									;If IPL region writing is disabled, fill extra RAM
	Mov	EDI,extraRAM								; with IPL ROM
	Mov	CL,10h
	Rep	MovSD
%endif

	;Reset internal variables ----------------
	Mov	dword [t8kHz],T8_CYC-1					;Reset timer clocks
	Mov	dword [t64kHz],T64_CYC-1
	Mov	dword [t64Cnt],0
	Mov	byte [portMod],0

	Call	FixSPC,0FFC0h,0,0,0,0,0

ENDP


;
;Debug SPC700

PROC SetSPCDbg, pTrace, opts
USES	EDX

	Mov	EDX,[pDebug]

	Mov	EAX,[pTrace]
	Cmp	EAX,-1
	JE		short .NoFunc
		Mov	[pDebug],EAX
	.NoFunc:

	Mov	dword [pOpFetch],SPCFetch				;Disable instruction tracing

	Cmp	byte [opts],-1
	JE		short .NoOpts								;Leave options as they are

	Test	byte [opts],SPC_TRACE					;If trace is enabled, set fetch pointer to watch
	JZ		short .TraceOff
	Test	EAX,EAX										;Make sure function pointer isn't null
	JZ		short .TraceOff
		Mov	dword [pOpFetch],SPCTrace
	.TraceOff:

	Mov	EAX,[opts]									;Save options
	Mov	[dbgOpt],EAX

	.NoOpts:
	Mov	EAX,EDX

ENDP


;
;Fix SPC700 After Load

PROC FixSPC, inPC, inA, inY, inX, inPSW, inSP
USES ECX,EDX,EBX,ESI,EDI

	;Load registers --------------------------
	XOr	EAX,EAX
	Mov	AX,[inPC]
	Mov	[regPC],AX
	Mov	AL,[inA]
	Mov	AH,[inY]
	Mov	[regYA],EAX
	Mov	AL,[inX]
	Mov	[regX],AL
	Mov	AL,[inSP]
	Mov	[regSP],AL

	Mov	DL,[inPSW]
	Mov	EBX,PSW
	Mov	AH,8
	.Flag:
		Mov	dword [EBX],0
		ShR	DL,1
		SetC	[1+EBX]
		Add	EBX,4
	Dec	AH
	JNZ	short .Flag

	Mov	RAM,[pAPURAM]
	Or		[PSW+P-1],RAM								;Load location of SPC RAM into PSW.P

	Mov	AL,[RAM+t0]									;Initialize timer counters
	Dec	AL
	Mov	[t0Step],AL
	Mov	AL,[RAM+t1]
	Dec	AL
	Mov	[t1Step],AL
	Mov	AL,[RAM+t2]
	Dec	AL
	Mov	[t2Step],AL
	Mov	byte [3+t0Step],0

	XOr	EAX,EAX										;Erase out port values
	Mov	[outPort],EAX

	Mov	EAX,[RAM+port0]							;Copy port values to InP
	Mov	[inPortCp],EAX

	Mov	AL,[RAM+control]							;Copy control register for comparisons
	And	AL,87h
	Mov	[tControl],AL
	Mov	[RAM+control],AL

	;Copy the correct extra RAM --------------
	Test	byte [RAM+control],80h
	RetNZ

	LEA	ESI,[RAM+ipl]
	Mov	EDI,extraRAM
	Mov	ECX,10h
	Rep	MovSD

ENDP


;
;Get CPU Registers

PROC GetSPCRegs, pPC, pA, pY, pX, pPSW, pSP
USES EBX

	Mov	EBX,PSW+1
	Mov	AL,80h
	.Flag2:
		Mov	AH,[EBX]
		Add	EBX,4
	ShR	AX,1
	JNC	short .Flag2
	Mov	EBX,[pPSW]
	Mov	[EBX],AL

	Mov	EBX,[pPC]
	Mov	AX,[regPC]
	Mov	[EBX],AX

	Mov	EBX,[pA]
	Mov	AX,[regYA]
	Mov	[EBX],AL

	Mov	EBX,[pY]
	Mov	[EBX],AH

	Mov	EBX,[pX]
	Mov	AL,[regX]
	Mov	[EBX],AL

	Mov	EBX,[pSP]
	Mov	AL,[regSP]
	Mov	[EBX],AL

ENDP


;
;Save SPC700's Current State of Operation

PROC SaveSPC, pStateS
USES ALL

	Mov	EBX,[pStateS]

	;RAM -------------------------------------
	Mov	EDI,[EBX+pRAM]
	Test	EDI,EDI
	JZ		.NoRAM
		Mov	ESI,[pAPURAM]

		Test	byte [ESI+control],80h
		JNZ	short .ROM
			Mov	ECX,10000h/4
			Rep	MovSD
			Jmp	short .NoRAM
		.ROM:

		Mov	ECX,0FFC0h/4
		Rep	MovSD

		Mov	ESI,extraRAM
		Mov	ECX,40h/4
		Rep	MovSD
	.NoRAM:

	;Registers -------------------------------
	Mov	AX,[regPC]
	Mov	[EBX+rPC],AX

	Mov	AX,[regYA]
	Mov	[EBX+rA],AX

	Mov	AL,[regX]
	Mov	[EBX+rX],AL

	Mov	AL,[regSP]
	Mov	[EBX+rSP],AL

	Mov	ECX,PSW+1
	Mov	AL,80h
	.Flag:
		Mov	AH,[ECX]
		Add	ECX,4
	ShR	EAX,1
	JNC	short .Flag
	Mov	[EBX+rPSW],AL

	;Timers ----------------------------------
	Mov	EAX,[t8kHz]
	Mov	[EBX+tmr8kHz],EAX

	Mov	EAX,[t64kHz]
	Mov	[EBX+tmr64kHz],EAX

	Mov	EAX,[t64Cnt]
	Mov	[EBX+tmr64Cnt],EAX

	Mov	EAX,[t0Step]
	Mov	[EBX+upCnt],EAX

	;Ports -----------------------------------
	Mov	AL,[portMod]
	Mov	[EBX+inPMod],AL

	Mov	EAX,[outPort]
	Mov	[EBX+outP],EAX

	;Clear reserved bytes --------------------
	XOr	EAX,EAX
	Mov	[EBX+_rs1],EAX
	Mov	[EBX+_rs2],EAX
	Mov	[EBX+_rs3],EAX
	Mov	[EBX+_rs3+4],EAX

ENDP


;
;Restore SPC700's State of Operation

PROC RestoreSPC, pStateL
USES ALL

	Mov	EBX,[pStateL]
	XOr	EAX,EAX

	;RAM -------------------------------------
	Mov	ESI,[EBX+pRAM]
	Test	ESI,ESI
	JZ		.NoRAM
		Mov	EDI,[pAPURAM]
		Mov	ECX,0FFC0h/4
		Rep	MovSD

		Mov	EDI,extraRAM
		Mov	ECX,40h/4
		Rep	MovSD

		Mov	EDI,[pAPURAM]

		Mov	EAX,[EDI+port0]
		Mov	[inPortCp],EAX

		Mov	AL,[EDI+control]
		And	AL,87h
		Mov	[tControl],AL
		Mov	[EDI+control],AL

		Mov	ESI,iplROM
		Test	AL,80h
		JNZ	short .ROM
			Mov	ESI,extraRAM
		.ROM:

		Mov	EDI,[pAPURAM]
		Mov	DI,ipl
		Mov	ECX,40h/4
		Rep	MovSD
	.NoRAM:

	;Registers -------------------------------
	Mov	AX,[EBX+rPC]
	Mov	[regPC],AX

	Mov	AX,[EBX+rA]
	Mov	[regYA],AX

	Mov	AL,[EBX+rX]
	Mov	[regX],AL

	Mov	AL,[EBX+rSP]
	Mov	[regSP],AL

	Mov	AL,[EBX+rPSW]
	Mov	ECX,PSW
	Mov	EDX,7
	XOr	EDI,EDI
	.Flag:
		Mov	dword [EDX*4+ECX],EDI
		ShL	AL,1
		SetC	[1+EDX*4+ECX]
	Dec	EDX
	JNS	short .Flag

	Mov	EDI,[pAPURAM]
	Or		[PSW+P-1],EDI

	;Timers ----------------------------------
	Mov	EAX,[EBX+tmr8kHz]
	Mov	[t8kHz],EAX

	Mov	EAX,[EBX+tmr64kHz]
	Mov	[t64kHz],EAX

	Mov	EAX,[EBX+tmr64Cnt]
	Mov	[t64Cnt],EAX

	Mov	EAX,[EBX+upCnt]
	Mov	[t0Step],EAX
	Mov	byte [3+t0Step],0

	;Ports -----------------------------------
	Mov	AL,[EBX+inPMod]
	Mov	[portMod],AL

	Mov	EAX,[EBX+outP]
	Mov	[outPort],EAX
ENDP


;
;Write to APU RAM

PROC SetAPURAM, addr, val
USES EBX

	;Write value to memory -------------------
	Mov	EBX,[pAPURAM]
	Mov	BX,[addr]
	Mov	AL,[val]
	Mov	[EBX],AL

	;Check write -----------------------------
	Cmp	BX,ipl
	JAE	short .WROM

	Dec	BH
	Cmp	BX,0FFF0h
	JB		short .Done

	;Handle function register ----------------
	Push	ECX,EDI,EBP
	Mov	EDI,[pAPURAM]
	Mov	EBP,.Return

	Mov	AL,BL
	XOr	EBX,EBX
	Mov	BH,AL
	Add	EBX,FuncBase-0F000h
	Jmp	EBX
	.Return:
	Pop	EBP,EDI,ECX
	RetN

	;Handle IPL ROM --------------------------
	.WROM:
	XOr	BX,BX
%if IPLW
	Test	byte [tControl],80h
	RetZ
%endif

	Push	ECX
	MovZX	ECX,word [addr]
	Sub	ECX,ipl
%if IPLW
	Mov	AL,[EBX+ECX+ipl]
	Mov	[ECX+extraRAM],AL
	Mov	AL,[ECX+iplROM]
%else
	Mov	AL,[ECX+extraRAM]
%endif
	Mov	[EBX+ECX+ipl],AL
	Pop	ECX

	.Done:

ENDP


;
;Write to SPC700 Port(s)

PROC InPort, portb, valb
USES ECX

	;Set flag in portMod ---------------------
	Mov	ECX,[portb]
	And	ECX,3
	Mov	AL,1
	ShL	AL,CL
	Or 	[portMod],AL

	;Write value to port ---------------------
	Mov	AL,[valb]
	Mov	[ECX+inPortCp],AL
	Add	ECX,[pAPURAM]
	Mov	[ECX+0F4h],AL

ENDP

PROC InPortW, portw, valw
USES ECX

	;Set flag in portMod ---------------------
	Mov	ECX,[portw]
	And	ECX,3
	Mov	AL,3
	ShL	AL,CL
	Or 	[portMod],AL

	;Write value to port ---------------------
	Mov	AX,[valw]
	Mov	[ECX+inPortCp],AX
	Add	ECX,[pAPURAM]
	Mov	[ECX+0F4h],AX

ENDP


;
;Emulate SPC700

PROC EmuSPC, cyc

	Test	byte [dbgOpt],SPC_HALT
	JZ		short .NoHalt
		Or		byte [dbgOpt],DSP_HALT				;Prevent the DSP from any further processing
		Push	EDX
		XOr	EAX,EAX									;if (EDX>T8_CYC) EAX=EDX-T8_CYC, else EAX=0
		Mov	EDX,[cyc]
		Sub	EDX,T8_CYC
		SetB	AL
		Dec	EAX
		And	EAX,EDX
		Pop	EDX
		Retn
	.NoHalt:

	PushAD												;Save all registers and load EAX with the number of

	Mov	EAX,[cyc]									; clock cycles to execute

%if DEBUG												;EBP = Location to jump to after handling an opcode
	Mov	EBP,[pOpFetch]
%else
	Mov	EBP,SPCFetch
%endif

	;Setup clock cycle execution -------------
	;clkLeft contains the number of clock cycles to emulate until timer 2 increases or it's time to quit

	XOr	EDX,EDX
	Mov	[clkTotal],EAX

	Sub	EAX,[t64kHz]								;if (cyc > t64kHz) clkExec = t64kHz
	SetA	DL												;else clkExec = cyc
	Dec	EDX
	And	EAX,EDX
	Add	EAX,[t64kHz]

	Mov	[clkExec],EAX								;Save the number we want to execute this lap
	Mov	[clkLeft],EAX

	;Load x86 registers ----------------------
	Mov	EDI,[pAPURAM]
	Mov	ESI,[regPC]
	Mov	EAX,[regYA]
	Mov	CH,[regX]
	XOr	EDX,EDX										;EDX = Number of emulated clock cycles, initialize to 0

	;Update DSP data register ----------------
	;(Incase it was modifed during DSP emulation)
	MovZX	EBX,byte [RAM+dspAddr]
	Mov	CL,[EBX+dsp]
	Mov	[RAM+dspData],CL

%if DEBUG
	Jmp	EBP											;Jump into emulation routine
%else
	Jmp	SPCFetch
%endif

ALIGN 16
SPCExit:
	Mov	[regPC],ESI									;Save emulated registers
	Mov	[regYA],EAX
	Mov	[regX],CH
	PopAD													;Restore x86 registers before call
	Mov	EAX,[clkTotal]								;Return clock cycles left to emulate

ENDP


;
;Opcode Fetcher
;
;Fetches the next opcode and jumps to the appropriate handler.  Keeps track of clock cycles emulated
;and timers.

Fetch:

	;In the debug build, before the fetcher handles the next opcode a user defined function is
	;called.

ALIGN	16
SPCTrace:
	Sub	[clkLeft],EDX								;Have we executed all clock cycles?
	JLE	SPCTimers									;  Yes, Update timers

SPCBreak:
	;Call SPCTrace ---------------------------
	Push	dword [t0Step]								;Pass down counters
	CmpPSW												;DL = PSW
	ShR	ECX,8											;CL = X
	Push	dword [regSP]								;Pass SP
	Push	EDX											; ""  PSW
	Push	ECX											; ""  X
	Push	EAX											; ""  YA
	Push	ESI											; ""  PC

	Mov	EAX,[t64kHz]								;Pass number of cycles left until 64kHz increase
	Mov	CL,CPU_CYC
	Div	CL
	Mov	[23+ESP],AL

	Call	[pDebug]										;Call tracing routine

	;Update registers ------------------------
	Pop	EAX
	Mov	SI,AX
	Pop	EAX
	And	EAX,0FFFFh
	Pop	ECX
	MovZX	ECX,CL
	ShL	ECX,8
	Pop	EDX
	ExpPSW
	Pop	EDX
	Mov	[regSP],DX
	Pop	EDX

	Mov	EBP,SPCExit
	Mov	EDI,[pAPURAM]								;Restore EDI
	XOr	EDX,EDX										;Reset DH since timers have already been handled

	Test	byte [dbgOpt],SPC_HALT | SPC_RETURN
	JNZ	SPCTimers

	Test	byte [dbgOpt],SPC_TRACE
	JNZ	short .Trace
		Mov	dword [pOpFetch],SPCFetch
	.Trace:
	Mov	EBP,[pOpFetch]								;Restore EBP

	;Fetching begins by first subtracting the number of clock cycles emulated by the last
	;instruction (passed in EDX).  If more cycles need to be emulated before updating timer 2, then
	;get the next opcode and emulate it.  To emulate, each opcode handler is aligned on a page
	;boundary.  An instruction opcode acts as an index to this 64K table of functions, and is used
	;to jump directly to the handler without needing to use a jump table.

ALIGN	16
SPCFetch:												;(All opcode handlers return to this point)
	Sub	[clkLeft],EDX								;Have we executed all clock cycles?
	JLE	short SPCTimers							;  Yes, Update timers

	XOr	EDX,EDX
	Mov	DH,[ESI]										;Get next opcode (EDX = Opcode << 8)
	Add	EDX,OpcBase									;Add the base of the emulation table to the opcode
	Inc	PC												;Move PC to first operand or next instruction
	Jmp	EDX											;Jump to handler

	;clkExec contains the number of clock cycles we wanted to emulate this round.  By subtracting
	;clkLeft (which will be a number <= 0) we get the actual number of cycles emulated, which will
	;be subtracted from the total number to be emulated (clkTotal) and used to update the timers.

ALIGN	16
	SPCTimers:
	Mov	EBX,[clkExec]								;EBX = Actual number of clock cycles emulated
	Sub	EBX,[clkLeft]

	Mov	DL,[tControl]								;DL = Control register

	;Update timers ---------------------------
	Sub	[t64kHz],EBX								;Subtract cycles from timer clock / Has it rolled over?
	JNS	.NoTInc2										;  No, Don't update timer 2
		Test	DL,4										;Is timer 2 enabled?
		SetNZ	CL											;  Yes, CL = 1

		.Tt:												;*** It's possible for t64kHz to be < -T64_CYC ***
			Inc	dword [t64Cnt]						;Increase 64kHz counter (external use)
			Add	DH,CL
		Add	dword [t64kHz],T64_CYC				;Restore timer 2 clock
		JL		short .Tt								;Repeat until t64kHz is positive

		Sub	[t2Step],DH								;Is counter 2 ready to be increased?
		JNC	short .NoC2Inc							;  No
			Mov	DH,[RAM+t2]							;Restore the number of steps needed to inc c2
			Dec	DH
			Mov	[t2Step],DH
			Inc	byte [RAM+c2]						;Increase counter 2
			And	byte [RAM+c2],0Fh					;Only lower 4-bits are operable
%if CNTBK
%if DSPINTEG
			Call	CatchUp
%else
			Mov	EBP,SPCExit							;Signal the emu to exit so the DSP can catch up
%endif
%endif
		.NoC2Inc:
	.NoTInc2:

	Sub	[t8kHz],EBX
	JNS	short .NoTInc
		XOr	CL,CL
		Add	dword [t8kHz],T8_CYC					;Reset clock pulse counter

		ShR	DL,1										;Decrease t0Step if counter is enabled
		SbB	[t0Step],CL								;Have all timer pulses occurred?
		JNC	short .NoC0Inc							;  Nope, Skip counter inc
			Mov	DH,[RAM+t0]
			Dec	DH
			Inc	byte [RAM+c0]						;Increase counter
			Mov	[t0Step],DH							;Reset timer
			And	byte [RAM+c0],0Fh					;Counters are only 4-bit
%if CNTBK
%if DSPINTEG
			Call	CatchUp
%else
			Mov	EBP,SPCExit							;Exit CPU so DSP can create noise
%endif
%endif
		.NoC0Inc:

		ShR	DL,1
		SbB	[t1Step],CL
		JNC	short .NoC1Inc
			Mov	DH,[RAM+t1]
			Dec	DH
			Inc	byte [RAM+c1]
			Mov	[t1Step],DH
			And	byte [RAM+c1],0Fh
%if CNTBK
%if DSPINTEG
			Call	CatchUp
%else
			Mov	EBP,SPCExit
%endif
%endif
		.NoC1Inc:
	.NoTInc:

	;After the timers are updated, subtract the number of cycles emulated from the total number
	;passed in.  If we've emulated all clock cycles requested by the caller, then quit.  Otherwise
	;add another 32 and return to the fetcher.

	Sub	[clkTotal],EBX								;Have we executed all clock cycles?
	JLE	SPCExit										;  Yes, Quit

	Add	dword [clkLeft],T64_CYC					;Execute another T64_CYC cycles / single 64kHz pulse
	Mov	EBX,[clkLeft]
	Mov	[clkExec],EBX								;Store the number we want to execute

	XOr	EDX,EDX										;Reset EDX for return to fetcher

	Jmp	EBP											;Return to fetcher


;
;Counter Speed Hack
;
;Modifies the up counters so the next counter to increase will increase on the next 64kHz pulse.
;This guarantees the program will never have to poll the counters more than four consecutive times,
;which greatly increases emulation speed.
;
;In:
;   EBX -> Counter just read (FD-FFh)
;
;Out:
;   Counter = 0
;
;Destroys:
;   nothing

%if SPEED
ALIGN	16
CntHack:

	Test	byte [EBX],-1								;Is counter >0?
	JNZ	.Reset										;  Yes, No need to speed up then

	Test	byte [tControl],7							;Are any timers enabled?
	JNZ	.Reset

		Push	EAX,ECX,EDX

		;This code was added because some games repeatedly check or keep a running total of the number
		;of counter increases.  If more than 64 machine cycles go by between counter reads, we'll
		;assume that the program is doing other things besides polling the counter, and essentially
		;disable the speed hack.

		Mov	EAX,[t64Cnt]
		Sub	EAX,[oldT64]
		Cmp	EAX,(64 * CPU_CYC) / T64_CYC		;Wait for no more than 64 machine cycles
		Mov	EAX,[t64Cnt]
		Mov	[oldT64],EAX
		JA		.NoHack

		;Since we don't know which timers the program is using, the emulator breaks on all counter
		;increases.  Because of this, the hack needs to advance the next closest counter, regardless
		;of which counter is actually being read.  So the first step is to figure out how many clock
		;cycles are needed for each counter to increase, then choose the smallest value.
		;
		;Basically the following statement is used to convert each counter to cycles.  If the timer
		;is turned off, 2^32-1 is returned to guarentee that it's not chosen as a valid value.
		;
		; if (APURAM.control & 1)
		;    EDX = t0Step * T8_CYC + t8kHz; //T64_CYC for timer 2
		; else
		;    EDX = 0xffffffff;

		MovZX	ECX,byte [t2Step]						;Get up counter 2
		XOr	EDX,EDX
		ShL	ECX,7										;ECX *= 128
		Test	byte [tControl],4						;Is the timer enabled?
		SetNZ	DL											;  Yes, EDX will = 0, -1 otherwise
		LEA	ECX,[ECX*2+ECX]						;ECX *= 3 (384 = T64_CYC)
		Dec	EDX
		Add	ECX,[t64kHz]							;ECX += t64kHz
		Or 	EDX,ECX									;EDX = -1 or cycles needed to increase counter 2

		MovZX	ECX,byte [t1Step]						;Do the same for counter 1
		XOr	EAX,EAX
		ShL	ECX,10									;ECX *= 1024 (128*8)
		Test	byte [tControl],2
		SetNZ	AL
		LEA	ECX,[ECX*2+ECX]						;ECX *= 3 (3072 = T8_CYC)
		Dec	EAX
		Add 	ECX,[t8kHz]
		Or 	EAX,ECX

		Cmp	EAX,EDX									;Does counter 1 have less time than counter 2?
		JB 	short .Cnt1								;  Yes, Leave EAX with counter 1
			Mov	EAX,EDX								;EAX = Clocks left until counter 2 increases
		.Cnt1:

		MovZX	ECX,byte [t0Step]						;And for counter 0
		XOr	EDX,EDX
		ShL	ECX,10
		Test	byte [tControl],1
		SetNZ	DL
		LEA	ECX,[ECX*2+ECX]
		Dec	EDX
		Add	ECX,[t8kHz]
		Or 	EDX,ECX

		Cmp	EAX,EDX									;Does counter 1 or 2 have less time than counter 0?
		JB 	short .Cnt2
			Mov	EAX,EDX
		.Cnt2:

		;Once we've figured out the least number of cycles needed for a counter increase, we need to
		;make sure we have enough total cycles left to emulate.  Then make sure we have enough cycles
		;to increase the 64kHz timer.

		Cmp	EAX,[clkTotal]							;Are there enough clocks left for shortest cnt to inc?
		JB 	short .Clk
			Mov	EAX,[clkTotal]						;  No, Change EAX to the # of cycles left to emulate
		.Clk:

		Cmp	EAX,T64_CYC								;Are there enough cycles for a 64kHz pulse?
		JB		short .NoHack							;  No, Quit since nothing will be modified

		;Now's the time to actually skip emulating clock cycles.  All we care about are 64kHz timer
		;increases, so first we divide the number of cycles to skip by T64_CYC and throw away the
		;remainder.  The number of pulses are added to T64Cnt, as if they really were emulated.
		;The pulses are then subtracted from t2Step.  Next the number of pulses is divided by 8 to
		;give us the number of 8kHz pulses to skip.  Since up to 7 64kHz pulses could occur before
		;one 8kHz pulse, the remainder is converted back into clock cycles and subtracted from t8kHz.
		;Finally, the total number of cycles skipped is subtracted from clkTotal.
		;
		; T64Cnt += EAX / T64_CYC;
		; t2Step -= EAX / T64_CYC;                       //# of 64kHz pulses
		; t8kHz -= ((EAX % T8_CYC) / T64_CYC) * T64_CYC; //Fraction of 8kHz pulses
		; t0Step -= EAX / T8_CYC;                        //# of 8kHz pulses
		; clkTotal -= EAX - (EAX % T64_CYC);             //Actual cycles skipped

		XOr	EDX,EDX									;EAX = # of 64kHz pulses
		Mov	ECX,T64_CYC
		Div	ECX

		Add	[t64Cnt],EAX							;Add pulses to 64kHz counter
		Sub	[t2Step],AL								;Subtract pulses from up counter 2

		Mov	EDX,EAX
		Mov	ECX,EAX
		And	EAX,7										;EAX = # of cycles to be subtracted from 8kHz timer
		ShR	EDX,3										;EDX = # of 8kHz pulses to be subtracted from up 0 & 1
		ShL	EAX,7
		ShL	ECX,7
		LEA	EAX,[EAX*2+EAX]
		LEA	ECX,[ECX*2+ECX]						;ECX = Total number of emulated clock cycles skipped

		Sub	[t8kHz],EAX
		Sub	[t0Step],DL
		Sub	[t1Step],DL
		Sub	[clkTotal],ECX

		Pop	EDX,ECX,EAX
		Ret

ALIGN 16
		.NoHack:
		Pop	EDX,ECX,EAX
		Ret

ALIGN 16
	.Reset:
	Mov	byte [EBX],0

Ret
%endif


;
;Sleep Speed Hack
;
;Simulates executing the SLEEP instruction x times by reducing clkTotal to a value less than 3, and
;updating all timers and counters accordingly.
;
;In:
;   nothing
;
;Destroys:
;   EDX

%if SPEED
SleepHack:
	Push	EAX,ECX

	Mov	EAX,[clkExec]								;EAX = Cycles that have been emulated
	Sub	EAX,[clkLeft]
	Sub	[clkTotal],EAX

	Mov	EAX,[clkTotal]								;EAX = Total cycles left to emulate
	XOr	EDX,EDX
	Mov	ECX,3
	Div	ECX											;EAX = # of SLEEP instructions that can be emulated
	Sub	[clkTotal],EDX								;Make clkTotal divisible by 3
	Mov	[clkExec],EDX
	Mov	[clkLeft],EDX

	;Timers 0 and 1 --------------------------
	Mov	EAX,[clkTotal]
	XOr	EDX,EDX
	Mov	ECX,T8_CYC
	Div	ECX											;EAX = # of 8kHz cycles

	Sub	[t8kHz],EDX
	JNS	short .No8Inc
		Add	dword [t8kHz],T8_CYC
		Inc	EAX
	.No8Inc:

	Push	EAX
	MovZX	EDX,byte [t0Step]
	Inc	EDX
	Sub	EDX,EAX
	JG		short .NoC0Inc
		MovZX	ECX,byte [RAM+t0]
		Dec	CL

		Neg	EAX
		XOr	EDX,EDX
		Div	ECX

		Add	[RAM+c0],AL
		And	byte [RAM+c0],0Fh

		Mov	CL,DL
		Mov	DL,[RAM+t0]
		Sub	DL,CL
	.NoC0Inc:
	Mov	[t0Step],DL

	Pop	EAX
	MovZX	EDX,byte [t1Step]
	Inc	EDX
	Sub	EDX,EAX
	JG		short .NoC1Inc
		MovZX	ECX,byte [RAM+t1]
		Dec	CL

		Neg	EAX
		XOr	EDX,EDX
		Div	ECX

		Add	[RAM+c1],AL
		And	byte [RAM+c1],0Fh

		Mov	CL,DL
		Mov	DL,[RAM+t1]
		Sub	DL,CL
	.NoC1Inc:
	Mov	[t1Step],DL

	;Timer 2 ---------------------------------
	Mov	EAX,[clkTotal]
	XOr	EDX,EDX
	Mov	ECX,T64_CYC
	Div	ECX

	Sub	[t64kHz],EDX
	JNS	short .No64Inc
		Add	dword [t64kHz],T64_CYC
		Inc	EAX
	.No64Inc:

	MovZX	EDX,byte [t2Step]
	Inc	EDX
	Sub	EDX,EAX
	JG		short .NoC2Inc
		MovZX	ECX,byte [RAM+t2]
		Dec	CL

		Neg	EAX
		XOr	EDX,EDX
		Div	ECX

		Add	[RAM+c2],AL
		And	byte [RAM+c2],0Fh

		Mov	CL,DL
		Mov	DL,[RAM+t2]
		Sub	DL,CL
	.NoC2Inc:
	Mov	[t2Step],DL

	Mov	EAX,[clkLeft]
	Mov	[clkTotal],EAX

	Pop	ECX,EAX

Ret
%endif

;
;Macros

;
;Load pointers
;   Load DPI or ABSL (aliases for EBX) with the value needed by the instruction.

;dp - Load DPI with the 8-bit immediate value
%macro Ldp 0
	Mov	EBX,dword [PSW+P-1]						;;EBX-> Direct Page 0 or 1
	Mov	BL,[OP1]										;;BL-> Location in DP
%endmacro

;dp - Load DPI with the 2nd 8-bit immediate value
%macro Ldp2 0
	Mov	EBX,dword [PSW+P-1]
	Mov	BL,[OP2]
%endmacro

;(X) - Load DPI with the value in X
%macro LX 0
	Mov	EBX,dword [PSW+P-1]
	Mov	BL,X
%endmacro

;(Y) - Load DPI with the value in Y
%macro LY 0
	Mov	EBX,dword [PSW+P-1]
	Mov	BL,Y
%endmacro

;dp+X - Load DPI with the 8-bit immediate value + X
%macro LdpX 0
	Ldp
	Add	BL,X
%endmacro

;dp+Y - Load DPI with the 8-bit immediate value + Y
%macro LdpY 0
	Ldp
	Add	BL,Y
%endmacro

;abs - Load ABSL with the 16-bit immediate value
%macro Labs 0
	Mov	EBX,[pAPURAM]
	Mov	BX,[OP1]
%endmacro

;abs+X - Load ABSL with the 16-bit immediate value + X
%macro LabsX 0
	Mov	EBX,[pAPURAM]
	Mov	BL,X
	Add	BX,[OP1]
%endmacro

;abs+Y - Load ABSL with the 16-bit immediate value + Y
%macro LabsY 0
	Mov	EBX,[pAPURAM]
	Mov	BL,Y
	Add	BX,[OP1]
%endmacro

;[dp+X] - Load ABSL with the 16-bit value at [dp+X]
%macro LadpX 0
	LdpX
	Mov	BX,[DPI]
%endmacro

;[dp]+Y - Load ABSL with the 16-bit value at [dp] + Y
%macro LadpY 0
	Ldp
	Mov	BX,[DPI]
	Add	BL,Y
	AdC	BH,0
%endmacro

;[abs+X] - Load ABSL with the 16-bit value at [abs+X]
%macro LaabsX 0
	Mov	EBX,[pAPURAM]
	Mov	BL,X
	Add	BX,[OP1]
%endmacro

;mem.bit - Load ABSL with the first 13-bits of the immediate value, and DL with the last 3
;   ABSL-> byte in mem 0-1FFFh
;   DL  -> bit in byte
%macro Lmbit 0
	Labs
	Mov	DL,BH
	ShR	DL,5
	And	BH,1Fh
%endmacro


;
;Get flags from CPU
;   The x86 CPU updates its flags with the execution of each instruction.  So these macros update
;   the PSW with the x86 equivilants.

;Get Carry flag
%macro GetC 0
	SetC	[PSW+CF]
%endmacro

;Get Not Carry flag
%macro GetCs 0
	SetNC	[PSW+CF]
%endmacro

;Get Negative and Zero flags
%macro GetNZ 0
	SetS	[PSW+N]
	SetZ	[PSW+Z]
%endmacro

;Get Negative, Zero, and Carry flags
%macro GetNZC 0
	GetC
	GetNZ
%endmacro

;Get Negative, Zero, and Not Carry flags (for compare instructions)
%macro GetNZCs 0
	GetCs
	GetNZ
%endmacro

;Get Negative, Overflow, and Zero flags
%macro GetNVZ 0
	GetNZ
	SetO	[PSW+V]
%endmacro

;Get Negative, Overflow, Half-Carry, Zero, and Carry flags (for addition instructions)
%macro GetNVHZC 0
	GetC
	GetNVZ
%if HALFC
	Mov	DH,AH
	LAHF
	ShL	AH,4
	SetC	[PSW+H]
	Mov	AH,DH
%endif
%endmacro

;Get Negative, Overflow, Half-Carry, Zero, and Not Carry flags (for subtraction instructions)
%macro GetNVHZCs 0
	GetCs
	GetNVZ
%if HALFC
	Mov	DH,AH
	LAHF
	ShL	AH,4
	SetNC	[PSW+H]
	Mov	AH,DH
%endif
%endmacro


;
;Memory checkers
;   Certain sections of memory can't be written to or read from, or trigger events when read or
;   modified.  This is the part that really slows emulation down, for every instruction involving
;   memory multiple checks have to be performed.

;Peform functions when specific memory locations have been modified.
;   Opt - bit-7 ROM checking needs to be performed
;         bit-0 Data written was 16-bit
%macro WrPost 1

%if	%1 & 80h
	Cmp	BX,ipl										;;Was memory in the IPL ROM region modified?
	JAE	short %%WROM								;;  Yes
%endif

	Dec	BH
	Cmp	BX,0FFF0h									;;Was a function register written to?
%if DEBUG
	JAE	short %%WReg								;;  Yes, Jump to handler
		Jmp	EBP										;;  No, Jump to next opcode
	%%WReg:
%else
	JB		SPCFetch
%endif
		Mov	CL,BL										;;EBX->Function register handler
		XOr	EBX,EBX
		Mov	BH,CL
		Add	EBX,FuncBase-0F000h
%if	%1 & 1
		Push	EBX										;;Save low function register handler
		Add	EBX,100h									;;Jump to next function register
		Mov	EBP,FuncBase+10h
%endif
		Jmp	EBX

%if	%1 & 80h
	%%WROM:
%if IPLW
		Test	byte [tControl],80h					;;Is ROM reading enabled?
%if DEBUG
		JZ 	short %%WNext							;;  No, write was okay
%else
		JZ		SPCFetch
%endif
%endif
			Sub	BX,ipl								;;Convert BX to index within IPL ROM region
			MovZX	EBX,BX
%if IPLW
			Mov	CL,[EBX+RAM+ipl]					;;Get the byte written
			Mov	[EBX+extraRAM],CL
			Mov	CL,[EBX+iplROM]					;;Replace ROM byte
%else
			Mov	CL,[EBX+extraRAM]
%endif
			Mov	[EBX+RAM+ipl],CL
%if DEBUG
		%%WNext:
		Jmp	EBP
%else
		Jmp	SPCFetch
%endif
%endif

%endmacro

;Perform functions when specific memory locations have been read
;   Opt - bit-1 Next instruction should be fetched after check
;         bit-0 Data read was 16-bit

%macro ResetCnt 0
	Inc	BH												;;EBX -> Counter
%if SPEED
	Call	CntHack										;;Call speed hack
%else
	Mov	byte [EBX],0								;;Reset counter
%endif
%endmacro

%macro CheckDSP 0
%if DSPBK && DSPINTEG
	Cmp	BX,0F3h
	JNE	short %%NotF3
		Call	CatchUp
	%%NotF3:
%endif
%endmacro

%macro BreakDSP 0
%if DSPBK && (DSPINTEG == 0)
	Cmp	BX,0FFF3h
	JNE	short %%NotF3
		Push	EAX
		Mov	EBP,SPCExit
		Mov	EAX,[clkLeft]
		Add	[clkExec],EAX
		Sub	[clkLeft],EAX
		Pop	EAX
	%%NotF3:
%endif
%endmacro

%macro RdPost 1
	Dec	BH
	BreakDSP
	Cmp	BX,0FFFDh									;;Was a counter read from?

%if %1 & 2
%if DEBUG
	JAE	short %%RNext								;;  Yes
		Jmp	EBP										;;Jump back to fetch
	%%RNext:
	ResetCnt
	Jmp	EBP
%else
	JB		SPCFetch
	ResetCnt
	Jmp	SPCFetch
%endif
%else
	JB 	short %%RNext								;;  No, Continue with opcode
	ResetCnt
	%%RNext:
%endif

%endmacro


;
;Clean up after executing an instruction
;   Updates flags to reflect instruction, checks memory read/writes, and decreases the clock counter
;   the number of cycles it takes for the instruction to execute on the SPC700.  If the opcode has
;   any operands, PC is moved past them.

;   %1 = Number of clock cycles instruction takes to execute
;   %2 = Number of bytes in expression
;   %3 = Check memory
;        01h - perform post write checking
;        02h - perform post read checking
;        04h - 16-bit operation
;        08h - check absolute address
;   %4 = Flags to be updated
;        NZ, NZC(s), NVZ, NVZC(s)
%macro CleanUp 2-*

	%if %0 >= 4											;;Is flag paramater not blank?
		Get%4												;;  Yes, Get modified flags
	%endif

	%if %2 > 1											;;Is number of instruction bytes greater than 1?
		%rep %2-1										;;  Yes, Increase PC for each operand byte
			Inc	PC
		%endrep
	%endif

	Mov	EDX,%1*CPU_CYC								;;EDX = Amount to decrease Clock
;;	Mov	EDX,%1										;;(Incase I implement a dynamic divider)
;;	IMul	EDX,[cpuCyc]

	%if %0 >= 3											;;Is memory check parameter not blank?
		%ifidn %3,na									;;  Is parameter equal to nothing (0)?
%if DEBUG
			Jmp	EBP									;;    Yes, Fetch next opcode
%else
			Jmp	SPCFetch
%endif
		%endif
		%if %3 & 1										;;  Is write check bit in parameter set?
			%if %3 & 8									;;    Is it possible to write to IPL ROM region?
				WrPost 80h								;;      Yes, Check for ROM writes
			%else											;;      No, Just check for function register writes
				%if %3 & 4								;;      Was a 16-bit value written?
					WrPost 1								;;        Yes
				%else
					WrPost 0								;;        No
				%endif
			%endif
		%endif

		%if %3 & 2										;;  Is read check bit in parameter set?
			%if %3 & 4
				RdPost 3
			%else
				RdPost 2
			%endif
		%endif
	%else
%if DEBUG
		Jmp	EBP										;;  No, Grab next opcode
%else
		Jmp	SPCFetch
%endif
	%endif

%endmacro


;
;Stack Operations

;Pop byte off stack
;   Val - r/m to pop
%macro PopB 1
	Inc	byte [regSP]								;;Increase SP
	Mov	EDI,[regSP]									;;EDI -> Current stack position
	Mov	%1,[EDI]										;;Get value from stack
	XOr	DI,DI											;;EDI -> SPC RAM base
%endmacro

;Pop word off stack
;   Val - r/m to pop
%macro PopW 1
	Mov	EDI,[regSP]
	Mov	%1,[1+EDI]
	Add	byte [regSP],2
	XOr	DI,DI
%endmacro

;Push byte onto stack
;   Val - r/m to push
%macro PushB 1
	Mov	EDI,[regSP]
	Dec	byte [regSP]								;;Decrease SP
	Mov	[EDI],%1										;;Put value in stack
	XOr	DI,DI
%endmacro

;Push word onto stack
;   Val - r/m to push
%macro PushW 1
	Mov	EDI,[regSP]
	Mov	[EDI-1],%1
	Sub	byte [regSP],2
	XOr	DI,DI
%endmacro


;
;Opcode Handlers
;
;Ins - Desc
; Flags
;Form
;
;Ins:   Mnemonic for instruction
;
;Desc:  Long name for instruction
;
;Flags: Results of flag after execution
;       ? - Unknown
;       * - Reflects result of instruction
;       0 - Clear
;       1 - Set
;
;Form:  Format implemented
;
;
;Once the table is built, entrance can be called indirectly at EmuSPC.
;
;Upon entrance to an opcode handler, the registers will be:
;   AL  = SPC.A
;   AH  = SPC.Y
;   EBX = ?
;   CH  = SPC.X
;   EDX-> Entry point of current opcode handler
;   SI  = SPC.PC+1
;   ESI-> First byte after opcode
;   EDI-> Base of SPC RAM
;   EBP-> SPCFetch (SPCTrace if debugging is enabled)
;
;   CL, EDX, and EBX may be freely modified.  All other registers must reflect the results of the
;   operation or remain unchanged.
;
;
;Opcodes:
;
;   00 NOP     10 BPL     20 CLRP    30 BMI     40 SETP    50 BVC     60 CLRC    70 BVS
;   01 TCALL   11 TCALL   21 TCALL   31 TCALL   41 TCALL   51 TCALL   61 TCALL   71 TCALL
;   02 SET1    12 CLR1    22 SET1    32 CLR1    42 SET1    52 CLR1    62 SET1    72 CLR1
;   03 BBS     13 BBC     23 BBS     33 BBC     43 BBS     53 BBC     63 BBS     73 BBC
;   04 OR      14 OR      24 AND     34 AND     44 EOR     54 EOR     64 CMP     74 CMP
;   05 OR      15 OR      25 AND     35 AND     45 EOR     55 EOR     65 CMP     75 CMP
;   06 OR      16 OR      26 AND     36 AND     46 EOR     56 EOR     66 CMP     76 CMP
;   07 OR      17 OR      27 AND     37 AND     47 EOR     57 EOR     67 CMP     77 CMP
;   08 OR      18 OR      28 AND     38 AND     48 EOR     58 EOR     68 CMP     78 CMP
;   09 OR      19 OR      29 AND     39 AND     49 EOR     59 EOR     69 CMP     79 CMP
;   0A OR1     1A DECW    2A OR1     3A INCW    4A AND1    5A CMPW    6A AND1    7A ADDW
;   0B ASL     1B ASL     2B ROL     3B ROL     4B LSR     5B LSR     6B ROR     7B ROR
;   0C ASL     1C ASL     2C ROL     3C ROL     4C LSR     5C LSR     6C ROR     7C ROR
;   0D PUSH    1D DEC     2D PUSH    3D INC     4D PUSH    5D MOV     6D PUSH    7D MOV
;   0E TSET1   1E CMP     2E CBNE    3E CMP     4E TCLR1   5E CMP     6E DBNZ    7E CMP
;   0F BRK     1F JMP     2F BRA     3F CALL    4F PCALL   5F JMP     6F RET     7F RETI
;
;   80 SETC    90 BCC     A0 EI      B0 BCS     C0 DI      D0 BNE     E0 CLRV    F0 BEQ
;   81 TCALL   91 TCALL   A1 TCALL   B1 TCALL   C1 TCALL   D1 TCALL   E1 TCALL   F1 TCALL
;   82 SET1    92 CLR1    A2 SET1    B2 CLR1    C2 SET1    D2 CLR1    E2 SET1    F2 CLR1
;   83 BBS     93 BBC     A3 BBS     B3 BBC     C3 BBS     D3 BBC     E3 BBS     F3 BBC
;   84 ADC     94 ADC     A4 SBC     B4 SBC     C4 MOV     D4 MOV     E4 MOV     F4 MOV
;   85 ADC     95 ADC     A5 SBC     B5 SBC     C5 MOV     D5 MOV     E5 MOV     F5 MOV
;   86 ADC     96 ADC     A6 SBC     B6 SBC     C6 MOV     D6 MOV     E6 MOV     F6 MOV
;   87 ADC     97 ADC     A7 SBC     B7 SBC     C7 MOV     D7 MOV     E7 MOV     F7 MOV
;   88 ADC     98 ADC     A8 SBC     B8 SBC     C8 CMP     D8 MOV     E8 MOV     F8 MOV
;   89 ADC     99 ADC     A9 SBC     B9 SBC     C9 MOV     D9 MOV     E9 MOV     F9 MOV
;   8A EOR1    9A SUBW    AA MOV1    BA MOVW    CA MOV1    DA MOVW    EA NOT1    FA MOV
;   8B DEC     9B DEC     AB INC     BB INC     CB MOV     DB MOV     EB MOV     FB MOV
;   8C DEC     9C DEC     AC INC     BC INC     CC MOV     DC DEC     EC MOV     FC INC
;   8D MOV     9D MOV     AD CMP     BD MOV     CD MOV     DD MOV     ED NOTC    FD MOV
;   8E POP     9E DIV     AE POP     BE DAA     CE POP     DE CBNE    EE POP     FE DBNZ
;   8F MOV     9F XCN     AF MOV     BF MOV     CF MUL     DF DAS     EF SLEEP   FF STOP


;
;NOP - No Operation
; N V P B H I Z C
;
;NOp
%macro Opc00 0
	CleanUp	2,1										;Do nothing
%endmacro


;
;ADC - Add with Carry
; N V P B H I Z C
; * *     *   * *
;AdC A,#imm
%macro Opc88 0
	ShR	byte [PSW+CF],1
	AdC	A,[OP1]
	CleanUp	2,2,na,NVHZC
%endmacro

;AdC A,dp
%macro Opc84 0
	Ldp
	CheckDSP
	ShR	byte [PSW+CF],1
	AdC	A,[DPI]
	CleanUp	3,2,RD,NVHZC
%endmacro

;AdC A,dp+X
%macro Opc94 0
	LdpX
	CheckDSP
	ShR	byte [PSW+CF],1
	AdC	A,[DPI]
	CleanUp	4,2,RD,NVHZC
%endmacro

;AdC A,abs
%macro Opc85 0
	Labs
	CheckDSP
	ShR	byte [PSW+CF],1
	AdC	A,[ABSL]
	CleanUp	4,3,RA,NVHZC
%endmacro

;AdC A,abs+X
%macro Opc95 0
	LabsX
	CheckDSP
	ShR	byte [PSW+CF],1
	AdC	A,[ABSL]
	CleanUp	5,3,RA,NVHZC
%endmacro

;AdC A,abs+Y
%macro Opc96 0
	LabsY
	CheckDSP
	ShR	byte [PSW+CF],1
	AdC	A,[ABSL]
	CleanUp	5,3,RA,NVHZC
%endmacro

;AdC A,(X)
%macro Opc86 0
	LX
	CheckDSP
	ShR	byte [PSW+CF],1
	AdC	A,[DPI]
	CleanUp	3,1,RD,NVHZC
%endmacro

;AdC A,[dp+X]
%macro Opc87 0
	LadpX
	CheckDSP
	ShR	byte [PSW+CF],1
	AdC	A,[ABSL]
	CleanUp	6,2,RA,NVHZC
%endmacro

;AdC A,[dp]+Y
%macro Opc97 0
	LadpY
	CheckDSP
	ShR	byte [PSW+CF],1
	AdC	A,[ABSL]
	CleanUp	6,2,RA,NVHZC
%endmacro

;AdC dp,#imm
%macro Opc98 0
	Mov	DH,[OP1]
	Ldp2
	ShR	byte [PSW+CF],1
	AdC	[DPI],DH
	CleanUp	5,3,WD,NVHZC
%endmacro

;AdC dp,dp
%macro Opc89 0
	Ldp
	CheckDSP
	Mov	DH,[DPI]
	RdPost 0
	Ldp2
	ShR	byte [PSW+CF],1
	AdC	[DPI],DH
	CleanUp	6,3,WD,NVHZC
%endmacro

;AdC (X),(Y)
%macro Opc99 0
	LY
	CheckDSP
	Mov	DH,[DPI]
	RdPost 0
	LX
	ShR	byte [PSW+CF],1
	AdC	[DPI],DH
	CleanUp	5,1,WD,NVHZC
%endmacro


;
;ADDW - Add Word
; N V P B H I Z C
; * *     *   * *
;AddW YA,dp
%macro Opc7A 0
	Ldp
	Add	YA,[DPI]
	CleanUp	5,2,RD16,NVHZC
%endmacro


;
;AND - Bit-wise Logical And
; N V P B H I Z C
; *           *
;And A,#imm
%macro Opc28 0
	And	A,[OP1]
	CleanUp	2,2,na,NZ
%endmacro

;And A,dp
%macro Opc24 0
	Ldp
	CheckDSP
	And	A,[DPI]
	CleanUp	3,2,RD,NZ
%endmacro

;And A,dp+X
%macro Opc34 0
	LdpX
	CheckDSP
	And	A,[DPI]
	CleanUp	4,2,RD,NZ
%endmacro

;And A,abs
%macro Opc25 0
	Labs
	CheckDSP
	And	A,[ABSL]
	CleanUp	4,3,RA,NZ
%endmacro

;And A,abs+X
%macro Opc35 0
	LabsX
	CheckDSP
	And	A,[ABSL]
	CleanUp	5,3,RA,NZ
%endmacro

;And A,abs+Y
%macro Opc36 0
	LabsY
	CheckDSP
	And	A,[ABSL]
	CleanUp	5,3,RA,NZ
%endmacro

;And A,(X)
%macro Opc26 0
	LX
	CheckDSP
	And	A,[DPI]
	CleanUp	3,1,RD,NZ
%endmacro

;And A,[dp+X]
%macro Opc27 0
	LadpX
	CheckDSP
	And	A,[ABSL]
	CleanUp	6,2,RA,NZ
%endmacro

;And A,[dp]+Y
%macro Opc37 0
	LadpY
	CheckDSP
	And	A,[ABSL]
	CleanUp	6,2,RA,NZ
%endmacro

;And dp,#imm
%macro Opc38 0
	Ldp2
	Mov	DH,[OP1]
	And	[DPI],DH
	CleanUp	5,3,WD,NZ
%endmacro

;And dp,dp
%macro Opc29 0
	Ldp
	CheckDSP
	Mov	DH,[DPI]
	RdPost 0
	Ldp2
	And	[DPI],DH
	CleanUp	6,3,WD,NZ
%endmacro

;And (X),(Y)
%macro Opc39 0
	LY
	CheckDSP
	Mov	DH,[DPI]
	RdPost 0
	LX
	And	[DPI],DH
	CleanUp	5,1,WD,NZ
%endmacro


;
;AND1 - And Carry with Absolute Bit
; N V P B H I Z C
;               *
;And1 C,mem.bit
%macro Opc4A 0
	Cmp	byte [PSW+CF],0							;Is carry zero?
	JZ 	short %%Nc									;  Yes, Result will be zero anyway so quit
		Lmbit
		Mov	BL,[ABSL]
		BT 	EBX,EDX									;(Bit Test only uses the first five bits of EDX)
		CleanUp	4,3,na,C
ALIGN 16
	%%Nc:
	CleanUp	4,3
%endmacro

;And1 C,/mem.bit
%macro Opc6A 0
	Cmp	byte [PSW+CF],0
	JZ 	short %%Ncn
		Lmbit
		Mov	BL,[ABSL]
		BT 	EBX,EDX
		CleanUp	4,3,na,Cs
ALIGN 16
	%%Ncn:
	CleanUp	4,3
%endmacro


;
;ASL - Arithmetic Shift Left
; N V P B H I Z C
; *           * *
;ASL A
%macro Opc1C 0
	ShL	A,1
	CleanUp	2,1,na,NZC
%endmacro

;ASL dp
%macro Opc0B 0
	Ldp
	ShL	byte [DPI],1
	CleanUp	4,2,WD,NZC
%endmacro

;ASL dp+X
%macro Opc1B 0
	LdpX
	ShL	byte [DPI],1
	CleanUp	5,2,WD,NZC
%endmacro

;ASL abs
%macro Opc0C 0
	Labs
	ShL	byte [ABSL],1
	CleanUp	5,3,WA,NZC
%endmacro


;
;BBC - Branch If Bit Clear
; N V P B H I Z C

	;-----------------------------------------
	;Branch if bit clear
	;   %1 - Bit to test (0-7)

	%macro BBC 1
		Ldp												;;Load DP pointer
		Test	byte [DPI],1 << %1					;;Test requested bit, is it clear?
		JNZ	short %%BCDone							;;  No, Clean up
			MovSX	EDX,byte [OP2]						;;EDX = Relative adjustment
			Add	PC,DX									;;Adjust PC
			CleanUp	7,3,RD
	ALIGN 16
		%%BCDone:
		CleanUp	5,3,RD
	%endmacro

;BBC dp.0,rel
%macro Opc13 0
	BBC	0
%endmacro

;BBC dp.1,rel
%macro Opc33 0
	BBC	1
%endmacro

;BBC dp.2,rel
%macro Opc53 0
	BBC	2
%endmacro

;BBC dp.3,rel
%macro Opc73 0
	BBC	3
%endmacro

;BBC dp.4,rel
%macro Opc93 0
	BBC	4
%endmacro

;BBC dp.5,rel
%macro OpcB3 0
	BBC	5
%endmacro

;BBC dp.6,rel
%macro OpcD3 0
	BBC	6
%endmacro

;BBC dp.7,rel
%macro OpcF3 0
	BBC	7
%endmacro


;
;BBS - Branch If Bit Set
; N V P B H I Z C

	;-----------------------------------------
	;Branch if bit set
	;   %1 - Bit to test (0-7)

	%macro BBS 1
		Ldp
		Test	byte [DPI],1 << %1					;;Test the requested bit, is it set?
		JZ 	short %%BSDone							;;  No, Clean up
			MovSX	EDX,byte [OP2]						;;DX = Relative adjustment
			Add	PC,DX									;;Add relative displacement to PC
			CleanUp	7,3,RD
	ALIGN 16
		%%BSDone:
		CleanUp	5,3,RD
	%endmacro

;BBS dp.0,rel
%macro Opc03 0
	BBS	0
%endmacro

;BBS dp.1,rel
%macro Opc23 0
	BBS	1
%endmacro

;BBS dp.2,rel
%macro Opc43 0
	BBS	2
%endmacro

;BBS dp.3,rel
%macro Opc63 0
	BBS	3
%endmacro

;BBS dp.4,rel
%macro Opc83 0
	BBS	4
%endmacro

;BBS dp.5,rel
%macro OpcA3 0
	BBS	5
%endmacro

;BBS dp.6,rel
%macro OpcC3 0
	BBS	6
%endmacro

;BBS dp.7,rel
%macro OpcE3 0
	BBS	7
%endmacro


;
;Bxx - Conditional Branch
; N V P B H I Z C

	;-----------------------------------------
	;Branch on condition
	;   %1 - Flag to Test
	;   %2 - Condition of flag (0 or 1)

	%macro Bxx 2
		Test	byte [PSW+%1],1
		%if %2 == 1
			JZ 	short %%BxDone
		%else
			JNZ	short %%BxDone
		%endif
			MovSX	EBX,byte [OP1]
			Add	PC,BX
			CleanUp	4,2
	ALIGN 16
		%%BxDone:
		CleanUp	2,2
	%endmacro

;BCC rel - Branch if Carry Clear (JAE)
%macro Opc90 0
	Bxx	CF,0
%endmacro

;BCS rel - Branch if Carry Set (JB)
%macro OpcB0 0
	Bxx	CF,1
%endmacro

;BEQ rel - Branch if Equal (JE/JZ)
%macro OpcF0 0
	Bxx	Z,1
%endmacro

;BMI rel - Branch if Minus (JS)
%macro Opc30 0
	Bxx	N,1
%endmacro

;BNE rel - Branch if Not Equal (JNE/JNZ)
%macro OpcD0 0
	Bxx	Z,0
%endmacro

;BPL rel - Branch if Plus (JNS)
%macro Opc10 0
	Bxx	N,0
%endmacro

;BVC rel - Branch if Overflow Clear (JNO)
%macro Opc50 0
	Bxx	V,0
%endmacro

;BVS rel - Branch if Overflow Set (JO)
%macro Opc70 0
	Bxx	V,1
%endmacro


;
;BRA - Branch (JMP Short)
; N V P B H I Z C

;BRA rel
%macro Opc2F 0
	MovSX	EBX,byte [OP1]
	Add	PC,BX
	CleanUp	4,2
%endmacro


;
;BRK - Software Interrupt
; N V P B H I Z C
;       1   0
;Brk
%macro Opc0F 0
	Mov		byte [PSW+B],1
	Mov		byte [PSW+I],0
	PushW		PC
	CmpPSW
	PushB		PS
	Mov		PC,[0FFDEh+RAM]
	CleanUp	8,1
%endmacro


;
;CALL - Call Procedure
; N V P B H I Z C

;Call abs
%macro Opc3F 0
	Add		PC,2
	PushW		PC
	Mov		PC,[ESI-2]
	CleanUp	8,na
%endmacro


;
;CBNE - Compare with A, Branch if Not Equal
; N V P B H I Z C

;CBNE dp,rel
%macro Opc2E 0
	Ldp
	CheckDSP
	Cmp	A,[DPI]
	JE 	short %%NCBdp
		MovSX	EDX,byte [OP2]
		Add	PC,DX
		CleanUp	7,3,RD
	ALIGN 16
	%%NCBdp:
	CleanUp	5,3,RD
%endmacro

;CBNE dp+X,rel
%macro OpcDE 0
	LdpX
	CheckDSP
	Cmp	A,[DPI]
	JE 	short %%NCBdpx
		MovSX	EDX,byte [OP2]
		Add	PC,DX
		CleanUp	8,3,RD
	ALIGN 16
	%%NCBdpx:
	CleanUp	6,3,RD
%endmacro


;
;CLR1 - Clear Bit
; N V P B H I Z C

	;-----------------------------------------
	;Clear bit
	;   %1 - Bit to clear (0-7)

	%macro Clr 1
		Ldp
		And	byte [DPI],~(1 << %1)
		CleanUp	4,2,WD
	%endmacro

;Clr1 dp.0
%macro Opc12 0
	Clr	0
%endmacro

;Clr1 dp.1
%macro Opc32 0
	Clr	1
%endmacro

;Clr1 dp.2
%macro Opc52 0
	Clr	2
%endmacro

;Clr1 dp.3
%macro Opc72 0
	Clr	3
%endmacro

;Clr1 dp.4
%macro Opc92 0
	Clr	4
%endmacro

;Clr1 dp.5
%macro OpcB2 0
	Clr	5
%endmacro

;Clr1 dp.6
%macro OpcD2 0
	Clr	6
%endmacro

;Clr1 dp.7
%macro OpcF2 0
	Clr	7
%endmacro


;
;CLRC - Clear Carry Flag
; N V P B H I Z C
;               0
;ClrC
%macro Opc60 0
	Mov	byte [PSW+CF],0
	CleanUp	2,1
%endmacro


;
;CLRP - Clear Direct Page Flag
; N V P B H I Z C
;     0
;ClrP
%macro Opc20 0
	Mov	byte [PSW+P],0
	CleanUp	2,1
%endmacro


;
;CLRV - Clear Overflow and Half-carry Flags
; N V P B H I Z C
;   0     0
;ClrV
%macro OpcE0 0
	Mov	byte [PSW+V],0
	Mov	byte [PSW+H],0
	CleanUp	2,1
%endmacro


;
;CMP - Compare Two Operands
; N V P B H I Z C
; *           * *
;Cmp A,#imm
%macro Opc68 0
	Cmp	A,[OP1]
	CleanUp	2,2,na,NZCs
%endmacro

;Cmp X,#imm
%macro OpcC8 0
	Cmp	X,[OP1]
	CleanUp	2,2,na,NZCs
%endmacro

;Cmp Y,#imm
%macro OpcAD 0
	Cmp	Y,[OP1]
	CleanUp	2,2,na,NZCs
%endmacro

;Cmp A,dp
%macro Opc64 0
	Ldp
	CheckDSP
	Cmp	A,[DPI]
	CleanUp	3,2,RD,NZCs
%endmacro

;Cmp X,dp
%macro Opc3E 0
	Ldp
	CheckDSP
	Cmp	X,[DPI]
	CleanUp	3,2,RD,NZCs
%endmacro

;Cmp Y,dp
%macro Opc7E 0
	Ldp
	CheckDSP
	Cmp	Y,[DPI]
	CleanUp	3,2,RD,NZCs
%endmacro

;Cmp A,dp+X
%macro Opc74 0
	LdpX
	CheckDSP
	Cmp	A,[DPI]
	CleanUp	4,2,RD,NZCs
%endmacro

;Cmp A,abs
%macro Opc65 0
	Labs
	CheckDSP
	Cmp	A,[ABSL]
	CleanUp	4,3,RA,NZCs
%endmacro

;Cmp X,abs
%macro Opc1E 0
	Labs
	CheckDSP
	Cmp	X,[ABSL]
	CleanUp	4,3,RA,NZCs
%endmacro

;Cmp Y,abs
%macro Opc5E 0
	Labs
	CheckDSP
	Cmp	Y,[ABSL]
	CleanUp	4,3,RA,NZCs
%endmacro

;Cmp A,abs+X
%macro Opc75 0
	LabsX
	CheckDSP
	Cmp	A,[ABSL]
	CleanUp	5,3,RA,NZCs
%endmacro

;Cmp A,abs+Y
%macro Opc76 0
	LabsY
	CheckDSP
	Cmp	A,[ABSL]
	CleanUp	5,3,RA,NZCs
%endmacro

;Cmp A,(X)
%macro Opc66 0
	LX
	CheckDSP
	Cmp	A,[DPI]
	CleanUp	3,1,RD,NZCs
%endmacro

;Cmp A,[dp+X]
%macro Opc67 0
	LadpX
	CheckDSP
	Cmp	A,[ABSL]
	CleanUp	6,2,RA,NZCs
%endmacro

;Cmp A,[dp]+Y
%macro Opc77 0
	LadpY
	CheckDSP
	Cmp	A,[ABSL]
	CleanUp	6,2,RA,NZCs
%endmacro

;Cmp dp,#imm
%macro Opc78 0
	Ldp2
	Mov	DH,[OP1]
	Cmp	[DPI],DH
	CleanUp	5,3,RD,NZCs
%endmacro

;Cmp dp,dp
%macro Opc69 0
	Ldp
	CheckDSP
	Mov	DH,[DPI]
	RdPost 0
	Ldp2
	Cmp	[DPI],DH
	CleanUp	6,3,RD,NZCs
%endmacro

;Cmp (X),(Y)
%macro Opc79 0
	LY
	CheckDSP
	Mov	DH,[DPI]
	RdPost 0
	LX
	Cmp	[DPI],DH
	CleanUp	5,1,RD,NZCs
%endmacro


;
;CMPW - Compare Two Words
; N V P B H I Z C
; *           * *
;CmpW YA,dp
%macro Opc5A 0
	Ldp
	Cmp	YA,[DPI]
	CleanUp	4,2,RD16,NZCs
%endmacro


;
;DAA - Decimal Adjust After Addition
; N V P B H I Z C
; *           * *
;DAA A
%macro OpcDF 0
	Mov	DH,AH											;Save AH (Y)
	Mov	AH,[PSW+H]									;Set AF and CF in AH
	ShL	AH,4
	Or 	AH,[PSW+CF]
	SAHF													;Store AH into flags register
	Mov	AH,DH											;Restore AH
	DAA													;Execute DAA on AL (A)
	CleanUp	3,1,na,NZC
%endmacro


;
;DAS - Decimal Adjust After Subtraction
; N V P B H I Z C
; *           * *
;DAS A
%macro OpcBE 0
	Mov	DH,AH
	Mov	AH,[PSW+H]
	ShL	AH,4
	Or 	AH,[PSW+CF]
	XOr	AH,11h										;Reverse flags for x86
	SAHF
	Mov	AH,DH
	DAS
	CleanUp	3,1,na,NZCs
%endmacro


;
;DBNZ - Decrease byte, Branch if not Zero
; N V P B H I Z C

;DBNZ Y,rel
%macro OpcFE 0
	Dec	Y
	JZ 	short %%NDBy
		MovSX	EBX,byte [OP1]
		Add	PC,BX
		CleanUp	6,2
	ALIGN 16
	%%NDBy:
	CleanUp	4,2
%endmacro

;DBNZ dp,rel
%macro Opc6E 0
	Ldp
	Dec	byte [DPI]
	JZ 	short %%NDBdp
		MovSX	EDX,byte [OP2]
		Add	PC,DX
		CleanUp	7,3,WD
	ALIGN 16
	%%NDBdp:
	CleanUp	5,3,WD
%endmacro


;
;DEC - Decrease Byte by 1
; N V P B H I Z C
; *           *
;Dec A
%macro Opc9C 0
	Dec	A
	CleanUp	2,1,na,NZ
%endmacro

;Dec X
%macro Opc1D 0
	Dec	X
	CleanUp	2,1,na,NZ
%endmacro

;Dec Y
%macro OpcDC 0
	Dec	Y
	CleanUp	2,1,na,NZ
%endmacro

;Dec dp
%macro Opc8B 0
	Ldp
	Dec	byte [DPI]
	CleanUp	4,2,WD,NZ
%endmacro

;Dec dp+X
%macro Opc9B 0
	LdpX
	Dec	byte [DPI]
	CleanUp	5,2,WD,NZ
%endmacro

;Dec abs
%macro Opc8C 0
	Labs
	Dec	byte [ABSL]
	CleanUp	5,3,WA,NZ
%endmacro


;
;DECW - Decrease Word by 1
; N V P B H I Z C
; *           *
;DecW dp
%macro Opc1A 0
	Ldp
	Dec	word [DPI]
	CleanUp	6,2,WD16,NZ
%endmacro


;
;DI - Disable Interrupts
; N V P B H I Z C
;           0
;DI
%macro OpcC0 0
	Mov	byte [PSW+I],0
	CleanUp	3,1
%endmacro


;
;DIV - Divide
; N V P B H I Z C
; * *     *   *
;Div YA,X
%macro Opc9E 0
	Test	X,X											;Is divisor 0?
	JZ 	short %%Div0									;  Yes
		XOr	EDX,EDX									;DX:AX = YA
		MovZX	EBX,X										;EBX = X
		Div	BX
		Test	AH,AH										;Did the divide overflow?
		JNZ	short %%DivOF							;  Yes

		Mov	byte [PSW+H],0							;H and V are indeterminate after a divide, so I just
		Mov	byte [PSW+V],0							;  clear them.

		Mov	Y,DL
		Test	A,A
		CleanUp	12,1,na,NZ

ALIGN 16
	%%Div0:
		Mov	byte [PSW+V],0
		Mov	byte [PSW+H],0
		Mov	byte [PSW+N],0
		Mov	byte [PSW+Z],0
		Mov	YA,0FFh
		CleanUp	12,1,na

ALIGN 16
	%%DivOF:
		Mov	byte [PSW+V],1
		Mov	byte [PSW+H],1
		Test	AX,AX
		Mov	Y,DL
		CleanUp	12,1,na,NZ
%endmacro


;
;EI - Enable Interrupts
; N V P B H I Z C
;           1
;EI
%macro OpcA0 0
	Mov	byte [PSW+I],1
	CleanUp	3,1
%endmacro


;
;EOR - Bit-wise Logical Exclusive Or
; N V P B H I Z C
; *           *
;EOr A,#imm
%macro Opc48 0
	XOr	A,[OP1]
	CleanUp	2,2,na,NZ
%endmacro

;EOr A,dp
%macro Opc44 0
	Ldp
	CheckDSP
	XOr	A,[DPI]
	CleanUp	3,2,RD,NZ
%endmacro

;EOr A,dp+X
%macro Opc54 0
	LdpX
	CheckDSP
	XOr	A,[DPI]
	CleanUp	4,2,RD,NZ
%endmacro

;EOr A,abs
%macro Opc45 0
	Labs
	CheckDSP
	XOr	A,[ABSL]
	CleanUp	4,3,RA,NZ
%endmacro

;EOr A,abs+X
%macro Opc55 0
	LabsX
	CheckDSP
	XOr	A,[ABSL]
	CleanUp	5,3,RA,NZ
%endmacro

;EOr A,abs+Y
%macro Opc56 0
	LabsY
	CheckDSP
	XOr	A,[ABSL]
	CleanUp	5,3,RA,NZ
%endmacro

;EOr A,(X)
%macro Opc46 0
	LX
	CheckDSP
	XOr	A,[DPI]
	CleanUp	3,1,RD,NZ
%endmacro

;EOr A,[dp+X]
%macro Opc47 0
	LadpX
	CheckDSP
	XOr	A,[ABSL]
	CleanUp	6,2,RA,NZ
%endmacro

;EOr A,[dp]+Y
%macro Opc57 0
	LadpY
	CheckDSP
	XOr	A,[ABSL]
	CleanUp	6,2,RA,NZ
%endmacro

;EOr dp,#imm
%macro Opc58 0
	Ldp2
	Mov	DH,[OP1]
	XOr	[DPI],DH
	CleanUp	5,3,WD,NZ
%endmacro

;EOr dp,dp
%macro Opc49 0
	Ldp
	CheckDSP
	Mov	DH,[DPI]
	RdPost 0
	Ldp2
	XOr	[DPI],DH
	CleanUp	6,3,WD,NZ
%endmacro

;EOr (X),(Y)
%macro Opc59 0
	LY
	CheckDSP
	Mov	DH,[DPI]
	RdPost 0
	LX
	XOr	[DPI],DH
	CleanUp	5,1,WD,NZ
%endmacro


;
;EOR1 - XOr Carry with Absolute Bit
; N V P B H I Z C
;               *
;EOr1 C,mem.bit
%macro Opc8A 0
	Lmbit
	Push	EBX
	Mov	BL,[ABSL]
	BT 	EBX,EDX
	AdC	byte [PSW+CF],0
	And	byte [PSW+CF],1
	Pop	EBX
	CleanUp	5,3,RA
%endmacro


;
;INC - Increase Byte by 1
; N V P B H I Z C
; *           *
;Inc A
%macro OpcBC 0
	Inc	A
	CleanUp	2,1,na,NZ
%endmacro

;Inc X
%macro Opc3D 0
	Inc	X
	CleanUp	2,1,na,NZ
%endmacro

;Inc Y
%macro OpcFC 0
	Inc	Y
	CleanUp	2,1,na,NZ
%endmacro

;Inc dp
%macro OpcAB 0
	Ldp
	Inc	byte [DPI]
	CleanUp	4,2,WD,NZ
%endmacro

;Inc dp+X
%macro OpcBB 0
	LdpX
	Inc	byte [DPI]
	CleanUp	5,2,WD,NZ
%endmacro

;Inc abs
%macro OpcAC 0
	Labs
	Inc	byte [ABSL]
	CleanUp	5,3,WA,NZ
%endmacro


;
;INCW - Increase Word by 1
; N V P B H I Z C
; *           *
;INCW dp
%macro Opc3A 0
	Ldp
	Inc	word [DPI]
	CleanUp	6,2,WD16,NZ
%endmacro


;
;JMP - Indirect Jump
; N V P B H I Z C

;Jmp abs
%macro Opc5F 0
	Mov	PC,[OP1]
	CleanUp	3,na
%endmacro

;Jmp [abs+X]
%macro Opc1F 0
	LaabsX
	Mov	PC,[ABSL]
	CleanUp	6,na
%endmacro


;
;LSR - Logical Shift Right
; N V P B H I Z C
; *           * *
;LSR A
%macro Opc5C 0
	ShR	A,1
	CleanUp	2,1,na,NZC
%endmacro

;LSR dp
%macro Opc4B 0
	Ldp
	ShR	byte [DPI],1
	CleanUp	4,2,WD,NZC
%endmacro

;LSR dp+X
%macro Opc5B 0
	LdpX
	ShR	byte [DPI],1
	CleanUp	5,2,WD,NZC
%endmacro

;LSR abs
%macro Opc4C 0
	Labs
	ShR	byte [ABSL],1
	CleanUp	5,3,WA,NZC
%endmacro


;
;MOV - Load Immediate
; N V P B H I Z C
; *           *
;Mov A,#imm
%macro OpcE8 0
	Mov	A,[OP1]
	Test	A,A
	CleanUp	2,2,na,NZ
%endmacro

;Mov X,#imm
%macro OpcCD 0
	Mov	X,[OP1]
	Test	X,X
	CleanUp	2,2,na,NZ
%endmacro

;Mov Y,#imm
%macro Opc8D 0
	Mov	Y,[OP1]
	Test	Y,Y
	CleanUp	2,2,na,NZ
%endmacro


;
;MOV - Copy Register
; N V P B H I Z C
; *           *
;Mov A,X
%macro Opc7D 0
	Mov	A,X
	Test	A,A
	CleanUp	2,1,na,NZ
%endmacro

;Mov A,Y
%macro OpcDD 0
	Mov	A,Y
	Test	A,A
	CleanUp	2,1,na,NZ
%endmacro

;Mov X,A
%macro Opc5D 0
	Mov	X,A
	Test	X,X
	CleanUp	2,1,na,NZ
%endmacro

;Mov Y,A
%macro OpcFD 0
	Mov	Y,A
	Test	Y,Y
	CleanUp	2,1,na,NZ
%endmacro

;Mov X,SP
%macro Opc9D 0
	Mov	X,byte [regSP]
	Test	X,X
	CleanUp	2,1,na,NZ
%endmacro

;Mov SP,X (flags not affected)
%macro OpcBD 0
	Mov	byte [regSP],X
	CleanUp	2,1
%endmacro


;
;MOV - Load Memory Into Register
; N V P B H I Z C
; *           *
;Mov A,dp
%macro OpcE4 0
	Ldp
	CheckDSP
 	Mov	A,[DPI]
	Test	A,A
	CleanUp	3,2,RD,NZ
%endmacro

;Mov X,dp
%macro OpcF8 0
	Ldp
	CheckDSP
	Mov	X,[DPI]
	Test	X,X
	CleanUp	3,2,RD,NZ
%endmacro

;Mov Y,dp
%macro OpcEB 0
	Ldp
	CheckDSP
	Mov	Y,[DPI]
	Test	Y,Y
	CleanUp	3,2,RD,NZ
%endmacro

;Mov A,dp+X
%macro OpcF4 0
	LdpX
	CheckDSP
	Mov	A,[DPI]
	Test	A,A
	CleanUp	4,2,RD,NZ
%endmacro

;Mov X,dp+Y
%macro OpcF9 0
	LdpY
	CheckDSP
	Mov	X,[DPI]
	Test	X,X
	CleanUp	4,2,RD,NZ
%endmacro

;Mov Y,dp+X
%macro OpcFB 0
	LdpX
	CheckDSP
	Mov	Y,[DPI]
	Test	Y,Y
	CleanUp	4,2,RD,NZ
%endmacro

;Mov A,abs
%macro OpcE5 0
	Labs
	CheckDSP
	Mov	A,[ABSL]
	Test	A,A
	CleanUp	4,3,RA,NZ
%endmacro

;Mov X,abs
%macro OpcE9 0
	Labs
	CheckDSP
	Mov	X,[ABSL]
	Test	X,X
	CleanUp	4,3,RA,NZ
%endmacro

;Mov Y,abs
%macro OpcEC 0
	Labs
	CheckDSP
	Mov	Y,[ABSL]
	Test	Y,Y
	CleanUp	4,3,RA,NZ
%endmacro

;Mov A,abs+X
%macro OpcF5 0
	LabsX
	CheckDSP
	Mov	A,[ABSL]
	Test	A,A
	CleanUp	5,3,RA,NZ
%endmacro

;Mov A,abs+Y
%macro OpcF6 0
	LabsY
	CheckDSP
	Mov	A,[ABSL]
	Test	A,A
	CleanUp	5,3,RA,NZ
%endmacro

;Mov A,(X)
%macro OpcE6 0
	LX
	CheckDSP
	Mov	A,[DPI]
	Test	A,A
	CleanUp	3,1,RD,NZ
%endmacro

;Mov A,[dp+X]
%macro OpcE7 0
	LadpX
	CheckDSP
	Mov	A,[ABSL]
	Test	A,A
	CleanUp	6,2,RA,NZ
%endmacro

;Mov A,[dp]+Y
%macro OpcF7 0
	LadpY
	CheckDSP
	Mov	A,[ABSL]
	Test	A,A
	CleanUp	6,2,RA,NZ
%endmacro


;
;MOV - Store Register in Memory
; N V P B H I Z C

;Mov dp,A
%macro OpcC4 0
	Ldp
	Mov	[DPI],A
	CleanUp	4,2,WD
%endmacro

;Mov dp,X
%macro OpcD8 0
	Ldp
	Mov	[DPI],X
	CleanUp	4,2,WD
%endmacro

;Mov dp,Y
%macro OpcCB 0
	Ldp
	Mov	[DPI],Y
	CleanUp	4,2,WD
%endmacro

;Mov dp+X,A
%macro OpcD4 0
	LdpX
	Mov	[DPI],A
	CleanUp	5,2,WD
%endmacro

;Mov dp+Y,X
%macro OpcD9 0
	LdpY
	Mov	[DPI],X
	CleanUp	5,2,WD
%endmacro

;Mov dp+X,Y
%macro OpcDB 0
	LdpX
	Mov	[DPI],Y
	CleanUp	5,2,WD
%endmacro

;Mov abs,A
%macro OpcC5 0
	Labs
	Mov	[ABSL],A
	CleanUp	5,3,WA
%endmacro

;Mov abs,X
%macro OpcC9 0
	Labs
	Mov	[ABSL],X
	CleanUp	5,3,WA
%endmacro

;Mov abs,Y
%macro OpcCC 0
	Labs
	Mov	[ABSL],Y
	CleanUp	5,3,WA
%endmacro

;Mov abs+X,A
%macro OpcD5 0
	LabsX
	Mov	[ABSL],A
	CleanUp	6,3,WA
%endmacro

;Mov abs+Y,A
%macro OpcD6 0
	LabsY
	Mov	[ABSL],A
	CleanUp	6,3,WA
%endmacro

;Mov [dp+X],A
%macro OpcC7 0
	LadpX
	Mov	[ABSL],A
	CleanUp	7,2,WA
%endmacro

;Mov [dp]+Y,A
%macro OpcD7 0
	LadpY
	Mov	[ABSL],A
	CleanUp	7,2,WA
%endmacro

;Mov (X),A
%macro OpcC6 0
	LX
	Mov	[DPI],A
	CleanUp	4,1,WD
%endmacro


;
;MOV - Move Data to Memory
; N V P B H I Z C

;Mov dp,#imm
%macro Opc8F 0
	Ldp2
	Mov	DH,[OP1]
	Mov	[DPI],DH
	CleanUp	5,3,WD
%endmacro

;Mov dp,dp
%macro OpcFA 0
	Ldp
	Mov		DH,[DPI]
	RdPost	0
	Ldp2
	Mov		[DPI],DH
	CleanUp	5,3,WD
%endmacro


;
;MOV - Load Byte with Auto Increase
; N V P B H I Z C
; *           *
;Mov A,(X)+
%macro OpcBF 0
	LX
	CheckDSP
	Inc	X
	Mov	A,[DPI]
	Test	A,A
	CleanUp	4,1,RD,NZ
%endmacro


;
;MOV - Store Byte with Auto Increase
; N V P B H I Z C

;Mov (X)+,A
%macro OpcAF 0
	LX
	Inc	X
	Mov	[DPI],A
	CleanUp	4,1,WD
%endmacro


;
;MOV1 - Load Absolute Bit Into Carry
; N V P B H I Z C
;               *
;Mov1 C,mem.bit
%macro OpcAA 0
	Lmbit
	Push	EBX
	Mov	BL,[ABSL]
	BT 	EBX,EDX
	GetC
	Pop	EBX
	CleanUp	4,3,RA
%endmacro


;
;MOV1 - Store Carry in Absolute Bit
; N V P B H I Z C

;Mov1 mem.bit,C
%macro OpcCA 0
	Lmbit
	Mov	CL,DL
	Mov	DL,[PSW+CF]
	ShL	DL,CL
	Or		[ABSL],DL
	Mov	DL,0FEh
	Or		DL,[PSW+CF]
	ROL	DL,CL
	And	[ABSL],DL
	CleanUp	6,3,WA
%endmacro


;
;MOVW - Move Word
; N V P B H I Z C
; *           *
;MovW YA,dp
%macro OpcBA 0
	Ldp
	Mov	YA,[DPI]
	Test	YA,YA
	CleanUp	5,2,RD16,NZ
%endmacro

;MovW dp,YA (flags not affected)
%macro OpcDA 0
	Ldp
	Mov	[DPI],YA
	CleanUp	4,2,WD16
%endmacro


;
;MUL - Multiply
; N V P B H I Z C
; *           *
;Mul YA
%macro OpcCF 0
	Mul	Y
	Test	YA,YA
	CleanUp	9,1,na,NZ
%endmacro


;
;NOT1 - Complement Absolute Bit
; N V P B H I Z C

;Not1 mem.bit
%macro OpcEA 0
	Lmbit
	Mov	CL,DL
	Mov	DH,1
	ShL	DH,CL
	XOr	[ABSL],DH
	CleanUp	5,3,WA
%endmacro


;
;NOTC - Complement Carry Flag
; N V P B H I Z C
;               *
;NotC
%macro OpcED 0
	XOr	byte [PSW+CF],1
	CleanUp	3,1
%endmacro


;
;OR - Bit-wise Logical Inclusive Or
; N V P B H I Z C
; *           *
;Or A,#imm
%macro Opc08 0
	Or 	A,[OP1]
	CleanUp	2,2,na,NZ
%endmacro

;Or A,dp
%macro Opc04 0
	Ldp
	CheckDSP
	Or 	A,[DPI]
	CleanUp	3,2,RD,NZ
%endmacro

;Or A,dp+X
%macro Opc14 0
	LdpX
	CheckDSP
	Or 	A,[DPI]
	CleanUp	4,2,RD,NZ
%endmacro

;Or A,abs
%macro Opc05 0
	Labs
	CheckDSP
	Or 	A,[ABSL]
	CleanUp	4,3,RA,NZ
%endmacro

;Or A,abs+X
%macro Opc15 0
	LabsX
	CheckDSP
	Or 	A,[ABSL]
	CleanUp	5,3,RA,NZ
%endmacro

;Or A,abs+Y
%macro Opc16 0
	LabsY
	CheckDSP
	Or 	A,[ABSL]
	CleanUp	5,3,RA,NZ
%endmacro

;Or A,(X)
%macro Opc06 0
	LX
	CheckDSP
	Or 	A,[DPI]
	CleanUp	3,1,RD,NZ
%endmacro

;Or A,[dp+X]
%macro Opc07 0
	LadpX
	CheckDSP
	Or 	A,[ABSL]
	CleanUp	6,2,RA,NZ
%endmacro

;Or A,[dp]+Y
%macro Opc17 0
	LadpY
	CheckDSP
	Or 	A,[ABSL]
	CleanUp	6,2,RA,NZ
%endmacro

;Or dp,#imm
%macro Opc18 0
	Ldp2
	Mov	DH,[OP1]
	Or 	[DPI],DH
	CleanUp	5,3,WD,NZ
%endmacro

;Or dp,dp
%macro Opc09 0
	Ldp
	CheckDSP
	Mov	DH,[DPI]
	RdPost 0
	Ldp2
	Or 	[DPI],DH
	CleanUp	6,3,WD,NZ
%endmacro

;Or (X),(Y)
%macro Opc19 0
	LY
	CheckDSP
	Mov	DH,[DPI]
	RdPost 0
	LX
	Or 	[DPI],DH
	CleanUp	5,1,WD,NZ
%endmacro


;
;OR1 - Or Carry with Absolute Bit
; N V P B H I Z C
;               *
;Or1 C,mem.bit
%macro Opc0A 0
	Lmbit
	Mov	CL,[ABSL]
	BT 	ECX,EDX
	SetC	DL
	Or		[PSW+CF],DL
	CleanUp	5,3,RA
%endmacro

;Or1 C,/mem.bit
%macro Opc2A 0
	Lmbit
	Mov	CL,[ABSL]
	BT 	ECX,EDX
	SetNC	DL
	Or		[PSW+CF],DL
	CleanUp	5,3,RA
%endmacro


;
;PCALL - Call Procedure in Uppermost Page
; N V P B H I Z C

;PCall up
%macro Opc4F 0
	LEA		EBX,[RAM+up]
	Mov		BL,[OP1]
	Inc		PC
	PushW		PC
	Mov		ESI,EBX
	CleanUp	6,na
%endmacro


;
;POP - Pop Register Off Stack
; N V P B H I Z C

;Pop A
%macro OpcAE 0
	PopB	A
	CleanUp	4,1
%endmacro

;Pop X
%macro OpcCE 0
	PopB	X
	CleanUp	4,1
%endmacro

;Pop Y
%macro OpcEE 0
	PopB	Y
	CleanUp	4,1
%endmacro


;
;POP - Pop Flags Off Stack
; N V P B H I Z C
; ? ? ? ? ? ? ? ?
;Pop PSW
%macro Opc8E 0
	PopB	PS
	ExpPSW
	CleanUp	4,1
%endmacro


;
;PUSH - Push Register Onto Stack
; N V P B H I Z C

;Push A
%macro Opc2D 0
	PushB	A
	CleanUp	4,1
%endmacro

;Push X
%macro Opc4D 0
	PushB	X
	CleanUp	4,1
%endmacro

;Push Y
%macro Opc6D 0
	PushB	Y
	CleanUp	4,1
%endmacro

;Push PSW
%macro Opc0D 0
	CmpPSW
	PushB	PS
	CleanUp	4,1
%endmacro


;
;RET - Return From Call
; N V P B H I Z C

;Ret
%macro Opc6F 0
	PopW	PC
	CleanUp	5,1
%endmacro


;
;RETI - Return From Interrupt
; N V P B H I Z C
; ? ? ? ? ? 1 ? ?

;RetI
%macro Opc7F 0
	PopB	PS
	ExpPSW
	PopW	PC
	CleanUp	6,1
%endmacro


;
;ROL - Rotate 9-bits Left
; N V P B H I Z C
; *           * *
;RoL A
%macro Opc3C 0
	ShR	byte [PSW+CF],1
	RCL	A,1
	GetC
	Test	A,A
	CleanUp	2,1,na,NZ
%endmacro

;RoL dp
%macro Opc2B 0
	Ldp
	ShR	byte [PSW+CF],1
	RCL	byte [DPI],1
	GetC
	Cmp	byte [DPI],0
	CleanUp	4,2,WD,NZ
%endmacro

;RoL dp+X
%macro Opc3B 0
	LdpX
	ShR	byte [PSW+CF],1
	RCL	byte [DPI],1
	GetC
	Cmp	byte [DPI],0
	CleanUp	5,2,WD,NZ
%endmacro

;RoL abs
%macro Opc2C 0
	Labs
	ShR	byte [PSW+CF],1
	RCL	byte [ABSL],1
	GetC
	Cmp	byte [ABSL],0
	CleanUp	5,3,WA,NZ
%endmacro


;
;ROR - Rotate 9-bits Right
; N V P B H I Z C
; *           * *
;RoR A
%macro Opc7C 0
	ShR	byte [PSW+CF],1
	RCR	A,1
	GetC
	Test	A,A
	CleanUp	2,1,na,NZ
%endmacro

;RoR dp
%macro Opc6B 0
	Ldp
	ShR	byte [PSW+CF],1
	RCR	byte [DPI],1
	GetC
	Cmp	byte [DPI],0
	CleanUp	4,2,WD,NZ
%endmacro

;RoR dp+X
%macro Opc7B 0
	LdpX
	ShR	byte [PSW+CF],1
	RCR	byte [DPI],1
	GetC
	Cmp	byte [DPI],0
	CleanUp	5,2,WD,NZ
%endmacro

;RoR abs
%macro Opc6C 0
	Labs
	ShR	byte [PSW+CF],1
	RCR	byte [ABSL],1
	GetC
	Cmp	byte [ABSL],0
	CleanUp	5,3,WA,NZ
%endmacro


;
;SBC - Subtract with Carry
; N V P B H I Z C
; * *     *   * *
;SbC A,#imm
%macro OpcA8 0
	Cmp	byte [PSW+CF],1
	SbB	A,[OP1]
	CleanUp	2,2,na,NVHZCs
%endmacro

;SbC A,dp
%macro OpcA4 0
	Ldp
	CheckDSP
	Cmp	byte [PSW+CF],1
	SbB	A,[DPI]
	CleanUp	3,2,RD,NVHZCs
%endmacro

;SbC A,dp+X
%macro OpcB4 0
	LdpX
	CheckDSP
	Cmp	byte [PSW+CF],1
	SbB	A,[DPI]
	CleanUp	4,2,RD,NVHZCs
%endmacro

;SbC A,abs
%macro OpcA5 0
	Labs
	CheckDSP
	Cmp	byte [PSW+CF],1
	SbB	A,[ABSL]
	CleanUp	4,3,RA,NVHZCs
%endmacro

;SbC A,abs+X
%macro OpcB5 0
	LabsX
	CheckDSP
	Cmp	byte [PSW+CF],1
	SbB	A,[ABSL]
	CleanUp	5,3,RA,NVHZCs
%endmacro

;SbC A,abs+Y
%macro OpcB6 0
	LabsY
	CheckDSP
	Cmp	byte [PSW+CF],1
	SbB	A,[ABSL]
	CleanUp	5,3,RA,NVHZCs
%endmacro

;SbC A,(X)
%macro OpcA6 0
	LX
	CheckDSP
	Cmp	byte [PSW+CF],1
	SbB	A,[DPI]
	CleanUp	3,1,RD,NVHZCs
%endmacro

;SbC A,[dp+X]
%macro OpcA7 0
	LadpX
	CheckDSP
	Cmp	byte [PSW+CF],1
	SbB	A,[ABSL]
	CleanUp	6,2,RA,NVHZCs
%endmacro

;SbC A,[dp]+Y
%macro OpcB7 0
	LadpY
	CheckDSP
	Cmp	byte [PSW+CF],1
	SbB	A,[ABSL]
	CleanUp	6,2,RA,NVHZCs
%endmacro

;SbC dp,#imm
%macro OpcB8 0
	Ldp2
	Mov	DH,[OP1]
	Cmp	byte [PSW+CF],1
	SbB	[DPI],DH
	CleanUp	5,3,WD,NVHZCs
%endmacro

;SbC dp,dp
%macro OpcA9 0
	Ldp
	CheckDSP
	Mov	DH,[DPI]
	RdPost 0
	Ldp2
	Cmp	byte [PSW+CF],1
	SbB	[DPI],DH
	CleanUp	6,3,WD,NVHZCs
%endmacro

;SbC (X),(Y)
%macro OpcB9 0
	LY
	CheckDSP
	Mov	DH,[DPI]
	RdPost 0
	LX
	Cmp	byte [PSW+CF],1
	SbB	[DPI],DH
	CleanUp	5,1,WD,NVHZCs
%endmacro


;
;SET1 - Set Bit
; N V P B H I Z C

	;-----------------------------------------
	;Set bit
	;   %1 - Bit to Set (0-7)

	%macro Set 1
		Ldp
		Or 	byte [DPI],1 << %1
		CleanUp	4,2,WD
	%endmacro

;Set1 dp.0
%macro Opc02 0
	Set	0
%endmacro

;Set1 dp.1
%macro Opc22 0
	Set	1
%endmacro

;Set1 dp.2
%macro Opc42 0
	Set	2
%endmacro

;Set1 dp.3
%macro Opc62 0
	Set	3
%endmacro

;Set1 dp.4
%macro Opc82 0
	Set	4
%endmacro

;Set1 dp.5
%macro OpcA2 0
	Set	5
%endmacro

;Set1 dp.6
%macro OpcC2 0
	Set	6
%endmacro

;Set1 dp.7
%macro OpcE2 0
	Set	7
%endmacro


;
;SETC - Set Carry Flag
; N V P B H I Z C
;               1
;SetC
%macro Opc80 0
	Mov	byte [PSW+CF],1
	CleanUp	2,1
%endmacro


;
;SETP - Set Direct Page Flag
; N V P B H I Z C
;     1     0
;SetP
%macro Opc40 0
	Mov	byte [PSW+P],1
	Mov	byte [PSW+I],0
	CleanUp	2,1
%endmacro


;
;SLEEP - Suspend the SPC700
; N V P B H I Z C

;Sleep
%macro OpcEF 0
	Test	byte [portMod],80h						;Is the SPC700 sleeping?
	SetZ	DL												;  No, Erase port status
	Dec	DL
	And	[portMod],DL

	Test	byte [portMod],0Fh						;Has the 65816 written to the SPC700?
	JNZ	%%WakeUp										;  Yes, Wake up
		Or		byte [portMod],80h					;SPC700 is in sleep mode
		Dec	PC											;Sleep by repeating the instruction
%if SPEED
		Call	SleepHack
%endif
		CleanUp	3,1
ALIGN 16
	%%WakeUp:
	Mov	byte [portMod],0
	CleanUp	3,1
%endmacro


;
;STOP - Stop the SPC700
; N V P B H I Z C

;Stop
%macro OpcFF 0
	Dec	PC												;Stop execution by repeating the instruction
%if DEBUG
	Test	byte [dbgOpt],SPC_TRACE
	JNZ	short %%NoDbg
%endif
	Mov	EBX,[pDebug]
	Test	EBX,EBX
	JZ		short %%NoDbg
		Sub	dword [clkLeft],3*CPU_CYC
		Jmp	SPCBreak
	%%NoDbg:

	CleanUp	3,1
%endmacro


;
;SUBW - Subtract Word
; N V P B H I Z C
; * *     *   * *
;SubW YA,dp
%macro Opc9A 0
	Ldp
	Sub	YA,[DPI]
	CleanUp	5,2,RD16,NVHZCs
%endmacro


;
;TCALL - Call Pointer in Vector Table
; N V P B H I Z C

	;-----------------------------------------
	;Table Call
	;   %1 - Table Index

	%macro TCall 1
		PushW		PC
		Mov		PC,word [((15-%1)*2)+RAM+ipl]
		CleanUp	8,na
	%endmacro

;TCall 0
%macro Opc01 0
	TCall	0
%endmacro

;TCall 1
%macro Opc11 0
	TCall	1
%endmacro

;TCall 2
%macro Opc21 0
	TCall	2
%endmacro

;TCall 3
%macro Opc31 0
	TCall	3
%endmacro

;TCall 4
%macro Opc41 0
	TCall	4
%endmacro

;TCall 5
%macro Opc51 0
	TCall	5
%endmacro

;TCall 6
%macro Opc61 0
	TCall	6
%endmacro

;TCall 7
%macro Opc71 0
	TCall	7
%endmacro

;TCall 8
%macro Opc81 0
	TCall	8
%endmacro

;TCall 9
%macro Opc91 0
	TCall	9
%endmacro

;TCall 10
%macro OpcA1 0
	TCall	10
%endmacro

;TCall 11
%macro OpcB1 0
	TCall	11
%endmacro

;TCall 12
%macro OpcC1 0
	TCall	12
%endmacro

;TCall 13
%macro OpcD1 0
	TCall	13
%endmacro

;TCall 14
%macro OpcE1 0
	TCall	14
%endmacro

;TCall 15
%macro OpcF1 0
	TCall	15
%endmacro


;
;TCLR1 - Test and Clear Bits with A
; N V P B H I Z C
; *           *
;TClr1 abs
%macro Opc4E 0
	Labs
	Mov	DH,[ABSL]
	Not	A
	And	[ABSL],A
	Not	A
	Test	DH,A
	CleanUp	6,3,WA,NZ
%endmacro


;
;TSET1 - Test and Set Bits with A
; N V P B H I Z C
; *           *
;TSet1 abs
%macro Opc0E 0
	Labs
	Mov	DH,[ABSL]
	Or 	[ABSL],A
	Test	DH,A
	CleanUp	6,3,WA,NZ
%endmacro


;
;XCN - Exchange Nybbles
; N V P B H I Z C
; *           *
;XCN A
%macro Opc9F 0
	ROR	A,4
	Test	A,A
	CleanUp	5,1,na,NZ
%endmacro


;
;Function Register Handlers
;
;These are specific functions to handle data written to the function registers (F0h to FFh)

;
;Test (unused)

%macro Func0 0
%if DEBUG
	Jmp	EBP
%else
	Jmp	SPCFetch
%endif

;This is a bit of code to handle 16-bit writes to the F0h - FEh range.
;During a 16-bit write, EBP is loaded with the address here, and the address of the handler for the
;high byte is pushed onto the stack.  After handling the low byte, the emulator will jump here where
;EBP will be restored then control will be passed to the handler for the high byte.

ALIGN 16
%if DEBUG
	Mov	EBP,[pOpFetch]
%else
	Mov	EBP,SPCFetch
%endif
	Ret

%endmacro


;
;Control
;   Checks control register for ROM reading enable, port reset, and timer enable.

%macro Func1 0
	Push	EAX
	Mov	BL,[RAM+control]

	;ROM access ------------------------------
	Mov	AL,[tControl]
	XOr	AL,BL
	Test	AL,80h										;Did ROM access change?
	JZ 	short %%NoRA								;  No

		Push	ESI,ECX

		LEA	ESI,[RAM+ipl]							;Setup registers to move data from Extra RAM to
		Mov	EDI,extraRAM							; IPL region
		Mov	ECX,10h

		Test	BL,80h									;Is ROM readable?
		JZ		short %%NoRR							;  No, Perform move as intended
			Rep	MovSD									;Reverse original operation, move IPL to Extra RAM

			Mov	ESI,[pAPURAM]						;Setup registers to move ROM program into IPL region
			Add	ESI,ipl
			Mov	EDI,iplROM
			Mov	CL,10h
		%%NoRR:

		%%MovROM:
			Mov	EAX,[EDI]							;Move iplROM or extraRAM to IPL ROM region
			Mov	[ESI],EAX
			LEA	EDI,[4+EDI]
			LEA	ESI,[4+ESI]
		Dec	CL
		JNZ	short %%MovROM

		Mov	EDI,[pAPURAM]
		Pop	ECX,ESI

	%%NoRA:

	;Clear ports -----------------------------
	Test	BL,30h										;Was a clear ports command written?
	JZ 	short %%NoCP								;  No
		Mov	EAX,EBX
		Not	EAX										;Reverse command bits
		ShL	EAX,26									;Create a mask based on bits 5 & 4
		SAR	EAX,15
		SAR	AX,15
		And	dword [RAM+port0],EAX				;Reset in-ports
		And	dword [inPortCp],EAX
;		And	BL,Not 30h
;		Mov	[RAM.control],BL						;Reset clear ports bits
	%%NoCP:

	;Timer control ---------------------------
	Mov	AL,[tControl]								;AL = Timers currently enabled
	And	BL,87h
	Mov	[tControl],BL								;Store new timers enabled
	Mov	[RAM+control],BL

	Not	AL
	And	BL,AL											;BL=1 if new timer=1 and old timer=0
	And	BL,7
	JZ 	short %%NoTR								;  Quit if no timers are reset
		;Branching method ---------------------
		ShR	BL,1										;Do we need to reset timer 0?
		JNC 	short %%NoRT0							;  No
			Mov	AL,[RAM+t0]							;Get timer register
			Dec	AL
			Mov	[t0Step],AL							;Reset timer step
			Mov	byte [RAM+c0],0					;Reset counter
		%%NoRT0:

		ShR	BL,1
		JNC	short %%NoRT1
			Mov	AL,[RAM+t1]
			Dec	AL
			Mov	[t1Step],AL
			Mov	byte [RAM+c1],0
		%%NoRT1:

		ShR	BL,1
		JNC	short %%NoRT2
			Mov	AL,[RAM+t2]
			Dec	AL
			Mov	[t2Step],AL
			Mov	byte [RAM+c2],0
		%%NoRT2:

		;Branch free method -------------------
;		ShL	EBX,29									;EBX = abc00000000000000000000000000000
;		SAR	EBX,7										;EBX = aaaaaaaabc0000000000000000000000
;		ShR	EBX,8										;EBX = 00000000aaaaaaaabc00000000000000
;		SAR	BX,7										;EBX = 00000000aaaaaaaabbbbbbbbc0000000
;		SAR	BL,7										;EBX = 00000000aaaaaaaabbbbbbbbcccccccc
;		Not	EBX										;EBX = 11111111AAAAAAAABBBBBBBBCCCCCCCC
;
;		And	[t0Step],EBX							;Reset internal up counters
;		And	[RAM+c0],EBX							;Reset counters
;		Not	EBX
;		And	EBX,[RAM+t0]							;Get new timer values
;		Or 	[t0Step],EBX							;Update internal up counters
	%%NoTR:

	Pop	EAX

	Jmp	EBP

%endmacro


;
;DSP Address
;   Loads 0F3h with value indexed in DSP RAM.  In the SNES, reads from registers 80-FF mirror
;   registers 00-7F.

%macro Func2 0
	;Return any data -------------------------
	MovZX	EBX,byte [RAM+dspAddr]					;EBX = DSP register
	And	EBX,7Fh										;The MSB of the addr is ignored when getting data

	Mov	BL,[EBX+dsp]								;Get byte from DSP RAM
	Mov	[RAM+dspData],BL							;Store byte in DSP data reg
	Jmp	EBP
%endmacro


;
;DSP Data
;   Updates the DSP RAM with the value written to the data port, and calls the DSP emulator to
;   handle the new data.

%macro Func3 0
	Test	byte [dbgOpt],SPC_NODSP
	JNZ	short @@NoDSP
		Push	EDX,EAX
		MovZX	EBX,byte [RAM+dspAddr]
		Mov	AL,[RAM+dspData]
		Call	DSPIn

;		Mov	EBP,SPCExit								;DEBUG: Stop emulation if F3h is written to
;		Mov	EAX,[clkLeft]							;       MOVW 0F3h,YA will crash this
;		Add	[clkExec],EAX
;		Sub	[clkLeft],EAX

		Pop	EAX,EDX
	Jmp	EBP
ALIGN 16
	@@NoDSP:
	MovZX	EBX,byte [RAM+dspAddr]
	Mov	CL,[RAM+dspData]
	Mov	[EBX+dsp],CL
	Jmp	EBP
%endmacro


;
;Out Ports 0-3
;   Moves data to out port memory, and replaces register with in port value

%macro Func4 0
	Mov	CL,[RAM+port0]								;Get value written
	Mov	BL,[0+inPortCp]							;Get value to be read from in port memory (from console)
	Mov	[0+outPort],CL								;Place value written into out port memory (to console)
	Mov	[RAM+port0],BL								;Reload register with in port value

	Jmp	EBP
%endmacro

%macro Func5 0
	Mov	CL,[RAM+port1]
	Mov	BL,[1+inPortCp]
	Mov	[1+outPort],CL
	Mov	[RAM+port1],BL

	Jmp	EBP
%endmacro

%macro Func6 0
	Mov	CL,[RAM+port2]
	Mov	BL,[2+inPortCp]
	Mov	[2+outPort],CL
	Mov	[RAM+port2],BL

	Jmp	EBP
%endmacro

%macro Func7 0
	Mov	CL,[RAM+port3]
	Mov	BL,[3+inPortCp]
	Mov	[3+outPort],CL
	Mov	[RAM+port3],BL

	Jmp	EBP
%endmacro


;
;Unused
;   No special handling is associated with registers 0F8h and 0F9h

%macro Func8 0
	Jmp	EBP
%endmacro

%macro Func9 0
	Jmp	EBP
%endmacro


;
;Timers 0-2
;   In reality, reading the timer registers always returns 00

%macro FuncA 0
	Jmp	EBP
%endmacro

%macro FuncB 0
	Jmp	EBP
%endmacro

%macro FuncC 0
	Jmp	EBP
%endmacro


;
;Counters 0-2
;  The counter registers can't be written to.  In the unlikely event that they are, this hack will
;  reset them to keep things from going awry.

%macro FuncD 0
	Mov	byte [RAM+c0],0
	Jmp	EBP
%endmacro

%macro FuncE 0
	Mov	byte [RAM+c1],0
	Jmp	EBP
%endmacro

%macro FuncF 0
	Mov	byte [RAM+c2],0
	Jmp	EBP
%endmacro


;
;Build Opcode Table
;
;Arrange all the opcoder emulators in order.  (This was a lot easier with MASM's FORC macro.)

TIMES (($$-$) & ((256)-1)) NOp
OpcBase:

;%define test 1
Opc00
TIMES (($$-$) & ((256)-1)) NOp
Opc01
TIMES (($$-$) & ((256)-1)) NOp
Opc02
TIMES (($$-$) & ((256)-1)) NOp
Opc03
TIMES (($$-$) & ((256)-1)) NOp
Opc04
TIMES (($$-$) & ((256)-1)) NOp
Opc05
TIMES (($$-$) & ((256)-1)) NOp
Opc06
TIMES (($$-$) & ((256)-1)) NOp
Opc07
TIMES (($$-$) & ((256)-1)) NOp
Opc08
TIMES (($$-$) & ((256)-1)) NOp
Opc09
TIMES (($$-$) & ((256)-1)) NOp
Opc0A
TIMES (($$-$) & ((256)-1)) NOp
Opc0B
TIMES (($$-$) & ((256)-1)) NOp
Opc0C
TIMES (($$-$) & ((256)-1)) NOp
Opc0D
TIMES (($$-$) & ((256)-1)) NOp
Opc0E
TIMES (($$-$) & ((256)-1)) NOp
Opc0F
TIMES (($$-$) & ((256)-1)) NOp
Opc10
TIMES (($$-$) & ((256)-1)) NOp
Opc11
TIMES (($$-$) & ((256)-1)) NOp
Opc12
TIMES (($$-$) & ((256)-1)) NOp
Opc13
TIMES (($$-$) & ((256)-1)) NOp
Opc14
TIMES (($$-$) & ((256)-1)) NOp
Opc15
TIMES (($$-$) & ((256)-1)) NOp
Opc16
TIMES (($$-$) & ((256)-1)) NOp
Opc17
TIMES (($$-$) & ((256)-1)) NOp
Opc18
TIMES (($$-$) & ((256)-1)) NOp
Opc19
TIMES (($$-$) & ((256)-1)) NOp
Opc1A
TIMES (($$-$) & ((256)-1)) NOp
Opc1B
TIMES (($$-$) & ((256)-1)) NOp
Opc1C
TIMES (($$-$) & ((256)-1)) NOp
Opc1D
TIMES (($$-$) & ((256)-1)) NOp
Opc1E
TIMES (($$-$) & ((256)-1)) NOp
Opc1F
TIMES (($$-$) & ((256)-1)) NOp
Opc20
TIMES (($$-$) & ((256)-1)) NOp
Opc21
TIMES (($$-$) & ((256)-1)) NOp
Opc22
TIMES (($$-$) & ((256)-1)) NOp
Opc23
TIMES (($$-$) & ((256)-1)) NOp
Opc24
TIMES (($$-$) & ((256)-1)) NOp
Opc25
TIMES (($$-$) & ((256)-1)) NOp
Opc26
TIMES (($$-$) & ((256)-1)) NOp
Opc27
TIMES (($$-$) & ((256)-1)) NOp
Opc28
TIMES (($$-$) & ((256)-1)) NOp
Opc29
TIMES (($$-$) & ((256)-1)) NOp
Opc2A
TIMES (($$-$) & ((256)-1)) NOp
Opc2B
TIMES (($$-$) & ((256)-1)) NOp
Opc2C
TIMES (($$-$) & ((256)-1)) NOp
Opc2D
TIMES (($$-$) & ((256)-1)) NOp
Opc2E
TIMES (($$-$) & ((256)-1)) NOp
Opc2F
TIMES (($$-$) & ((256)-1)) NOp
Opc30
TIMES (($$-$) & ((256)-1)) NOp
Opc31
TIMES (($$-$) & ((256)-1)) NOp
Opc32
TIMES (($$-$) & ((256)-1)) NOp
Opc33
TIMES (($$-$) & ((256)-1)) NOp
Opc34
TIMES (($$-$) & ((256)-1)) NOp
Opc35
TIMES (($$-$) & ((256)-1)) NOp
Opc36
TIMES (($$-$) & ((256)-1)) NOp
Opc37
TIMES (($$-$) & ((256)-1)) NOp
Opc38
TIMES (($$-$) & ((256)-1)) NOp
Opc39
TIMES (($$-$) & ((256)-1)) NOp
Opc3A
TIMES (($$-$) & ((256)-1)) NOp
Opc3B
TIMES (($$-$) & ((256)-1)) NOp
Opc3C
TIMES (($$-$) & ((256)-1)) NOp
Opc3D
TIMES (($$-$) & ((256)-1)) NOp
Opc3E
TIMES (($$-$) & ((256)-1)) NOp
Opc3F
TIMES (($$-$) & ((256)-1)) NOp
Opc40
TIMES (($$-$) & ((256)-1)) NOp
Opc41
TIMES (($$-$) & ((256)-1)) NOp
Opc42
TIMES (($$-$) & ((256)-1)) NOp
Opc43
TIMES (($$-$) & ((256)-1)) NOp
Opc44
TIMES (($$-$) & ((256)-1)) NOp
Opc45
TIMES (($$-$) & ((256)-1)) NOp
Opc46
TIMES (($$-$) & ((256)-1)) NOp
Opc47
TIMES (($$-$) & ((256)-1)) NOp
Opc48
TIMES (($$-$) & ((256)-1)) NOp
Opc49
TIMES (($$-$) & ((256)-1)) NOp
Opc4A
TIMES (($$-$) & ((256)-1)) NOp
Opc4B
TIMES (($$-$) & ((256)-1)) NOp
Opc4C
TIMES (($$-$) & ((256)-1)) NOp
Opc4D
TIMES (($$-$) & ((256)-1)) NOp
Opc4E
TIMES (($$-$) & ((256)-1)) NOp
Opc4F
TIMES (($$-$) & ((256)-1)) NOp
Opc50
TIMES (($$-$) & ((256)-1)) NOp
Opc51
TIMES (($$-$) & ((256)-1)) NOp
Opc52
TIMES (($$-$) & ((256)-1)) NOp
Opc53
TIMES (($$-$) & ((256)-1)) NOp
Opc54
TIMES (($$-$) & ((256)-1)) NOp
Opc55
TIMES (($$-$) & ((256)-1)) NOp
Opc56
TIMES (($$-$) & ((256)-1)) NOp
Opc57
TIMES (($$-$) & ((256)-1)) NOp
Opc58
TIMES (($$-$) & ((256)-1)) NOp
Opc59
TIMES (($$-$) & ((256)-1)) NOp
Opc5A
TIMES (($$-$) & ((256)-1)) NOp
Opc5B
TIMES (($$-$) & ((256)-1)) NOp
Opc5C
TIMES (($$-$) & ((256)-1)) NOp
Opc5D
TIMES (($$-$) & ((256)-1)) NOp
Opc5E
TIMES (($$-$) & ((256)-1)) NOp
Opc5F
TIMES (($$-$) & ((256)-1)) NOp
Opc60
TIMES (($$-$) & ((256)-1)) NOp
Opc61
TIMES (($$-$) & ((256)-1)) NOp
Opc62
TIMES (($$-$) & ((256)-1)) NOp
Opc63
TIMES (($$-$) & ((256)-1)) NOp
Opc64
TIMES (($$-$) & ((256)-1)) NOp
Opc65
TIMES (($$-$) & ((256)-1)) NOp
Opc66
TIMES (($$-$) & ((256)-1)) NOp
Opc67
TIMES (($$-$) & ((256)-1)) NOp
Opc68
TIMES (($$-$) & ((256)-1)) NOp
Opc69
TIMES (($$-$) & ((256)-1)) NOp
Opc6A
TIMES (($$-$) & ((256)-1)) NOp
Opc6B
TIMES (($$-$) & ((256)-1)) NOp
Opc6C
TIMES (($$-$) & ((256)-1)) NOp
Opc6D
TIMES (($$-$) & ((256)-1)) NOp
Opc6E
TIMES (($$-$) & ((256)-1)) NOp
Opc6F
TIMES (($$-$) & ((256)-1)) NOp
Opc70
TIMES (($$-$) & ((256)-1)) NOp
Opc71
TIMES (($$-$) & ((256)-1)) NOp
Opc72
TIMES (($$-$) & ((256)-1)) NOp
Opc73
TIMES (($$-$) & ((256)-1)) NOp
Opc74
TIMES (($$-$) & ((256)-1)) NOp
Opc75
TIMES (($$-$) & ((256)-1)) NOp
Opc76
TIMES (($$-$) & ((256)-1)) NOp
Opc77
TIMES (($$-$) & ((256)-1)) NOp
Opc78
TIMES (($$-$) & ((256)-1)) NOp
Opc79
TIMES (($$-$) & ((256)-1)) NOp
Opc7A
TIMES (($$-$) & ((256)-1)) NOp
Opc7B
TIMES (($$-$) & ((256)-1)) NOp
Opc7C
TIMES (($$-$) & ((256)-1)) NOp
Opc7D
TIMES (($$-$) & ((256)-1)) NOp
Opc7E
TIMES (($$-$) & ((256)-1)) NOp
Opc7F
TIMES (($$-$) & ((256)-1)) NOp
Opc80
TIMES (($$-$) & ((256)-1)) NOp
Opc81
TIMES (($$-$) & ((256)-1)) NOp
Opc82
TIMES (($$-$) & ((256)-1)) NOp
Opc83
TIMES (($$-$) & ((256)-1)) NOp
Opc84
TIMES (($$-$) & ((256)-1)) NOp
Opc85
TIMES (($$-$) & ((256)-1)) NOp
Opc86
TIMES (($$-$) & ((256)-1)) NOp
Opc87
TIMES (($$-$) & ((256)-1)) NOp
Opc88
TIMES (($$-$) & ((256)-1)) NOp
Opc89
TIMES (($$-$) & ((256)-1)) NOp
Opc8A
TIMES (($$-$) & ((256)-1)) NOp
Opc8B
TIMES (($$-$) & ((256)-1)) NOp
Opc8C
TIMES (($$-$) & ((256)-1)) NOp
Opc8D
TIMES (($$-$) & ((256)-1)) NOp
Opc8E
TIMES (($$-$) & ((256)-1)) NOp
Opc8F
TIMES (($$-$) & ((256)-1)) NOp
Opc90
TIMES (($$-$) & ((256)-1)) NOp
Opc91
TIMES (($$-$) & ((256)-1)) NOp
Opc92
TIMES (($$-$) & ((256)-1)) NOp
Opc93
TIMES (($$-$) & ((256)-1)) NOp
Opc94
TIMES (($$-$) & ((256)-1)) NOp
Opc95
TIMES (($$-$) & ((256)-1)) NOp
Opc96
TIMES (($$-$) & ((256)-1)) NOp
Opc97
TIMES (($$-$) & ((256)-1)) NOp
Opc98
TIMES (($$-$) & ((256)-1)) NOp
Opc99
TIMES (($$-$) & ((256)-1)) NOp
Opc9A
TIMES (($$-$) & ((256)-1)) NOp
Opc9B
TIMES (($$-$) & ((256)-1)) NOp
Opc9C
TIMES (($$-$) & ((256)-1)) NOp
Opc9D
TIMES (($$-$) & ((256)-1)) NOp
Opc9E
TIMES (($$-$) & ((256)-1)) NOp
Opc9F
TIMES (($$-$) & ((256)-1)) NOp
OpcA0
TIMES (($$-$) & ((256)-1)) NOp
OpcA1
TIMES (($$-$) & ((256)-1)) NOp
OpcA2
TIMES (($$-$) & ((256)-1)) NOp
OpcA3
TIMES (($$-$) & ((256)-1)) NOp
OpcA4
TIMES (($$-$) & ((256)-1)) NOp
OpcA5
TIMES (($$-$) & ((256)-1)) NOp
OpcA6
TIMES (($$-$) & ((256)-1)) NOp
OpcA7
TIMES (($$-$) & ((256)-1)) NOp
OpcA8
TIMES (($$-$) & ((256)-1)) NOp
OpcA9
TIMES (($$-$) & ((256)-1)) NOp
OpcAA
TIMES (($$-$) & ((256)-1)) NOp
OpcAB
TIMES (($$-$) & ((256)-1)) NOp
OpcAC
TIMES (($$-$) & ((256)-1)) NOp
OpcAD
TIMES (($$-$) & ((256)-1)) NOp
OpcAE
TIMES (($$-$) & ((256)-1)) NOp
OpcAF
TIMES (($$-$) & ((256)-1)) NOp
OpcB0
TIMES (($$-$) & ((256)-1)) NOp
OpcB1
TIMES (($$-$) & ((256)-1)) NOp
OpcB2
TIMES (($$-$) & ((256)-1)) NOp
OpcB3
TIMES (($$-$) & ((256)-1)) NOp
OpcB4
TIMES (($$-$) & ((256)-1)) NOp
OpcB5
TIMES (($$-$) & ((256)-1)) NOp
OpcB6
TIMES (($$-$) & ((256)-1)) NOp
OpcB7
TIMES (($$-$) & ((256)-1)) NOp
OpcB8
TIMES (($$-$) & ((256)-1)) NOp
OpcB9
TIMES (($$-$) & ((256)-1)) NOp
OpcBA
TIMES (($$-$) & ((256)-1)) NOp
OpcBB
TIMES (($$-$) & ((256)-1)) NOp
OpcBC
TIMES (($$-$) & ((256)-1)) NOp
OpcBD
TIMES (($$-$) & ((256)-1)) NOp
OpcBE
TIMES (($$-$) & ((256)-1)) NOp
OpcBF
TIMES (($$-$) & ((256)-1)) NOp
OpcC0
TIMES (($$-$) & ((256)-1)) NOp
OpcC1
TIMES (($$-$) & ((256)-1)) NOp
OpcC2
TIMES (($$-$) & ((256)-1)) NOp
OpcC3
TIMES (($$-$) & ((256)-1)) NOp
OpcC4
TIMES (($$-$) & ((256)-1)) NOp
OpcC5
TIMES (($$-$) & ((256)-1)) NOp
OpcC6
TIMES (($$-$) & ((256)-1)) NOp
OpcC7
TIMES (($$-$) & ((256)-1)) NOp
OpcC8
TIMES (($$-$) & ((256)-1)) NOp
OpcC9
TIMES (($$-$) & ((256)-1)) NOp
OpcCA
TIMES (($$-$) & ((256)-1)) NOp
OpcCB
TIMES (($$-$) & ((256)-1)) NOp
OpcCC
TIMES (($$-$) & ((256)-1)) NOp
OpcCD
TIMES (($$-$) & ((256)-1)) NOp
OpcCE
TIMES (($$-$) & ((256)-1)) NOp
OpcCF
TIMES (($$-$) & ((256)-1)) NOp
OpcD0
TIMES (($$-$) & ((256)-1)) NOp
OpcD1
TIMES (($$-$) & ((256)-1)) NOp
OpcD2
TIMES (($$-$) & ((256)-1)) NOp
OpcD3
TIMES (($$-$) & ((256)-1)) NOp
OpcD4
TIMES (($$-$) & ((256)-1)) NOp
OpcD5
TIMES (($$-$) & ((256)-1)) NOp
OpcD6
TIMES (($$-$) & ((256)-1)) NOp
OpcD7
TIMES (($$-$) & ((256)-1)) NOp
OpcD8
TIMES (($$-$) & ((256)-1)) NOp
OpcD9
TIMES (($$-$) & ((256)-1)) NOp
OpcDA
TIMES (($$-$) & ((256)-1)) NOp
OpcDB
TIMES (($$-$) & ((256)-1)) NOp
OpcDC
TIMES (($$-$) & ((256)-1)) NOp
OpcDD
TIMES (($$-$) & ((256)-1)) NOp
OpcDE
TIMES (($$-$) & ((256)-1)) NOp
OpcDF
TIMES (($$-$) & ((256)-1)) NOp
OpcE0
TIMES (($$-$) & ((256)-1)) NOp
OpcE1
TIMES (($$-$) & ((256)-1)) NOp
OpcE2
TIMES (($$-$) & ((256)-1)) NOp
OpcE3
TIMES (($$-$) & ((256)-1)) NOp
OpcE4
TIMES (($$-$) & ((256)-1)) NOp
OpcE5
TIMES (($$-$) & ((256)-1)) NOp
OpcE6
TIMES (($$-$) & ((256)-1)) NOp
OpcE7
TIMES (($$-$) & ((256)-1)) NOp
OpcE8
TIMES (($$-$) & ((256)-1)) NOp
OpcE9
TIMES (($$-$) & ((256)-1)) NOp
OpcEA
TIMES (($$-$) & ((256)-1)) NOp
OpcEB
TIMES (($$-$) & ((256)-1)) NOp
OpcEC
TIMES (($$-$) & ((256)-1)) NOp
OpcED
TIMES (($$-$) & ((256)-1)) NOp
OpcEE
TIMES (($$-$) & ((256)-1)) NOp
OpcEF
TIMES (($$-$) & ((256)-1)) NOp
OpcF0
TIMES (($$-$) & ((256)-1)) NOp
OpcF1
TIMES (($$-$) & ((256)-1)) NOp
OpcF2
TIMES (($$-$) & ((256)-1)) NOp
OpcF3
TIMES (($$-$) & ((256)-1)) NOp
OpcF4
TIMES (($$-$) & ((256)-1)) NOp
OpcF5
TIMES (($$-$) & ((256)-1)) NOp
OpcF6
TIMES (($$-$) & ((256)-1)) NOp
OpcF7
TIMES (($$-$) & ((256)-1)) NOp
OpcF8
TIMES (($$-$) & ((256)-1)) NOp
OpcF9
TIMES (($$-$) & ((256)-1)) NOp
OpcFA
TIMES (($$-$) & ((256)-1)) NOp
OpcFB
TIMES (($$-$) & ((256)-1)) NOp
OpcFC
TIMES (($$-$) & ((256)-1)) NOp
OpcFD
TIMES (($$-$) & ((256)-1)) NOp
OpcFE
TIMES (($$-$) & ((256)-1)) NOp
OpcFF
TIMES (($$-$) & ((256)-1)) NOp

FuncBase:
Func0
TIMES (($$-$) & ((256)-1)) NOp
Func1
TIMES (($$-$) & ((256)-1)) NOp
Func2
TIMES (($$-$) & ((256)-1)) NOp
Func3
TIMES (($$-$) & ((256)-1)) NOp
Func4
TIMES (($$-$) & ((256)-1)) NOp
Func5
TIMES (($$-$) & ((256)-1)) NOp
Func6
TIMES (($$-$) & ((256)-1)) NOp
Func7
TIMES (($$-$) & ((256)-1)) NOp
Func8
TIMES (($$-$) & ((256)-1)) NOp
Func9
TIMES (($$-$) & ((256)-1)) NOp
FuncA
TIMES (($$-$) & ((256)-1)) NOp
FuncB
TIMES (($$-$) & ((256)-1)) NOp
FuncC
TIMES (($$-$) & ((256)-1)) NOp
FuncD
TIMES (($$-$) & ((256)-1)) NOp
FuncE
TIMES (($$-$) & ((256)-1)) NOp
FuncF
TIMES (($$-$) & ((256)-1)) NOp
	Jmp	EBP
