#include <gtk/gtk.h>
#include "file.h"

fileWidgets widgets;
SPCInfo* info;

/**
   Update info->id666 with data from widgets
   then write to info->filename
**/
void writeInfo() {
  char* tmp;
  int i;

  info->id666.isBinary = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widgets.toggle));

  //text entries
#define GET_CHAR(name,len) \
  tmp = gtk_entry_get_text(GTK_ENTRY(widgets.name)); \
  memcpy(info->id666.name,tmp,len);

  GET_CHAR(song,32);
  GET_CHAR(game,32);
  GET_CHAR(dumper,16);
  GET_CHAR(comments,32);
  GET_CHAR(artist,32);

  //entries converted into numbers
#define GET_NUM(name) \
  tmp = gtk_entry_get_text(GTK_ENTRY(widgets.name)); \
  sscanf(tmp,"%d",&(info->id666.name));

  GET_NUM(day);
  GET_NUM(month);
  GET_NUM(year);
  GET_NUM(fade);

  //song length
  tmp = gtk_entry_get_text(GTK_ENTRY(widgets.lenMins));
  sscanf(tmp,"%d",&(info->id666.length));
  info->id666.length *= 60;
  int tmp2;
  tmp = gtk_entry_get_text(GTK_ENTRY(widgets.lenSecs));
  sscanf(tmp,"%d",&tmp2);
  info->id666.length += tmp2;
  info->id666.length *= 64000;

  //muted channels
  info->id666.muted = 0;

#define GET_MUTE(name,val) \
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widgets.name))) { \
    info->id666.muted |= val; \
  }

  GET_MUTE(mute1,1);
  GET_MUTE(mute2,2);
  GET_MUTE(mute3,4);
  GET_MUTE(mute4,8);
  GET_MUTE(mute5,16);
  GET_MUTE(mute6,32);
  GET_MUTE(mute7,64);
  GET_MUTE(mute8,128);

  //move to spcload??
  for(i=0; i < NUM_EMULATORS; i++) {
    tmp = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(widgets.emulator)->entry));
    if(strcmp(emulators[i],tmp)==0) {
      info->id666.emulator = i;
      break;
    }
  }

  WriteID666(info->filename,&(info->id666));
}


void saveFile(GtkWidget *widget) {
  writeInfo();
  gtk_widget_destroy(widget);  
}


/**
   Create a tab.
   Either id666 or id666ext, depending on isExt
**/
void addFileTab(GtkWidget *spc_file, SPCInfo *info, GtkWidget *notebook, gboolean isExt) {
  GtkWidget *table6;
  GtkWidget *songLabel;
  GtkWidget *gameLabel;
  GtkWidget *artistLabel;
  GtkWidget *dumperLabel;
  GtkWidget *dateDumpedLabel;
  GtkWidget *emulatorLabel;
  GtkWidget *hbox22;
  GtkWidget *label88;
  GtkWidget *label89;
  GtkWidget *songLenLabel;
  GtkWidget *hbox23;
  GtkWidget *label91;
  GtkWidget *fadeLabel;
  GtkWidget *commentsLabel;
  GtkWidget *muteLabel;
  GtkWidget *hbox24;
  GtkWidget *tabLabel;
  GList *glist = NULL;
  char tmp[32];
  int i;

  table6 = gtk_table_new (10, 2, FALSE);
  gtk_widget_show (table6);
  gtk_container_add (GTK_CONTAINER (notebook), table6);
  gtk_table_set_row_spacings (GTK_TABLE (table6), 1);
  gtk_table_set_col_spacings (GTK_TABLE (table6), 3);

  //tab label
  tabLabel = gtk_label_new (isExt ? "ID666 Extended" : "ID666");
  gtk_widget_show (tabLabel);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), isExt? 1 : 0), tabLabel);

  //Toggle Button
  widgets.toggle = gtk_check_button_new_with_label(isExt ? "Enable ID666 Extended Tag" : "Binary ID666 Tag");
  gtk_widget_show (widgets.toggle);
  gtk_table_attach (GTK_TABLE (table6), widgets.toggle, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widgets.toggle),
			       isExt ? info->hasID666ext : info->id666.isBinary);

#define DO_LABEL(label,labeltext,pos) \
  label = gtk_label_new(labeltext); \
  gtk_widget_show(label); \
  gtk_table_attach (GTK_TABLE (table6), label, 0, 1, pos, pos+1, \
                    (GtkAttachOptions) (GTK_FILL), 0, 0, 0); \
  gtk_misc_set_alignment (GTK_MISC (label), 1, 0.5); 

#define DO_ENTRY(entry,pos,length,tagname) \
  entry = gtk_entry_new (); \
  gtk_widget_show (entry); \
  gtk_widget_set_usize(entry, 120, -1); \
  gtk_table_attach (GTK_TABLE (table6), entry, 1, 2, pos, pos+1, \
                    (GtkAttachOptions) (GTK_FILL), \
                    0, 0, 0); \
  gtk_entry_set_max_length (GTK_ENTRY (entry), length); \
  if(isExt) { \
    if(info->hasID666ext) { \
      gtk_entry_set_text(GTK_ENTRY(entry),info->id666ext.tagname); \
    } else { \
      gtk_widget_set_sensitive(GTK_WIDGET(entry),FALSE); \
    } \
  } else { \
    gtk_entry_set_text(GTK_ENTRY(entry),info->id666.tagname); \
  }


  //Song
  DO_LABEL(songLabel,"Song: ",1);
  DO_ENTRY(widgets.song,1,31,song);

  //Game
  DO_LABEL(gameLabel,"Game: ",2);
  DO_ENTRY(widgets.game,2,31,game);

  //Artist
  DO_LABEL(artistLabel,"Artist: ",3);
  DO_ENTRY(widgets.artist,3,31,artist);

  //Dumper
  DO_LABEL(dumperLabel,"Dumper: ",4);
  DO_ENTRY(widgets.dumper,4,15,dumper);

  //Date Dumped
  DO_LABEL(dateDumpedLabel,"Date Dumped: ",5);

  hbox22 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox22);
  gtk_table_attach (GTK_TABLE (table6), hbox22, 1, 2, 5, 6,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (GTK_FILL), 0, 0);

#define DO_NUM_ENTRY(box,entry,length,name,extname) \
  entry = gtk_entry_new (); \
  gtk_widget_show (entry); \
  gtk_widget_set_usize(entry, length*6+12, -1); \
  gtk_box_pack_start (GTK_BOX (box), entry, FALSE, TRUE, 0); \
  gtk_entry_set_max_length (GTK_ENTRY (entry), length); \
  if(isExt) { \
    if(info->hasID666ext && info->id666ext.extname) { \
      snprintf(tmp,32,"%." #length "d",info->id666ext.extname); \
      gtk_entry_set_text(GTK_ENTRY(entry),tmp); \
    } else { \
      gtk_widget_set_sensitive(GTK_WIDGET(entry),FALSE); \
    } \
  } else if(info->id666.name) { \
    snprintf(tmp,32,"%." #length "d",info->id666.name); \
    gtk_entry_set_text(GTK_ENTRY(entry),tmp); \
  }

#define DO_BOX_LABEL(box,label,text) \
  label = gtk_label_new (text); \
  gtk_widget_show (label); \
  gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);

  DO_NUM_ENTRY(hbox22,widgets.year,4,year,year);
  DO_BOX_LABEL(hbox22,label88,"/");
  DO_NUM_ENTRY(hbox22,widgets.month,2,month,month);
  DO_BOX_LABEL(hbox22,label89,"/");
  DO_NUM_ENTRY(hbox22,widgets.day,2,day,day);

  //Emulator
  DO_LABEL(emulatorLabel,"Emulator: ",6);

  widgets.emulator = gtk_combo_new();
  gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(widgets.emulator)->entry),FALSE);
  gtk_widget_show (widgets.emulator);
  gtk_widget_set_usize(widgets.emulator, 120, -1);
  gtk_table_attach (GTK_TABLE (table6), widgets.emulator, 1, 2, 6, 7,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (GTK_FILL), 0, 0);
  for(i = 0; i < NUM_EMULATORS; i++) {
    glist = g_list_append(glist, emulators[i]);
  }
  gtk_combo_set_popdown_strings( GTK_COMBO(widgets.emulator), glist) ;
  g_list_free(glist);
  glist = NULL;
  gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(widgets.emulator)->entry),getEmulator(info->id666.emulator));

  //Comments
  DO_LABEL(commentsLabel,"Comments ",7);
  DO_ENTRY(widgets.comments,7,32,comments);

  //Song Length & Fade
  DO_LABEL(songLenLabel,"Song Length: ",8);

  hbox23 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox23);
  gtk_table_attach (GTK_TABLE (table6), hbox23, 1, 2, 8, 9,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (GTK_FILL), 0, 0);

  DO_NUM_ENTRY(hbox23,widgets.lenMins,2,length/(64000*60),loop);
  DO_BOX_LABEL(hbox23,label91,":");
  DO_NUM_ENTRY(hbox23,widgets.lenSecs,2,length/64000%60,loop);
  DO_BOX_LABEL(hbox23,fadeLabel," Fade: ");
  DO_NUM_ENTRY(hbox23,widgets.fade,3,fade/64000,fade/64000);

  //Muted Channels
  DO_LABEL(muteLabel,"Muted Channels: ",9);

  hbox24 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox24);
  gtk_table_attach (GTK_TABLE (table6), hbox24, 1, 2, 9, 10,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (GTK_FILL), 0, 0);

#define DO_MUTE(button,label,val) \
  button = gtk_check_button_new_with_label (label); \
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), \
			       (isExt ? info->id666ext.muted : info->id666.muted) & val); \
  gtk_widget_show (button); \
  gtk_box_pack_start (GTK_BOX (hbox24), button, FALSE, FALSE, 0);

  DO_MUTE(widgets.mute1,"1",1);
  DO_MUTE(widgets.mute2,"2",2);
  DO_MUTE(widgets.mute3,"3",4);
  DO_MUTE(widgets.mute4,"4",8);
  DO_MUTE(widgets.mute5,"5",16);
  DO_MUTE(widgets.mute6,"6",32);
  DO_MUTE(widgets.mute7,"7",64);
  DO_MUTE(widgets.mute8,"8",128);

}


/*****

Create and return file dialog box

*****/
GtkWidget *spc_file = NULL;
GtkWidget* create_spc_file (SPCInfo *inf)
{
  GtkWidget *vbox7;
  GtkWidget *hbox17;
  GtkWidget *filenameEntryLabel;
  GtkWidget *filenameEntry;
  GtkWidget *notebook;
  GtkWidget *hbox9;
  GtkWidget *filler;
  GtkWidget *save;
  GtkWidget *cancel;

  info = inf;

  if (spc_file != NULL) {
    gdk_window_raise(spc_file->window);
    return spc_file;
  }

  spc_file = gtk_window_new (GTK_WINDOW_DIALOG);
  gtk_window_set_title (GTK_WINDOW (spc_file), "ID666 Tag");
  gtk_window_set_default_size (GTK_WINDOW (spc_file), 50, 50);
  gtk_window_set_policy(GTK_WINDOW(spc_file),0,0,0);
  gtk_container_border_width(GTK_CONTAINER(spc_file), 5);

  vbox7 = gtk_vbox_new (FALSE, 5);
  gtk_widget_show (vbox7);
  gtk_container_add (GTK_CONTAINER (spc_file), vbox7);

  hbox17 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox17);
  gtk_box_pack_start (GTK_BOX (vbox7), hbox17, FALSE, TRUE, 0);

  //Filename
  filenameEntryLabel = gtk_label_new ("Filename: ");
  gtk_widget_show (filenameEntryLabel);
  gtk_box_pack_start (GTK_BOX (hbox17), filenameEntryLabel, FALSE, FALSE, 0);

  filenameEntry = gtk_entry_new ();
  gtk_widget_set_usize(filenameEntry, 90, -1);
  gtk_widget_show (filenameEntry);
  gtk_entry_set_text(GTK_ENTRY(filenameEntry), info->filename);
  gtk_editable_set_editable(GTK_EDITABLE(filenameEntry), FALSE);
  gtk_box_pack_start (GTK_BOX (hbox17), filenameEntry, TRUE, TRUE, 0);

  //id666/ext notebook
  notebook = gtk_notebook_new ();
  gtk_widget_show (notebook);
  gtk_box_pack_start (GTK_BOX (vbox7), notebook, FALSE, TRUE, 0);

  //add tabs
  addFileTab(spc_file,info,notebook,FALSE);
  //  addFileTab(spc_file,info,notebook,TRUE);

  //buttons box
  hbox9 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox9);
  gtk_box_pack_start (GTK_BOX (vbox7), hbox9, FALSE, TRUE, 0);

  filler = gtk_label_new ("");
  gtk_widget_show (filler);
  gtk_box_pack_start (GTK_BOX (hbox9), filler, TRUE, TRUE, 0);

  save = gtk_button_new_with_label ("Save");
  gtk_widget_show (save);
  gtk_widget_set_usize (save, 80, -1);
  gtk_box_pack_start (GTK_BOX (hbox9), save, FALSE, FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (save), 5);

  cancel = gtk_button_new_with_label ("Cancel");
  gtk_widget_show (cancel);
  gtk_widget_set_usize (cancel, 80, -1);
  gtk_box_pack_start (GTK_BOX (hbox9), cancel, FALSE, FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (cancel), 5);
  /**
  apply = gtk_button_new_with_label ("Apply");
  gtk_widget_show (apply);
  gtk_widget_set_usize (apply, 80, -1);
  gtk_box_pack_start (GTK_BOX (hbox9), apply, FALSE, FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (apply), 5);
  **/
  gtk_signal_connect_object ((gpointer) cancel, "clicked",
                            GTK_SIGNAL_FUNC(gtk_widget_destroy),
                            GTK_OBJECT(spc_file));

  gtk_signal_connect_object ((gpointer) save, "clicked",
                            GTK_SIGNAL_FUNC(saveFile),
                            GTK_OBJECT(spc_file));

  gtk_signal_connect(GTK_OBJECT(spc_file), "destroy", 
		     GTK_SIGNAL_FUNC(gtk_widget_destroyed), 
		     &spc_file);


  return spc_file;
}

