/* Structures and some other things in this file were taken from
   "spc2wav.c".
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "spcload.h"
#include "sexyconfig.h"

const char emulators[7][12] = {"unknown","ZSNES","Snes9x","ZST2SPC","ETC","SNEShout","ZSNESW"};

const char* getEmulator(int emu) {
  if(emu>6) {
    emu=0;
  }
  return emulators[emu];
}

void KillSPC(SPCInfo *info) {
 //free(info->pAPUBase);
}

void LoadID666Ext(FILE *fp, ID666ext *id666ext) {
  int id;
  unsigned int scount=0;
  unsigned int size;
  
  fread(&size,1,4,fp);
  
  while((id=fgetc(fp))>=0 && scount<size) {
    int type=fgetc(fp);
    scount++;
    
    if(type) {
      unsigned int len;
      unsigned char buf[256];
      unsigned char *ib;
      
      len=fgetc(fp);
      len|=fgetc(fp)<<8;
      scount+=2;
      if(len<=256) {
	fread(buf,1,len,fp);
	buf[255]=0;
	//printf("%02x %08x %s\n",id,*(u32*)buf,buf);
	switch(id) {
	case 0x01:strcpy(id666ext->song,buf);break;
	case 0x02:strcpy(id666ext->game,buf);break;
	case 0x03:strcpy(id666ext->artist,buf);break;
	case 0x04:strcpy(id666ext->dumper,buf);break;
	case 0x05:
	  ib=buf;
	  id666ext->year=*ib;
	  ib++;
	  id666ext->year|=*ib<<8;
	  ib++;
	  id666ext->month=*ib;
	  ib++;	 
	  id666ext->day=*ib;
	  ib++;
	  break;
	case 0x07:strcpy(id666ext->comments,buf);break;
	case 0x10:strcpy(id666ext->OSTTitle,buf);break;
	case 0x13:strcpy(id666ext->publisher,buf);break;
	case 0x30:id666ext->intro=*(unsigned int*)buf;break;
	case 0x31:id666ext->loop=*(unsigned int*)buf;break;
	case 0x32:id666ext->end+=*(unsigned int*)buf;break;
	case 0x33:id666ext->fade=*(unsigned int*)buf;break;
	case 0x36:id666ext->amplification=*(unsigned int*)buf;break;
	}
      } else {
	fseek(fp,len,SEEK_CUR);
      }
      scount+=len;
    } else {
      char data[2]; 
      data[0]=fgetc(fp);
      data[1]=fgetc(fp);
      scount+=2;
      switch(id) {
      case 0x06:id666ext->emulator=*(unsigned int*)data;break;
      case 0x11:id666ext->OSTDisc=*(unsigned int*)data;break;
      case 0x12:
	id666ext->OSTTrackChar=data[0];
	id666ext->OSTTrack=data[1];
	break;
      case 0x14:id666ext->copyrightYear=*(unsigned int*)data;break;
      case 0x34:id666ext->muted=*(unsigned int*)data;break;
      case 0x35:id666ext->loopTimes=*(unsigned int*)data;break;
      }
    }
  }
}

int LoadID666(FILE *fp, ID666 *id666) {
  unsigned char buf[0x100 - 0x2E];
  unsigned char *ib=buf;
  int isbinary = 1;
  
  fread(buf,1,0x100 - 0x2E,fp);
  
  memcpy(id666->song,ib,32);
  id666->song[31]=0;
  ib+=32;

  memcpy(id666->game,ib,32);
  id666->game[31]=0;
  ib+=32;

  memcpy(id666->dumper,ib,16);
  id666->dumper[15]=0;
  ib+=16;

  memcpy(id666->comments,ib,32);
  id666->comments[31]=0;
  ib+=32;

  /* Ok, now comes the fun part. */
  
  /* Date check */
  if(ib[2] == '/'  && ib[5] == '/' )
    isbinary = 0;
  
  /* Reserved bytes check */
  if(ib[0xD2 - 0x2E - 112] >= '0' && ib[0xD2 - 0x2E - 112] <= '9' && ib[0xD3 - 0x2E - 112] == 0x00)
    isbinary = 0;
  //printf("%02x\n",ib[0xD2 - 0x2E - 112]);
  
  //printf("%d\n", isbinary);
  
  id666->isBinary = isbinary;
  
  if(isbinary) {
    id666->year=*ib;
    ib++;
    id666->year|=*ib<<8;
    ib++;
    id666->month=*ib;
    ib++;	 
    id666->day=*ib;
    ib++;

    ib+=7;

    id666->length=*ib;
    ib++;
    
    id666->length|=*ib<<8;
    ib++;
    
    id666->length|=*ib<<16;
    id666->length*=64000;
    ib++;
    
    id666->fade=*ib;
    ib++;
    id666->fade|=*ib<<8; 
    ib++;
    id666->fade|=*ib<<16;
    ib++;
    id666->fade|=*ib<<24;
    id666->fade*=64;
    ib++;

    memcpy(id666->artist,ib,32);
    id666->artist[31]=0;
    ib+=32;

    id666->muted=*ib;
    ib++;

    id666->emulator=*ib;
    ib++;    
  } else {
    int year, month, day;
    int tmp;
    char buf[64];
    
    year=month=day=0;
    ib[10]	= 0;
    sscanf(ib, "%02d/%02d/%04d", &month, &day, &year);
    id666->year = year;
    id666->day = day;
    id666->month = month;
    ib+=11;
    
    memcpy(buf, ib, 3);
    buf[3] = 0;	
    tmp = 0;
    sscanf(buf, "%d", &tmp);
    id666->length = tmp * 64000;
    ib+=3;

    memcpy(buf, ib, 5);
    buf[5] = 0;
    tmp = 0;
    sscanf(buf, "%d", &tmp);
    id666->fade = tmp * 64;
    ib+=5;

    memcpy(id666->artist,ib,32);
    id666->artist[31]=0;
    ib+=32;

    //I have no idea if this is right or not.
    id666->muted=*ib;
    ib++;
    //

    memcpy(buf, ib, 1);
    buf[1] = 0;
    tmp = 0;
    sscanf(buf, "%d", &tmp);
    id666->emulator=tmp;
    ib++;
  }
  return 1;
}

/**
   Loads an SPC
   If pAPURAM is set, get ready for playing
**/
int LoadSPC(char *fn, SPCInfo *info, void *pAPURAM) {
  FILE *fp;
 
  if(!(fp=fopen(fn,"rb"))) 
    return(0); //can't open file
  
  memset(info,0,sizeof(SPCInfo));
  
  fread(&(info->header), 1, sizeof(SPCHeader), fp);
  LoadID666(fp,&info->id666);
  if(pAPURAM) {
    fread(&(info->data), 1, sizeof(SPCData), fp);//only read data if we're going to play
  }

  unsigned char verify[4]={'x','i','d','6'};
  unsigned char bv[4];
  fseek(fp,66048,SEEK_SET);
  fread(bv,1,4,fp);
  if(memcmp(bv,verify,4)) {
    info->hasID666ext=0; //no extended ID666
  } else {
    info->hasID666ext=1;
    LoadID666Ext(fp,&info->id666ext);
  }
  fclose(fp);
  
  info->totallength=info->id666.length+info->id666.fade;
  if(!info->totallength || SEXY_CONFIG.overrideSong)
    info->totallength=(SEXY_CONFIG.lengthMins*60+SEXY_CONFIG.lengthSecs)*64000; //DEFAULT LENGTH

  strncpy(info->filename,fn,255);
  info->filename[255] = 0;

  return(1);
}

int WriteID666(char *fn, ID666 *id666) {
  FILE *fp;
  unsigned char buf[0x100 - 0x2E];
  unsigned char *ob=buf;

  if(!(fp=fopen(fn,"r+b"))) 
    return(0); //can't open file

  memcpy(ob,id666->song,32);
  ob+=32;

  memcpy(ob,id666->game,32);
  ob+=32;

  memcpy(ob,id666->dumper,16);
  ob+=16;

  memcpy(ob,id666->comments,32);
  ob+=32;

  if(id666->isBinary) {
    //year
    ob+=2;
    *ob = id666->month;
    ob++;	 
    *ob = id666->day;
    ob++;

    ob+=7;

    //length
    ob+=3;
    
    //fade
    ob+=4;

    memcpy(ob,id666->artist,32);
    ob += 32;

    *ob = id666->muted;
    ob++;

    *ob = id666->emulator;
    ob++;    
  } else {
    snprintf(ob,11,"%.2d/%.2d/%.4d",id666->month,id666->day,id666->year);
    ob += 11;

    snprintf(ob,4,"%i",(id666->length)/64000);
    ob += 3;

    snprintf(ob,5,"%i",(id666->fade)/64);
    ob += 5;

    memcpy(ob,id666->artist,32);
    ob += 32;

    //I have no idea if this is right or not.
    *ob = id666->muted;
    ob++;
    //

    snprintf(ob,2,"%d",id666->emulator);
    ob += 1;
  }
  fseek(fp,0x2E,SEEK_SET); //move to start of ID666
  fwrite(buf,1,0x100 - 0x2E,fp);
  fclose(fp);

  return 1;
}

int WriteID666ext(char *fn, ID666ext *id666ext) {
  return 0;
}
