#include <gtk/gtkwidget.h>
#include <gtk/gtkwindow.h>
#include <gdk/gdk.h>

#include <xmms/plugin.h>
#include <xmms/util.h>
#include <xmms/titlestring.h>

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "Types.h"
#include "SNESAPU/SPC700.h"
#include "SNESAPU/DSP.h"
#include "SNESAPU/APU.h"
#include "sexyconfig.h"

#include "dialogs/about.h"
#include "dialogs/config.h"
#include "dialogs/file.h"

static volatile int RequestSeek = -1;
static volatile int RequestStop = 0;
static volatile int SongEnded = 0;

static pthread_mutex_t playmutex = PTHREAD_MUTEX_INITIALIZER;

static SPCInfo SPCINFO;

extern InputPlugin sexy_ip;

static void *pAPURAM;

InputPlugin *get_iplugin_info(void) {
  sexy_ip.description = "sexySPC SPC Player 0.5.0";
  return &sexy_ip;
}



/**
   Initialise the plugin
**/
void init(void) {
  //Load config
  readSexyConfig();

  //Allocate RAM -----------------------------
  void *pAPUBase = malloc(131072);

  pAPURAM = (void*)(((u32)pAPUBase + 0xFFFF) & ~0xFFFF);

  InitAPU(pAPURAM);
}




/**
   Show about dialog
**/
static void about(void) {
  gtk_widget_show(create_spc_about());
}




/**
   Show config dialog
**/
static void config(void) {
  gtk_widget_show(create_spc_config());
}



/**
   Show file dialog
**/
static void file(void) {
  gtk_widget_show(create_spc_file(&SPCINFO));
}




/**
   Returns one if this plugin can handle the file
**/
static int testfile(char *fn) {
  FILE *fp;
  char buf[0x1b+1];
  char *testie="SNES-SPC700 Sound File Data";
 
  if(!strncasecmp(fn,"http://",7)) 
    return(0); // Don't handle http://blahblah
  if(!(fp=fopen(fn,"rb"))) 
    return(0);
  if(fread(buf,1,0x1b,fp)!=0x1b) {
    fclose(fp);
    return(0);
  }
  fclose(fp);

  if(strncmp(buf,testie,0x1b)) 
    return(0);
  return(1);
}



/**
   Make the title string
**/
char *makeTitle(SPCInfo info, char *filename) {
  TitleInput *ti;
  XMMS_NEW_TITLEINPUT(ti);

#define DOTITLE(dest,source) (*ti).dest = (strlen(source) ? source : NULL)
  DOTITLE(performer,info.id666.dumper);
  DOTITLE(album_name,info.id666.game);
  DOTITLE(track_name,info.id666.song);
  (*ti).track_number = info.id666ext.OSTTrack;
  (*ti).year = info.id666.year;
  char tmp[11];
  sprintf(tmp,"%.4d/%.2d/%.2d",info.id666.year,info.id666.month,info.id666.day);
  DOTITLE(date,tmp);
  DOTITLE(genre,getEmulator(info.id666.emulator));
  DOTITLE(comment,info.id666.comments);
  DOTITLE(file_name,g_basename(filename));
  char *ext = strrchr(filename,'.');
  if(ext) ext++;
  DOTITLE(file_ext,ext);
  ext = strrchr(filename,'/');
  DOTITLE(file_path,g_strndup(filename,ext-filename+1));
  
  return xmms_get_titlestring(SEXY_CONFIG.titleOverride ? SEXY_CONFIG.titleFormat : xmms_get_gentitle_format(), ti);
}



/**
   Converts bits to what XMMS wants
**/
int ToXMMSBits(void) {
  switch(abs(SEXY_CONFIG.bits)) {
  case SAMPLE_8_BIT: return(FMT_U8);
  case SAMPLE_16_BIT: return(FMT_S16_LE);
  }
  return(0);
}




static pthread_t dethread;
void sexyd_update(void) {
  char buffer[512*4];
  char *Buffer=buffer;
  int count=512*4;
  int mask;

  EmuAPU(Buffer, 512, 1);
  sexy_ip.add_vis_pcm(sexy_ip.output->written_time(), ToXMMSBits(), channels(&SEXY_CONFIG), count/4, Buffer);
  
  mask = ~((((bits(&SEXY_CONFIG) / 8) * channels(&SEXY_CONFIG))) - 1);
 
  while(count>0 && !RequestStop) {
    int t=sexy_ip.output->buffer_free() & mask;
    if(t>count) {
      sexy_ip.output->write_audio(Buffer,count);
    } else {
      unsigned long long tmp;
      if(t)
	sexy_ip.output->write_audio(Buffer,t);

      tmp=count-t;
      tmp*=1000*1000;	// Multiply so we'll have microseconds after the following divide.
      tmp/=rate(&SEXY_CONFIG)*channels(&SEXY_CONFIG)*abs(bits(&SEXY_CONFIG));
      usleep(tmp);
    }
    count-=t;
    Buffer+=t;
  }
}




static void SeekWrapper(long t)	{ // BLAH, SNESAPU crashes on SeekAPU.  Why?
  char Buffer[2048*4];

  t /= 2;

  while(t > 0) {
    if(t>=2048)
      EmuAPU(Buffer, 2048, 1);
    else 
      EmuAPU(Buffer, t, 1);
    t-= 2048;
  }
}




/**
   Main play loop
**/
static void *playloop(void *arg) {
  static long inseek;
  inseek=0;

  for(;;) {
    pthread_mutex_lock(&playmutex);

    if(RequestStop) 
      break;

    if(t64Cnt>=SPCINFO.totallength && SPCINFO.totallength && !SEXY_CONFIG.playIndefinitely)  {
      if(!sexy_ip.output->buffer_playing())
	break;
      if(!inseek) usleep(1000);
    } else if(!inseek) {
      sexyd_update();
    }

    if(RequestSeek != -1) {
      unsigned long t=RequestSeek;

      inseek=t*64000-t64Cnt;

      if(inseek<0) {	/* Reverse seek requested. */
	SetSPC(&SPCINFO, pAPURAM);
	inseek=t*64000;
	sexy_ip.output->flush(0);
      }
      RequestSeek = -1;
    }

    if(inseek) {
      if(inseek<64000) {
	SeekWrapper(inseek);
	inseek=0;
	sexy_ip.output->flush(t64Cnt / 64);
      } else {
	SeekWrapper(64000);
        inseek-=64000;
	sexy_ip.output->flush(t64Cnt / 64);
      }
    }
    pthread_mutex_unlock(&playmutex);
  }

  sexy_ip.output->close_audio();
  SongEnded = 1;
  pthread_mutex_unlock(&playmutex);

  pthread_exit(0);
  return(0);
}


/**
   Sets up everything so that the SPC is ready to play
**/
void SetSPC(SPCInfo *info, void *pAPURAM) {

  SPCRegisters *pReg;

  ResetAPU(0x10000);
  SetAPUOpt(MIX_FLOAT,channels(&SEXY_CONFIG),bits(&SEXY_CONFIG),
	    rate(&SEXY_CONFIG),interpolation(&SEXY_CONFIG),SEXY_CONFIG.dspOptions);

  memcpy(pAPURAM, info->data.APURAM,65536);
  memcpy(&dsp,info->data.dsp,128);
  memcpy(&extraRAM,info->data.extraRAM,64);
  
  pReg = &(info->header.reg);
  FixAPU(*(u16*)pReg->pc, pReg->a, pReg->y, pReg->x,pReg->psw, pReg->sp);
  
  if(SEXY_CONFIG.playIndefinitely) {
    SetAPULength( -1, 0);
  } else {
    SetAPULength(info->totallength-info->id666.fade, info->id666.fade);
  }

  SetDSPAmp(SEXY_CONFIG.amplification*65536);
}


static int paused;
/**
   Start playing
**/
void play(char *fn) {
  sexy_ip.output->open_audio(ToXMMSBits(), rate(&SEXY_CONFIG), channels(&SEXY_CONFIG));
  paused = 0;
  
  if(LoadSPC(fn,&SPCINFO,pAPURAM)) {
    SetSPC(&SPCINFO,pAPURAM);
    RequestSeek = -1;
    RequestStop = 0;
    SongEnded = 0;
    sexy_ip.set_info(makeTitle(SPCINFO,fn),SPCINFO.totallength/64,rate(&SEXY_CONFIG)*channels(&SEXY_CONFIG)*abs(bits(&SEXY_CONFIG)),rate(&SEXY_CONFIG),channels(&SEXY_CONFIG));
    
    if(pthread_mutex_init(&playmutex, NULL))
      puts("Mutex failure!");
    pthread_create(&dethread,0,playloop,0);
  }
}




/**
   Stop playing
**/
void stop(void) {
  RequestStop = 1;
  pthread_mutex_lock(&playmutex);
  if(!SongEnded) {
    if(paused)
      sexy_ip.output->pause(0);
    paused = 0;
  }
  pthread_mutex_unlock(&playmutex);

  pthread_join(dethread,0);
  pthread_mutex_destroy(&playmutex);
  KillSPC(&SPCINFO);
}




/**
   Pause
**/
void sexy_pause(short p) {
  sexy_ip.output->pause(p);
  paused = p;
}




void seek(int time) {
  pthread_mutex_lock(&playmutex);
  RequestSeek = time;
  pthread_mutex_unlock(&playmutex);
}




int gettime(void) {
  if(SongEnded && !RequestStop)
    return(-1);
  else
    return sexy_ip.output->output_time();
}




void getsonginfo(char *fn, char **title, int *length) {
  SPCInfo tmp;
  if(LoadSPC(fn,&tmp,0)) {
    *length=tmp.totallength/64;
    *title=makeTitle(tmp,fn);
  }
}




InputPlugin sexy_ip = {
  0,
  0,
  "SPC Player",
  init,
  about,
  config,
  testfile,
  0,
  play,
  stop,
  sexy_pause,
  seek,
  0,
  gettime,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  getsonginfo,
  file,
  0
};
