;
;SNES Digital Signal Processor Emulator
;                                                        Copyright (C)1999-2004 Alpha-II Productions
;

CPU	PPRO
BITS	32


;
;Header files

%include "Macro.Inc"
%include "APU.Inc"
%include "SPC700.Inc"
%define INTERNAL
%include "DSP.Inc"


;
;Equates

;Envelope mode masks ------------------------
E_TYPE	EQU	00001b								;Type of adj: Constant(1/64 or 1/256) / Exp.(255/256)
E_DIR		EQU	00010b								;Direction: Decrease / Increase
E_DEST	EQU	00100b								;Destination: Default(0 or 1) / Other(x/8 or .75)
E_ADSR	EQU	01000b								;Envelope mode: Gain/ADSR
E_IDLE	EQU	80h									;Envelope speed is set to 0

E_DEC		EQU	00000b								;Linear decrease
E_EXP		EQU	00001b								;Exponential decrease
E_INC		EQU	00010b								;Linear increase
E_BENT	EQU	00110b								;Bent line increase
E_REL		EQU	01000b								;Release mode
E_SUST	EQU	01001b								;Sustain mode
E_ATT		EQU	01010b								;Attack mode
E_DECAY	EQU	01101b								;Decay mode
E_DIRECT	EQU	10000111b							;Direct gain

;Envelope precision -------------------------
E_SHIFT	EQU	4										;Amount to shift envelope to get 8-bit signed value

;Envelope adjustment rates ------------------
A_GAIN	EQU	(1 << E_SHIFT)						;Amount to adjust envelope values
A_LIN		EQU	(128*A_GAIN)/64					;Linear rate to increase/decrease envelope
A_KOF		EQU	(128*A_GAIN)/256					;Rate to decrease envelope during release
A_BENT	EQU	(128*A_GAIN)/256					;Rate to increase envelope after bend
A_NOATT	EQU	(64*A_GAIN)							;Rate to increase if attack rate is set to 0ms
A_EXP		EQU	0										;Rate to decrease envelope exponentially (Not used)

;Envelope destination values ----------------
D_MAX		EQU	(128*A_GAIN)-1						;Maximum envelope value
D_ATTACK	EQU	(128*A_GAIN*63)/64				;Destination of attack rate
D_BENT	EQU	(128*A_GAIN*3)/4					;First destination of bent line
D_EXP		EQU	(128*A_GAIN)/8						;Minimum decay destination value
D_MIN		EQU	0										;Minimum envelope value

;Array sizes --------------------------------
MIX_SIZE	EQU	1024									;Size of mixing buffer in samples
FIRBUF	EQU	2*2*64								;stereo * ring loop * 256kHz / 32kHz
ECHOBUF	EQU	2*((192000*240)/1000)			;Size of echo buffer (stereo * 192kHz * 240ms)

LOWTAPS	EQU	32										;Number of taps in low-pass FIR filter
LOWBUF	EQU	2*2*LOWTAPS


;
;Structures

;Interpolation lookup table
STRUC IntTab
	d0		resw	1
	d1		resw	1
	d2		resw	1
	d3		resw	1
ENDSTRUC



;
;Data

%ifndef	WIN32
SECTION .data ALIGN=256
%else
SECTION .data ALIGN=32
%endif

				;12-bit Gaussian curve generated by SNES DSP
	gaussTab	DW	    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0 ;s0
				DW	   16,   16,   16,   16,   16,   16,   16,   16,   16,   16,   16,   32,   32,   32,   32,   32
				DW	   32,   32,   48,   48,   48,   48,   48,   64,   64,   64,   64,   64,   80,   80,   80,   80
				DW	   96,   96,   96,   96,  112,  112,  112,  128,  128,  128,  144,  144,  144,  160,  160,  160
				DW	  176,  176,  176,  192,  192,  208,  208,  224,  224,  240,  240,  240,  256,  256,  272,  272
				DW	  288,  304,  304,  320,  320,  336,  336,  352,  368,  368,  384,  384,  400,  416,  432,  432
				DW	  448,  464,  464,  480,  496,  512,  512,  528,  544,  560,  576,  576,  592,  608,  624,  640
				DW	  656,  672,  688,  704,  720,  736,  752,  768,  784,  800,  816,  832,  848,  864,  880,  896
				DW	  928,  944,  960,  976,  992, 1024, 1040, 1056, 1072, 1104, 1120, 1136, 1168, 1184, 1216, 1232
				DW	 1248, 1280, 1296, 1328, 1344, 1376, 1392, 1424, 1440, 1472, 1504, 1520, 1552, 1584, 1600, 1632
				DW	 1664, 1696, 1712, 1744, 1776, 1808, 1840, 1872, 1888, 1920, 1952, 1984, 2016, 2048, 2080, 2112
				DW	 2144, 2192, 2224, 2256, 2288, 2320, 2352, 2400, 2432, 2464, 2496, 2544, 2576, 2608, 2656, 2688
				DW	 2736, 2768, 2800, 2848, 2880, 2928, 2976, 3008, 3056, 3088, 3136, 3184, 3216, 3264, 3312, 3360
				DW	 3392, 3440, 3488, 3536, 3584, 3632, 3680, 3728, 3776, 3824, 3872, 3920, 3968, 4016, 4064, 4112
				DW	 4160, 4208, 4272, 4320, 4368, 4416, 4480, 4528, 4576, 4640, 4688, 4752, 4800, 4864, 4912, 4976
				DW	 5024, 5088, 5136, 5200, 5248, 5312, 5376, 5424, 5488, 5552, 5616, 5664, 5728, 5792, 5856, 5920	;e0
				DW	 5984, 6048, 6096, 6160, 6224, 6288, 6352, 6416, 6480, 6560, 6624, 6688, 6752, 6816, 6880, 6944 ;s1
				DW	 7024, 7088, 7152, 7216, 7296, 7360, 7424, 7504, 7568, 7632, 7712, 7776, 7856, 7920, 7984, 8064
				DW	 8128, 8208, 8272, 8352, 8432, 8496, 8576, 8640, 8720, 8800, 8864, 8944, 9008, 9088, 9168, 9232
				DW	 9312, 9392, 9472, 9536, 9616, 9696, 9776, 9840, 9920,10000,10080,10160,10240,10304,10384,10464
				DW	10544,10624,10704,10784,10848,10928,11008,11088,11168,11248,11328,11408,11488,11568,11648,11712
				DW	11792,11872,11952,12032,12112,12192,12272,12352,12432,12512,12592,12672,12752,12832,12896,12976
				DW	13056,13136,13216,13296,13376,13456,13536,13616,13680,13760,13840,13920,14000,14080,14144,14224
				DW	14304,14384,14464,14528,14608,14688,14768,14832,14912,14992,15056,15136,15216,15280,15360,15440
				DW	15504,15584,15648,15728,15808,15872,15952,16016,16080,16160,16224,16304,16368,16432,16512,16576
				DW	16640,16720,16784,16848,16912,16976,17056,17120,17184,17248,17312,17376,17440,17504,17568,17632
				DW	17696,17744,17808,17872,17936,18000,18048,18112,18176,18224,18288,18336,18400,18448,18512,18560
				DW	18624,18672,18720,18784,18832,18880,18928,18976,19040,19088,19136,19184,19232,19280,19312,19360
				DW	19408,19456,19504,19536,19584,19632,19664,19712,19744,19792,19824,19856,19904,19936,19968,20016
				DW	20048,20080,20112,20144,20176,20208,20240,20272,20304,20320,20352,20384,20400,20432,20464,20480
				DW	20512,20528,20544,20576,20592,20608,20640,20656,20672,20688,20704,20720,20736,20752,20752,20768
				DW	20784,20800,20800,20816,20832,20832,20848,20848,20848,20864,20864,20864,20864,20864,20880,20880	;e1
				DW	20880,20880,20864,20864,20864,20864,20864,20848,20848,20848,20832,20832,20816,20800,20800,20784 ;s2
				DW	20768,20752,20752,20736,20720,20704,20688,20672,20656,20640,20608,20592,20576,20544,20528,20512
				DW	20480,20464,20432,20400,20384,20352,20320,20304,20272,20240,20208,20176,20144,20112,20080,20048
				DW	20016,19968,19936,19904,19856,19824,19792,19744,19712,19664,19632,19584,19536,19504,19456,19408
				DW	19360,19312,19280,19232,19184,19136,19088,19040,18976,18928,18880,18832,18784,18720,18672,18624
				DW	18560,18512,18448,18400,18336,18288,18224,18176,18112,18048,18000,17936,17872,17808,17744,17696
				DW	17632,17568,17504,17440,17376,17312,17248,17184,17120,17056,16976,16912,16848,16784,16720,16640
				DW	16576,16512,16432,16368,16304,16224,16160,16080,16016,15952,15872,15808,15728,15648,15584,15504
				DW	15440,15360,15280,15216,15136,15056,14992,14912,14832,14768,14688,14608,14528,14464,14384,14304
				DW	14224,14144,14080,14000,13920,13840,13760,13680,13616,13536,13456,13376,13296,13216,13136,13056
				DW	12976,12896,12832,12752,12672,12592,12512,12432,12352,12272,12192,12112,12032,11952,11872,11792
				DW	11712,11648,11568,11488,11408,11328,11248,11168,11088,11008,10928,10848,10784,10704,10624,10544
				DW	10464,10384,10304,10240,10160,10080,10000, 9920, 9840, 9776, 9696, 9616, 9536, 9472, 9392, 9312
				DW	 9232, 9168, 9088, 9008, 8944, 8864, 8800, 8720, 8640, 8576, 8496, 8432, 8352, 8272, 8208, 8128
				DW	 8064, 7984, 7920, 7856, 7776, 7712, 7632, 7568, 7504, 7424, 7360, 7296, 7216, 7152, 7088, 7024
				DW	 6944, 6880, 6816, 6752, 6688, 6624, 6560, 6480, 6416, 6352, 6288, 6224, 6160, 6096, 6048, 5984	;e2
				DW	 5920, 5856, 5792, 5728, 5664, 5616, 5552, 5488, 5424, 5376, 5312, 5248, 5200, 5136, 5088, 5024 ;s3
				DW	 4976, 4912, 4864, 4800, 4752, 4688, 4640, 4576, 4528, 4480, 4416, 4368, 4320, 4272, 4208, 4160
				DW	 4112, 4064, 4016, 3968, 3920, 3872, 3824, 3776, 3728, 3680, 3632, 3584, 3536, 3488, 3440, 3392
				DW	 3360, 3312, 3264, 3216, 3184, 3136, 3088, 3056, 3008, 2976, 2928, 2880, 2848, 2800, 2768, 2736
				DW	 2688, 2656, 2608, 2576, 2544, 2496, 2464, 2432, 2400, 2352, 2320, 2288, 2256, 2224, 2192, 2144
				DW	 2112, 2080, 2048, 2016, 1984, 1952, 1920, 1888, 1872, 1840, 1808, 1776, 1744, 1712, 1696, 1664
				DW	 1632, 1600, 1584, 1552, 1520, 1504, 1472, 1440, 1424, 1392, 1376, 1344, 1328, 1296, 1280, 1248
				DW	 1232, 1216, 1184, 1168, 1136, 1120, 1104, 1072, 1056, 1040, 1024,  992,  976,  960,  944,  928
				DW	  896,  880,  864,  848,  832,  816,  800,  784,  768,  752,  736,  720,  704,  688,  672,  656
				DW	  640,  624,  608,  592,  576,  576,  560,  544,  528,  512,  512,  496,  480,  464,  464,  448
				DW	  432,  432,  416,  400,  384,  384,  368,  368,  352,  336,  336,  320,  320,  304,  304,  288
				DW	  272,  272,  256,  256,  240,  240,  240,  224,  224,  208,  208,  192,  192,  176,  176,  176
				DW	  160,  160,  160,  144,  144,  144,  128,  128,  128,  112,  112,  112,   96,   96,   96,   96
				DW	   80,   80,   80,   80,   64,   64,   64,   64,   64,   48,   48,   48,   48,   48,   32,   32
				DW	   32,   32,   32,   32,   32,   16,   16,   16,   16,   16,   16,   16,   16,   16,   16,   16
				DW	    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0 ;e3

				;Jump table for DSP register writes (see DSPIn)
	dspRegs	DD	RVolL,	RVolR,	RPitch,	RPitch,	RNull,	RADSR,	RADSR,	RGain
				DD	RNull,	RNull,	RNull,	RNull,	RMVolL,	REFB,		RNull,	RFCI
				DD	RVolL,	RVolR,	RPitch,	RPitch,	RNull,	RADSR,	RADSR,	RGain
				DD	RNull,	RNull,	RNull,	RNull,	RMVolR,	RNull,	RNull,	RFCI
				DD	RVolL,	RVolR,	RPitch,	RPitch,	RNull,	RADSR,	RADSR,	RGain
				DD	RNull,	RNull,	RNull,	RNull,	REVolL,	RPMOn,	RNull,	RFCI
				DD	RVolL,	RVolR,	RPitch,	RPitch,	RNull,	RADSR,	RADSR,	RGain
				DD	RNull,	RNull,	RNull,	RNull,	REVolR,	RNull,	RNull,	RFCI
				DD	RVolL,	RVolR,	RPitch,	RPitch,	RNull,	RADSR,	RADSR,	RGain
				DD	RNull,	RNull,	RNull,	RNull,	RKOn,		RNull,	RNull,	RFCI
				DD	RVolL,	RVolR,	RPitch,	RPitch,	RNull,	RADSR,	RADSR,	RGain
				DD	RNull,	RNull,	RNull,	RNull,	RKOf,		RNull,	RNull,	RFCI
				DD	RVolL,	RVolR,	RPitch,	RPitch,	RNull,	RADSR,	RADSR,	RGain
				DD	RNull,	RNull,	RNull,	RNull,	RFlg,		RNull,	RNull,	RFCI
				DD	RVolL,	RVolR,	RPitch,	RPitch,	RNull,	RADSR,	RADSR,	RGain
				DD	RNull,	RNull,	RNull,	RNull,	RNull,	REDl,		RNull,	RFCI

				;Registers that require special handling for output type
	dspRegsI	DD	RVolL,	RVolR,	RMVolL,	RMVolR,	REVolL,	REVolR,	REFB,		RFCI	;Integer
	dspRegsM	DD	RVolLM,	RVolRM,	RMVolLM,	RMVolRM,	REVolLM,	REVolRM,	REFBM,	RFCI	;Monaural
	dspRegsF	DD	RVolLF,	RVolRF,	RMVolLF,	RMVolRF,	REVolLF,	REVolRF,	REFBF,	RFCF	;Floating-point

				;Pointers to each mixing routine:  none, integer (386), mmx, and float
	mixRout	DD	EmuDSPN,	EmuDSPI,	EmuDSPX, EmuDSPF

				;Pointers to interpolation functions for each mixing routine
	intRout	DD	NoneI,	LinearI,	CubicI,	GaussI
				DD	NoneX,	LinearX,	CubicX,	GaussX
				DD	NoneF,	LinearF,	CubicF,	GaussF

	;Masks for MMX gaussian interpolation ----
	loSmp		DD	00000FFFFh,00000FFFFh
	hiSmp		DD	0FFFF0000h,0FFFF0000h

	;Frequency table -------------------------
	freqTab	DD	   0
				DD	2048, 1536, 1280					;Number of samples between updates.  Used to determine
				DD	1024,  768,  640					;envelope rates and noise frequencies
				DD	 512,  384,  320
				DD	 256,  192,  160
				DD	 128,   96,   80
				DD	  64,   48,   40
				DD	  32,   24,   20
				DD	  16,   12,   10
				DD	   8,    6,    5
				DD	   4,    3
				DD	   2
				DD	   1

	;Floating point constants ----------------
	fn2_5		DD	-2.5									;Cubic interpolation
	fn1_5		DD	-1.5
	fn0_5		DD	-0.5
	fp0_5		DD	0.5
	fp1_5		DD	1.5

	fp0_46	DD	0.46									;Low-pass filter
	fp0_54	DD	0.54

	fp32km1	DD	32767.0

	fp6		DD	6.0									;VMax2dB()
	fp90		DD	90.0

	Scale32	fp64k,16									;Various
	Scale32	fp256,8									;Number of points of interpolation between samples
	Scale32	fp128,7									;Stereo separation
	Scale32	fpShR7,-7								;Voice volume
	Scale32	fpShR11,-11								;Main/Echo volume(7), 8 voices(3), Echo + Main(1)
	Scale32	fpShR15,-15								;Cubic interpolation
	Scale32	fpShR16,-16
	Scale32	fpShR23,-23								;EFBCT(16), EFB(7)
	Scale32	fpShR30,-30
	Scale32	fpShR31,-31
	Scale32	fpEShR,-(E_SHIFT+7)




;
; Variables

%ifndef	WIN32
SECTION .bss ALIGN=256
%else
SECTION .bss ALIGN=64

;The BSS must be aligned on at least a 256-byte boundary.  (If it's not, you'll know as soon as you
;play a song.)  This is tricky in Windows because the win32 object files can only specify 64-byte
;alignment.  Try padding with multiples of 64 until it works.

				resb	64								;Force page alignment
%endif

;Don't touch!  All arrays are carefully aligned on large boundaries to facillitate easier indexing
;and better cache utilization.

	;DSP Core ---------------------------- [0]
	mix		resb	8*128								;<VoiceMix> Mixing settings for each voice
	dsp		resb	128								;<DSPRAM> DSP registers

	;Look-up Tables -------------------- [480]
	rateTab	resd	32									;Update Rate Table
	brrTab	resd	16*64								;All possible range/nybble values for BRR
	cubicTab	resq	256								;<IntTab>

	;Globals -------------------------- [1D00]
	pAPURAM	resd	1									;-> APU RAM
	pTrace	resd	1									;-> Debugging vector

	pOutBuf	resd	1									;-> output buffer
	outLeft	resd	1									;Number of samples left to fill output buffer
	outCnt	resd	1									;t64 count at last call to EmuDSP
	outDec	resd	1									;Fractional number of samples to be generated

	;DSP Options ---------------------- [1D18]
	procType	resb	1									;Type of host CPU (returned by GetProcType)
	dspMix	resb	1									;Mixing routine

	dspChn	resb	1									;Number of channels being output

	dspSize	resb	1									;Size of samples in bytes

	dspRate	resd	1									;Output sample rate
	pitchBas	resd	1									;Base sample rate
	pitchAdj	resd	1									;Amount to adjust pitch rates [16.16]

	pInter	resd	1									;-> interpolation function
	dspInter	resb	1									;Interpolation method
	voiceMix	resb	1									;Voices that are currently being mixed

	dspOpts	resb	1									;Option flags passed to SetDSPOpt
	surround	resb	1									;Turn on surround sound
	pDecomp	resd	1									;-> sample decompression routine

	;Volume --------------------------- [1D34]
	volSepar	resd	1									;Stereo separation
	volRamp	resd	1									;Amount to ramp volume per sample

	volAmp	resd	1									;Amplification [16.16]
	volAtten	resd	1									;Global volume attenuation [1.16]
	volAdj	resd	1									;Amount to adjust main volumes [-15.16]

	volMainL	resd	1									;Main volume
	volMainR	resd	1
	volEchoL	resd	1									;Echo volume
	volEchoR	resd	1

	vMMaxL	resd	1									;Maximum absolute sample output
	vMMaxR	resd	1

	;Noise ---------------------------- [1D60]
	nRate		resd	1									;Noise sample rate reciprocal [.32]
	nAcc		resd	1									;Noise accumulator [.32] (>= 1 generate a new sample)
	nSmp		resd	1									;Current Noise sample
				resd	1

	;Echo filtering ------------------- [1D70]
	firCur	resd	1									;Index of the first sample to feed into the filter
	firRate	resd	1									;Rate to feed samples into filter
	firDec	resd	1									;Temp
	firEnabl	resb	1									;0 if filtering is disabled (each bit C0-C7 != 0)
	disEcho	resb	1									;0 if echo is enabled
				resb	2									;	[0] - FIR coefficients are all 0
															;	[4] - User has disabled echo
															;	[5] - Echo is disabled in FLG register
	firTaps	resd	2*8								;Filter coefficents (doubled for stereo MMX)

	;Echo ----------------------------- [1DC0]
	echoDel	resd	1									;Size of delay (in bytes)
	echoCur	resd	1									;Current sample in echo area
				resd	1
	efbct		resd	1									;User specified echo feedback crosstalk
	echoFB	resd	2									;Echo feedback
	echoFBCT	resd	2									;Echo feedback crosstalk

	;Low-pass filter ------------------ [1DE0]
	realRate	resd	1									;Real output sample rate, not emulated rate
	realOutS	resd	1									;Real size of output buffer

	lowAmp	resd	1									;Amplification to apply during filtering
	lowCur	resd	1
	lowRate	resd	1									;Rate to feed samples into filter [16.16]
	lowDec	resd	1									;Current sample fraction [0.16]
	lowRFI	resd	2									;Current RFI simulation sample
				resd	28

	;Single source playback ----------- [1E70]
	tBRR		resb	8									;Temporary buffer for storing BRR block
				resw	4									;Temporary buffer for single sound playback
	tBuf		resw	16									;
	tRate		resd	1
	tDec		resd	1
	tLoop		resd	1
	tBlk		resd	1
	tIdx		resd	1
	tP1		resd	1
	tP2		resd	1
				resd	1

	konWr		resd	1									;Last time the KON register was written to
	songLen	resd	1									;Length of song (in ticks)
	fadeLen	resd	1									;Length of fade (in ticks)
	outRate	resd	1

				resd	12

	;Storage buffers ------------------ [1F00]
	mixBuf	resd	2*2*MIX_SIZE					;Temporary mixing buffer (linear buffer)
	echoBuf	resd	ECHOBUF							;External echo memory, 240ms @ 192kHz (ring buffer)
	firBuf	resd	FIRBUF							;Unaltered echo samples fed into FIR filter (ring buffer)
	lowBuf	resd	LOWBUF							;Interpolated samples fed into low-pass FIR filter (ring buffer)
	lowTaps	resd	LOWTAPS							;Filter coefficients used for low-pass filter


;
; Code

SECTION .text ALIGN=16

;
;Get Processor Information

PROC GetProcInfo
USES ECX,EDX,EBX

	;Test for CPUID instruction --------------
	PushFD
	Mov	EDX,[ESP]									;EDX = EFLAGS
	XOr	dword [ESP],200000h						;Flip CPUID support flag
	PopFD													;Restore EFLAGS
	PushFD
	Pop	EAX											;EAX = New EFLAGS
	XOr	EAX,EDX										;Does processor support the CPUID instruction?
	JZ		.Done											;  No, Return 0

	XOr	EAX,EAX										;Get the number of extended functions
	CPUID
	Test	EAX,EAX										;Does CPU have any?
	JZ		.Done											;  No, Return 0

	;Test for MMX support --------------------
	Mov	EAX,1											;EAX = Get extended information
	CPUID
	XOr	EAX,EAX
	BT		EDX,23										;Does CPU have MMX support?
	JNC	.Done											;  No, Return 0

	;Check processor manufacturer ------------
	XOr	EAX,EAX
	CPUID

	Cmp	EBX,"Genu"									;Is processor an Intel?
	SetE	AL
	Cmp	EDX,"ineI"
	SetE	AH
	Or		AL,AH
	Cmp	ECX,"ntel"
	SetE	AH
	Or		AL,AH
	JZ		.NotIntel
		Mov	EAX,1
		CPUID
		BT		EDX,25									;Does CPU have SIMD support?
		SetC	AL											;  Yes
		MovZX	EAX,AL
		ShL	EAX,2
		Or		AL,1										;Return MMX support as well
		RetN
	.NotIntel:

	Cmp	EBX,"Auth"									;Is processor an AMD?
	SetE	AL
	Cmp	EDX,"enti"
	SetE	AH
	Or		AL,AH
	Cmp	ECX,"cAMD"
	SetE	AH
	Or		AL,AH
	JZ		.NotAMD
		Mov	EAX,80000001h
		CPUID
		BT		EDX,31									;Does CPU have 3DNow! support?
		SetC	AL											;  Yes
		MovZX	EAX,AL
		Add	EAX,EAX
		Or		AL,1
		RetN
	.NotAMD:

	Mov	EAX,1

	.Done:

ENDP


;
;Calculate Power of e
;
;Desc:
;   Calculates e to the power of x by using the formula:
;
;    2^(x*log2(e))
;
;   Where 2^x is calculated by:
;
;    2^int(x) * 2^frac(x)
;
;In:
;   ST = x
;
;Out:
;   ST = e^x

PROC Exp

	FStCW		[ESP-4]									;Save control state
	FStCW		[ESP-8]									;Set FPU to truncate when rounding
	Or			byte [ESP-7],1100b
	FLdCW		[ESP-8]

	FLdL2e												;									|x Log2(e)
	FMulP		ST1,ST									;									|x*Log2(e)

	FLd		ST											;									|ex ex
	FRndInt												;Get the integer portion	|ex floor(ex)

	FXch		ST1										;									|iex ex
	FSub		ST,ST1									;Get the fractional portion|iex ex-iex

	F2XM1													;Compute 2^frac				|iex pow(2,fex)-1
	FLd1													;									|iex p 1
	FAddP		ST1,ST									;									|iex p+1

	FXch													;Compute 2^int					|f iex
	FLd1													;									|f iex 1
	FScale												;									|f iex 1<<iex
	FStp		ST1										;									|f i

	FMulP		ST1,ST									;									|f*i

	FLdCW		[ESP-4]									;Restore control state

ENDP


;
;Initialize DSP

PROC InitDSP, pAPUBase
LOCALS ipD												;Integer, positive, delta
USES ECX,EDX,EBX,ESI,EDI

	Mov	EDI,[pAPUBase]								;Save pointer to SPC RAM
	Mov	[pAPURAM],EDI

	Mov	byte [dspMix],-1							;Reset values so SetDSPOpt will create new ones
	Mov	byte [dspChn],-1
	Mov	byte [dspSize],-1
	Mov	dword [dspRate],-1
	Mov	dword [realRate],32000
	Mov	dword [pitchBas],32000
	Mov	byte [dspInter],-1
	Mov	byte [dspOpts],0
	Mov	byte [disEcho],0
	Mov	dword [volSepar],0
	Mov	dword [efbct],10000h

	Mov	EDI,mix										;Erase all mixer settings
	XOr	EAX,EAX
	Mov	ECX,(8*128)/4
	Rep	StoSD

	;Build a look-up table for all possible expanded values in a BRR block.
	Mov	EDI,brrTab
	XOr	EBX,EBX										;EBX = Nybble to shift right by range
	Mov	CL,28											;ECX = Max range (+16 for 32-bit numbers)
	.Range:
		.Nybble:
			Mov	EAX,EBX								;EAX = Nybble >> Range
			SAR	EAX,CL
			And	EAX,~1								;All numbers used by DSP are even
			Mov	[EDI],EAX
			Add	EDI,4

		Add	EBX,10000000h							;Add 1 to uppermost nybble
		JNZ	short .Nybble

		Add	EDI,0C0h

	Dec	CL
	Cmp	CL,15
	JA 	short .Range

	Mov	BL,3
	XOr	ECX,ECX
	.Invalid:
		XOr	EAX,EAX									;Positive nybbles turn into 0 when range > 12
		Mov	CL,8
		Rep	StoSD

		Mov	EAX,-4096								;Negative nybbles turn into -4096 when range > 12
		Mov	CL,8
		Rep	StoSD

		Add	EDI,0C0h

	Dec	BL
	JNZ	short .Invalid

	;Build a look-up table to calculate a cubic spline with only four integer multiplies.
	;The table is built from the following equation, simplified for s:
	;
	; y = ax^3 + bx^2 + cx + d
	;
	;     3 (s[0] - s[1]) - s[-1] + s[2]
	; a = ------------------------------
	;                   2
	;
	;                      5 s[0] + s[2]
	; b = 2 s[1] + s[-1] - -------------
	;                            2
	;
	;     s[1] - s[-1]
	; c = ------------
	;          2
	;
	; d = s[0]
	;
	;y is the return sample
	;x is the delta from current sample
	;s is a four sample array with [0] being the current sample

	FInit													;Reset FPU, otherwise there'll be problems
	Mov	dword [ipD],0								;Start with a delta of 0 (calculate 256 points)

	Mov	EDI,cubicTab								;EDI->Cubic array				|FPU Stack after execution
	.NextC:
		;x1=(n/256)  x2=(n/256)^2  x3=(n/256)^3
		FILd	dword [ipD]								;Load (int) delta				|D
		FDiv	dword [fp256]							;Divide delta by 256			|D/256=X1
		FLd	ST											;Copy top of stack			|X1 X1
		FMul	ST,ST1									;Square point					|X1 X1*X1=X2
		FLd	ST											;									|X1 X2 X2
		FMul	ST,ST2									;Cube point						|X1 X2 X2*X1=X3

		;s[-1] *= -.5(x^3) + (x^2) - .5x ------
		FLd	dword [fn0_5]							;									|X1 X2 X3 -0.5
		FMul	ST,ST3									;									|X1 X2 X3 -0.5*X1=T1
		FAdd	ST,ST2									;									|X1 X2 X3 T1+X2
		FLd	dword [fn0_5]							;									|X1 X2 X3 T1 -0.5
		FMul	ST,ST2									;									|X1 X2 X3 T1 -0.5*X3=T2
		FAddP ST1,ST									;									|X1 X2 X3 T1+T2
		FMul	dword [fp32km1]						;Convert to fixed point (-.15)
		FIStP	word [EDI]								;Store value in cubicTab	|X1 X2 X3

		;s[0] *= 1.5(x^3) - 2.5(x^2) + 1 ------
		FLd	dword [fn2_5]							;									|X1 X2 X3 -2.5
		FMul	ST,ST2									;									|X1 X2 X3 -2.5*X2=T1
		FLd	dword [fp1_5]							;									|X1 X2 X3 T1 1.5
		FMul	ST,ST2									;									|X1 X2 X3 T1 1.5*X3=T2
		FLd1												;									|X1 X2 X3 T1 T2 1.0
		FAddP ST1,ST									;									|X1 X2 X3 T1 T2+1
		FAddP ST1,ST									;									|X1 X2 X3 T1+T2
		FMul	dword [fp32km1]
		FIStP	word [2+EDI]							;									|X1 X2 X3

		;s[1] *= -1.5(x^3) + 2(x^2) + .5x -----
		FLd	dword [fp0_5]							;									|X1 X2 X3 0.5
		FMul	ST,ST3									;									|X1 X2 X3 0.5*X1=T1
		FLd	ST2										;									|X1 X2 X3 T1 X2
		FAdd	ST,ST3									;									|X1 X2 X3 T1 X2+X2=T2
		FLd	dword [fn1_5]							;									|X1 X2 X3 T1 T2 -1.5
		FMul	ST,ST3									;									|X1 X2 X3 T1 T2 -1.5*X3=T3
		FAddP ST1,ST									;									|X1 X2 X3 T1 T2+T3
		FAddP ST1,ST									;									|X1 X2 X3 T1+T2
		FMul	dword [fp32km1]
		FIStP	word [4+EDI]							;									|X1 X2 X3

		;s[2] *= .5(x^3) - .5(x^2) ------------
		FLd	dword [fn0_5]							;									|X1 X2 X3 -0.5
		FMul	ST,ST2									;									|X1 X2 X3 -0.5*X2=T1
		FLd	dword [fp0_5]							;									|X1 X2 X3 T1 0.5
		FMul	ST,ST2									;									|X1 X2 X3 T1 0.5*X3=T2
		FAddP ST1,ST									;									|X1 X2 X3 T1+T2
		FMul	dword [fp32km1]
		FIStP	word [6+EDI]							;									|X1 X2 X3
		Add	EDI,8

		FStP	ST											;Pop X's off stack			|X1 X3
		FStP	ST											;									|X3
		FStP	ST											;									|(empty)

	Inc	byte [ipD]
	JNZ 	.NextC

%if SNESINT=0
	;Build a look-up table based on the Gaussian curve.
	;The table is built from the following equation:
	;
	;                  2
	;            (x/pi)
	;          - -------
	; y = A * e     2
	;
	;       ____
	;     \| 2pi * o
	; A = ----------
	;         n
	;
	;o is the mean value (2^15)
	;n is the number of points to interpolate between (4)

	Mov	dword [ipD],-512
	Mov	EDI,gaussTab								;EDI->Cubic array				|FPU Stack after execution
	FLd	dword [fpA]									;(sqrt(2 * pi) * 32768) / 4|A
	FLd	dword [fp512]								;									|A 512
	FLdPi													;									|A 512 3.14
	FDivP	ST1,ST										;									|A 512/3.14
	FLd	dword [fp256]								;Load 256 into FPU			|A pi 256.0
	.NextG:
		FILd	dword [ipD]								;Load (int) delta				|A pi 256 x

		FLd	ST											;									|A pi 256 x x
		FDiv	ST,ST3									;									|A pi 256 x x/pi
		FMul	ST,ST										;									|A pi 256 x p^2
		FMul	dword [fpShR1]							;									|A pi 256 x p/2
		FChS												;									|A pi 256 x -p
		Call	Exp										;									|A pi 256 x e^p
		FMul	ST,ST4									;									|A pi 256 x e*A
		FIStP	word [6+EDI]							;									|A pi 256 x

		FAdd	ST,ST1									;									|A pi 256 x+256
		FLd	ST											;									|A pi 256 x x
		FDiv	ST,ST3									;									|A pi 256 x x/pi
		FMul	ST,ST										;									|A pi 256 x p^2
		FMul	dword [fpShR1]							;									|A pi 256 x p/2
		FChS												;									|A pi 256 x -p
		Call	Exp										;									|A pi 256 x e^p
		FMul	ST,ST4									;									|A pi 256 x e*A
		FIStP	word [4+EDI]							;									|A pi 256 x

		FAdd	ST,ST1									;									|A pi 256 x+256
		FLd	ST											;									|A pi 256 x x
		FDiv	ST,ST3									;									|A pi 256 x x/pi
		FMul	ST,ST										;									|A pi 256 x p^2
		FMul	dword [fpShR1]							;									|A pi 256 x p/2
		FChS												;									|A pi 256 x -p
		Call	Exp										;									|A pi 256 x e^p
		FMul	ST,ST4									;									|A pi 256 x e*A
		FIStP	word [2+EDI]							;									|A pi 256 x

		FAdd	ST,ST1									;									|A pi 256 x+256
		FDiv	ST,ST2									;									|A pi 256 x/pi
		FMul	ST,ST										;									|A pi 256 p^2
		FMul	dword [fpShR1]							;									|A pi 256 p/2
		FChS												;									|A pi 256 -p
		Call	Exp										;									|A pi 256 e^p
		FMul	ST,ST3									;									|A pi 256 e*A
		FIStP	word [EDI]								;									|A pi 256
		Add	EDI,8

	Inc	byte [ipD]
	JNZ 	.NextG

	FStP	ST												;									|A 256
	FStP	ST												;									|256
	FStP	ST												;									|(empty)
%else
	;Interleave Gaussian table ---------------
	Mov	ESI,gaussTab
	Mov	EDI,mixBuf
	Mov	ECX,256*2
	Rep	MovSD

	Mov	ESI,mixBuf
	Mov	EDI,gaussTab
	XOr	CL,CL
	.NextG:
		Mov	AX,[ESI]
		Mov	[6+EDI],AX
		Mov	AX,[512+ESI]
		Mov	[4+EDI],AX
		Mov	AX,[1024+ESI]
		Mov	[2+EDI],AX
		Mov	AX,[1536+ESI]
		Mov	[0+EDI],AX
		Add	EDI,8
		Add	ESI,2
	Dec	CL
	JNZ	short .NextG
%endif

	Call	GetProcInfo
	Mov	[procType],AL

	Call	SetDSPOpt,1,2,16,32000,INT_GAUSS,0
	Call	SetDSPDbg,0

ENDP


;
;Reset DSP Settings

PROC ResetDSP
USES ECX,EBX,EDI

	XOr	EAX,EAX

	;Erase DSP Registers ---------------------
	Mov	EDI,dsp
	Mov	ECX,128/4
	Rep	StoSD
	Mov	byte [dsp+flg],0E0h						;Place DSP in power up mode

	;Erase internal mixing settings ----------
	Mov	AH,8
	Mov	EBX,dsp
	Mov	EDI,mix
	.ClrMix:
		Mov	[EDI+dspv],EBX
		Add	EBX,10h
		Add	EDI,4
		Mov	CL,mFlg-4
		Rep	StoSB
		And	byte [EDI],MFLG_MUTE					;Leave voice muted
		Or 	byte [EDI],MFLG_OFF					;Voice is inactive
		Inc	EDI
		Mov	CL,7Fh-mFlg
		Rep	StoSB
	Dec	AH
	JNZ	short .ClrMix

	;Erase global volume settings ------------
	Mov	[volMainL],EAX
	Mov	[volMainR],EAX
	Mov	[volEchoL],EAX
	Mov	[volEchoR],EAX

	;Erase noise settings --------------------
	Mov	[nRate],EAX
	Mov	[nAcc],EAX
	Mov	[nSmp],EAX

	;Echo region -----------------------------
	Mov	EDI,echoBuf
	Mov	ECX,ECHOBUF
	Rep	StoSD

	Mov	[echoCur],EAX								;Reset echo variables
	Mov	dword [echoDel],8							;Delay 1 sample
	Mov	[echoFB],EAX
	Mov	[4+echoFB],EAX
	Mov	[echoFBCT],EAX
	Mov	[4+echoFBCT],EAX

	;Echo filter -----------------------------
	Mov	EDI,firBuf									;Erase filter memory
	Mov	ECX,FIRBUF
	Rep	StoSD

	Mov	EDI,firTaps									;Reset filter coefficients
	Mov	CL,2*8
	Rep	StoSD

	Mov	[firCur],EAX								;Reset filter variables
	Mov	[firDec],EAX
	Mov	[firEnabl],AL

	;Reset low-pass filter -------------------
	Mov	EDI,lowBuf
	Mov	ECX,LOWBUF
	Rep	StoSD

	Mov	dword [lowCur],lowBuf
	Mov	[lowDec],EAX
	Mov	[lowRFI],EAX
	Mov	[4+lowRFI],EAX

	;Disable voices --------------------------
	Mov	[voiceMix],AL

	Mov	[vMMaxL],EAX
	Mov	[vMMaxR],EAX

	;Reset times -----------------------------
	Mov	[outLeft],EAX
	Mov	[outCnt],EAX
	Mov	[outDec],EAX
	Mov	[konWr],EAX
	Mov	dword [songLen],-1
	Mov	dword [fadeLen],1

	;Reset fade volume -----------------------
	Call	SetDSPVol,10000h

ENDP


;
;Calculate Anti-Aliasing Filter Coefficients
;
;Adapted from code in the SoundTouch sound processing library.

PROC AACoeffs
USES ALL

	FLdZ													;								|sum
	Mov	dword [ESP-4],13000						;Cutoff frequency 13kHz
	FILd	dword [ESP-4]								;fc2 = 2.0 * Cutoff 		|sum lowCut
	FIDiv	dword [realRate]							;								|sum lowCut/realRate
	FAdd	ST,ST											;								|sum ratio*2.0

	FLdPi													;wc = Pi * fc2				|sum fc2 Pi
	FMul	ST,ST1										;								|sum fc2 Pi*fc2

	FLdPi													;co = 2*PI / NumTaps		|sum fc2 wc Pi
	FAdd	ST,ST											;								|sum fc2 wc 2*Pi
	Mov	dword [ESP-4],LOWTAPS
	FIDiv	dword [ESP-4]								;								|sum fc2 wc 2Pi/num

	XOr	EBX,EBX										;EBX = Tap index
	XOr	EAX,EAX
	.Next:
		Mov		ECX,EBX								;cnt = i - NumTaps / 2
		Sub		ECX,LOWTAPS/2
		Mov		[ESP-4],ECX
		FILd		dword [ESP-4]						;								|sum fc2 wc co cnt

		Test		ECX,ECX								;if (cnt != 0)				|sum fc2 wc co cnt
		JZ			short .Zero

			FLd		ST									;temp = cnt * wc			|sum fc2 wc co cnt cnt
			FMul		ST,ST3							;								|sum fc2 wc co cnt cnt*wc

			FLd		ST									;sin(temp) * fc2 / temp	|sum fc2 wc co cnt temp temp
			FSin											;								|sum fc2 wc co cnt temp sin(temp)
			FMul		ST,ST5							;								|sum fc2 wc co cnt temp sin*fc2
			FDivRP	ST1,ST							;								|sum fc2 wc co cnt sin/temp

			Jmp		short .NotZero

		.Zero:
			FLd1											;h = 1.0						|sum fc2 wc co cnt 1.0

		.NotZero:

		FXch		ST1									;Hamming window			|sum fc2 wc co h cnt
		FMul		ST,ST2								;0.54+0.46*cos(co*cnt)	|sum fc2 wc co h cnt*co
		FCos												;								|sum fc2 wc co h cos(cnt)
		FMul		dword [fp0_46]						;								|sum fc2 wc co h w*0.46
		FAdd		dword [fp0_54]						;								|sum fc2 wc co h w+0.54

		FMulP		ST1,ST								;tap = w * h;				|sum fc2 wc co h*w

		FSt		dword [EBX*4+lowTaps]

		FAddP		ST4,ST								;sum += temp				|sum+tap fc2 wc co

	Inc	EBX
	Cmp	EBX,LOWTAPS
	JB		short .Next

	FStP	ST												;								|sum fc2 wc
	FStP	ST												;								|sum fc2
	FStP	ST												;								|sum

	;Normalize coefficients ------------------
	FMul	dword [fpShR31]
	XOr	EBX,EBX
	.Norm:
		FLd	dword [EBX*4+lowTaps]
		FDiv	ST1
		FIStP	dword [EBX*4+lowTaps]

	Inc	EBX
	Cmp	EBX,LOWTAPS
	JB		short .Norm

	FStP	ST												;								|(empty)

ENDP


;
;Set DSP Options

PROC SetDSPOpt, mixType, numChn, bits, rate, inter, opts
LOCALS fixVol,eraseBuf
USES ALL

	Mov	byte [fixVol],0
	Mov	byte [eraseBuf],0

	;=========================================
	;Verify parameters

	;mixType ---------------------------------
	MovZX	EAX,byte [dspMix]
	Mov	EDX,[mixType]
	Cmp	EDX,-1
	JE		short .DefMix

		Mov	AL,3
		Cmp	EDX,MIX_FLOAT
		JAE	short .DefMix

		Test	byte [procType],CPU_MMX
		SetNZ	AL
		Inc	AL
		Cmp	EDX,MIX_INT
		JAE	short .DefMix

		XOr	EAX,EAX

	.DefMix:
	Mov	[mixType],EAX

	;numChn ----------------------------------
	MovZX	EAX,byte [dspChn]
	Mov	EDX,[numChn]
	Cmp	EDX,-1
	JE		short .DefChn

		XOr	EAX,EAX
		Cmp	EDX,1
		SetNE	AL
		Inc	AL

	.DefChn:
	Mov	[numChn],EAX

	;bits ------------------------------------
	MovZX	EAX,byte [dspSize]
	Mov	EDX,[bits]
	Cmp	EDX,-1
	JE		short .DefBits

		Mov	EAX,EDX
		SAR	EAX,3

		Cmp	EDX,8
		JE		short .DefBits
		Cmp	EDX,16
		JE		short .DefBits
		Cmp	EDX,24
		JE		short .DefBits
		Cmp	EDX,32
		JE		short .DefBits
		Cmp	EDX,-32
		JE		short .DefBits

		Mov	EAX,2

	.DefBits:
	Mov	[bits],EAX

	;rate ------------------------------------
	Mov	EAX,[dspRate]
	Test	byte [dspOpts],DSP_ANALOG
	JZ		short .NoLow1
		Mov	EAX,[realRate]
	.NoLow1:

	Mov	EDX,[rate]
	Cmp	EDX,-1
	JE		short .DefRate

		Mov	EAX,EDX
		XOr	ECX,ECX

		Cmp	EDX,8000
		SetB	CL
		Cmp	EDX,192000
		SetA	CH
		Test	ECX,ECX
		JZ		short .DefRate

		Mov	EAX,32000

	.DefRate:
	Mov	[rate],EAX

	;inter -----------------------------------
	MovZX	EAX,byte [dspInter]
	Mov	EDX,[inter]
	Cmp	EDX,-1
	JE		short .DefInter

		Mov	EAX,EDX
		Cmp	EDX,3
		JBE	.DefInter

		Mov	EAX,3

	.DefInter:
	Mov	[inter],EAX

	;opts ------------------------------------
	MovZX	EAX,byte [dspOpts]
	Mov	EDX,[opts]
	Cmp	EDX,-1
	JE		short .DefOpts

		Mov	EAX,EDX

	.DefOpts:
	Mov	[opts],EAX

	;Verify mix type is capable of producing requested output
	Cmp	byte [mixType],0							;MIX_NONE is unaffected by output settings
	JZ		.MixOK

	Cmp	byte [bits],2								;ifn (bits >= 16-bit && numChn >= stereo &&
	SetAE	AL												;     mixType >= MMX)
	Cmp	byte [numChn],2
	SetAE	AH
	And	AL,AH
	Cmp	byte [mixType],2
	SetAE	AH
	And	AL,AH
	JNZ	short .MixOK

		Mov	byte [mixType],1						;Force mix type to 386

		Cmp	byte [bits],1							;if (bits == 8-bit) mixType = 386
		JE		short .MixOK

		Cmp	byte [numChn],1						;if (numChn == mono) mixType = 386; bits = 16-bit
		JA		short .Stereo
			Mov	byte [bits],2
			Jmp	short .MixOK
		.Stereo:

		Cmp	byte [bits],2							;if (bits != 16-bit) mixType = float
		JE		short .MixOK
			Mov	byte [mixType],3

	.MixOK:


	;=========================================
	;Options

	Mov	DL,[opts]

	;Select ADPCM routine --------------------
	Mov	dword [pDecomp],UnpckSrcOld
	Test	DL,DSP_OLDSMP
	JNZ	short .OldSmp
		Mov	dword [pDecomp],UnpckSrc
	.OldSmp:

	;Pseudo surround sound -------------------
	Mov	AL,[surround]
	XOr	AL,DL
	And	AL,DSP_SURND
	Or		[fixVol],AL									;Did surround sound flag change?

	Test	DL,DSP_SURND
	SetZ	AL
	Dec	AL
	Mov	[surround],AL

	;Reverse stereo --------------------------
	Mov	AL,[dspOpts]
	XOr	AL,DL
	And	AL,DSP_REVERSE
	Or		[fixVol],AL									;Did reverse flag change?

	;Disable echo ----------------------------
	Test	DL,DSP_NOECHO
	SetZ	AL
	Dec	AL
	And	AL,DSP_NOECHO
	And	byte [disEcho],~DSP_NOECHO
	Or		[disEcho],AL

	Mov	[dspOpts],DL								;Save option flags

	Call	SetDSPAmp,[volAmp]


	;=========================================
	;Interpolation method

	MovZX	EAX,byte [inter]							;Save interpolation type
	Mov	[dspInter],AL

	MovZX	EDX,byte [mixType]						;if (mixType != MIX_NONE)
	Test	EDX,EDX
	JZ		short .NoMix

		Dec	EDX
		ShL	EDX,2
		Or		EDX,EAX
		Mov	EAX,[EDX*4+intRout]
		Mov	[pInter],EAX

	.NoMix:


	;=========================================
	;Calculate sample rate change

	Mov	EAX,[rate]
	Mov	[realRate],EAX

	Test	byte [dspOpts],DSP_ANALOG
	JZ		short .NoLow
	And	byte [dspOpts],~DSP_ANALOG				;Clear flag, incase filter can't be implemented

	Cmp	byte [mixType],2							;if (mixType == MMX && rate >= 32000 &&
	JNE	short .NoLow								;    (opts & OPT_LOW))
	Cmp	dword [rate],32000
	JB		short .NoLow
		XOr	EDX,EDX
		Mov	EAX,32000 << 16
		Mov	[lowDec],EDX
		Div	dword [rate]
		Mov	[lowRate],EAX

		Mov	dword [rate],32000
		Or		byte [dspOpts],DSP_ANALOG

		Call	AACoeffs
	.NoLow:

	Mov	EAX,[rate]
	Cmp	EAX,[dspRate]								;Has sample rate changed?
	JE 	.SameRate									;  No
		Mov	[dspRate],EAX							;  Yes, Adjust a lot of items

		;Calculate amount to adjust SPC pitch values
		XOr	EDX,EDX									;EDX:EAX = Base pitch << 20
		Mov	EAX,[pitchBas]
		ShLD	EDX,EAX,20
		ShL	EAX,20

		Div	dword [dspRate]
		Mov	[pitchAdj],EAX

		;Calculate update rate for envelopes and noise
		Mov	ESI,freqTab
		Mov	EDI,rateTab
		Mov	EBX,32000

		Mov	ECX,31
		.CalcRT:
			Mov	EAX,[ECX*4+ESI]
			ShL	EAX,16
			Mul	dword [dspRate]
			Div	EBX

			Cmp	EAX,10000h
			JAE	short .RTOK
				Mov	EAX,10000h
			.RTOK:

			Mov	[ECX*4+EDI],EAX

		Dec	ECX
		JNZ	short .CalcRT
		Mov	[EDI],ECX

		;Volume ramping rate ------------------
		Mov	dword [ESP-4],32000
		FILd	dword [ESP-4]
		FIDiv	dword [dspRate]
		FMul	dword [fpShR11]
		FStP	dword [volRamp]

		;Reset FIR info -----------------------
		XOr	EAX,EAX
		Mov	[firDec],EAX
		Mov	[firCur],EAX
		Mov	[lowDec],EAX

		Mov	EAX,[dspRate]
		MovZX	EDX,word [2+dspRate]
		ShL	EAX,16
		Mov	ECX,32000
		Div	ECX
		Mov	[firRate],EAX							;firRate = (dspRate<<16) / 32kHz

		;Adjust voice rates -------------------
		Mov	ESI,7*16									;Adjust the current rates in each voice incase the
		Mov	EBX,7*80h								;  sample rate is being changed during emulation
		.Voice:
			MovZX	EAX,word [ESI+dsp+pitch]		;Set pitch
			Mul	dword [pitchAdj]
			ShRD	EAX,EDX,16
			Mov	[EBX+mix+mRate],EAX

			MovZX	EDI,byte [EBX+mix+eRIdx]		;Set envelope adjustment
			Mov	EAX,[EDI*4+rateTab]
			Mov	[EBX+mix+eRate],EAX
			Mov	[EBX+mix+eCnt],EAX

			Sub	ESI,16
		Add	EBX,-80h
		JNS	short .Voice

		;Adjust echo delay --------------------
		MovZX	EAX,byte [dsp+edl]
		ShL	AL,4
		Mul	dword [dspRate]
		Mov	ECX,1000
		Div	ECX
		Test	EAX,EAX
		SetZ	CL
		Or		AL,CL
		ShL	EAX,3
		Mov	[echoDel],EAX

		Mov	byte [eraseBuf],1
	.SameRate:


	;=========================================
	;Set sample size

	Mov	AL,[bits]
	Cmp	AL,[dspSize]								;If the sample size has changed, CL = 1
	JE		short .SameBits
		Mov	[dspSize],AL
	.SameBits:


	;=========================================
	;Set number of channels

	Mov	AL,[numChn]
	Cmp	AL,[dspChn]									;If the number of channels has changed, CL = 1
	SetNE	CL
	Or		[fixVol],CL
	Mov	[dspChn],AL


	;=========================================
	;Update areas affected by the mix type

	Mov	AL,[mixType]
	Cmp	AL,[dspMix]
	JE 	.SameMix

		Mov	[dspMix],AL

		Mov	byte [fixVol],1						;Force volumes to be recalculated
		Mov	byte [eraseBuf],1
	.SameMix:


	;=========================================
	;Erase sample buffers

	Test	byte [eraseBuf],-1
	JZ		short .NoErase

		XOr	EAX,EAX

		Mov	EDI,echoBuf
		Mov	ECX,ECHOBUF
		Rep	StoSD

		Mov	EDI,firBuf
		Mov	ECX,FIRBUF
		Rep	StoSD

		Mov	EDI,lowBuf
		Mov	ECX,LOWBUF
		Rep	StoSD

		Mov	[vMMaxL],EAX
		Mov	[vMMaxR],EAX
	.NoErase:


	;=========================================
	;Fixup volume handlers

	Test	byte [fixVol],-1
	JZ		.Done

		;Setup DSP register jump table --------
		Mov	ESI,dspRegsF							;ESI -> floating point register handlers
		Cmp	byte [dspMix],3
		JE		.SetRegs
		Mov	ESI,dspRegsM							;ESI -> mono integer register handlers
		Cmp	byte [dspChn],1
		JE		.SetRegs
		Mov	ESI,dspRegsI							;ESI -> stereo integer register handlers

		.SetRegs:
		Mov	EDI,dspRegs

		Mov	EAX,[ESI]								;EAX -> VOLL handler
		Mov	EDX,[4+ESI]								;EDX -> VOLR handler
		Mov	ECX,[28+ESI]							;ECX -> FC handler

		Test	byte [dspOpts],DSP_REVERSE			;If reversed, swap pointers to left and right volume
		JZ		short .NoRev1
			XChg	EAX,EDX
		.NoRev1:

		Mov	EBX,70h
		.CopyVol:
			Mov	[volL*4+EBX*4+EDI],EAX
			Mov	[volR*4+EBX*4+EDI],EDX
			Mov	[fc*4+EBX*4+EDI],ECX
		Sub	BL,10h
		JNS	short .CopyVol

		Mov	EAX,[8+ESI]								;EAX -> MVOLL handler
		Mov	EDX,[12+ESI]							;EDX -> MVOLR handler
		Mov	EBX,[16+ESI]							;EBX -> EVOLL handler
		Mov	ECX,[20+ESI]							;ECX -> EVOLR handler

		Test	byte [dspOpts],DSP_REVERSE
		JZ		short .NoRev2
			XChg	EAX,EDX
			XChg	EBX,ECX
		.NoRev2:

		Mov	[mvolL*4+EDI],EAX
		Mov	[mvolR*4+EDI],EDX
		Mov	[evolL*4+EDI],EBX
		Mov	[evolR*4+EDI],ECX

		Mov	EAX,[24+ESI]							;EAX -> EFB handler
		Mov	[efb*4+EDI],EAX

		;Reinitialize registers ---------------
		XOr	EDX,EDX
		Mov	ECX,70h
		.NextVoice:
			LEA	EBX,[ECX+volL]
			Mov	AL,[ECX+dsp+volL]
			Call	InitReg
			Mov	EAX,[ECX*8+mix+mTgtL]
			Mov	[ECX*8+mix+mChnL],EAX

			LEA	EBX,[ECX+volR]
			Mov	AL,[ECX+dsp+volR]
			Call	InitReg
			Mov	EAX,[ECX*8+mix+mTgtR]
			Mov	[ECX*8+mix+mChnR],EAX

			LEA	EBX,[ECX+fc]
			Mov	AL,[ECX+dsp+fc]
			Call	InitReg

		Sub	CL,10h
		JNC	short .NextVoice

		Mov	EBX,mvolL
		Mov	AL,[dsp+mvolL]
		Call	InitReg

		Mov	EBX,mvolR
		Mov	AL,[dsp+mvolR]
		Call	InitReg

		Mov	EBX,evolL
		Mov	AL,[dsp+evolL]
		Call	InitReg

		Mov	EBX,evolR
		Mov	AL,[dsp+evolR]
		Call	InitReg

		Mov	EBX,efb
		Mov	AL,[dsp+efb]
		Call	InitReg

	.Done:

ENDP


;
;Debug DSP

PROC SetDSPDbg, pTraceFunc
USES	EDX

	Mov	EDX,[pTrace]

	Mov	EAX,[pTraceFunc]
	Cmp	EAX,-1
	JE		short .NoFunc
		Mov	[pTrace],EAX
	.NoFunc:
	Mov	EAX,EDX

ENDP


;
;Fix DSP After Loading Saved State

PROC FixDSP
USES ALL

	;Enable voices currently keyed on --------
	Mov	byte [voiceMix],0

	Mov	EBX,kon
	Mov	AL,[dsp+kon]
	Call	InitReg

	;Setup global paramaters -----------------
	Mov	EBX,mvolL
	Mov	AL,[dsp+mvolL]
	Call	InitReg

	Mov	EBX,mvolR
	Mov	AL,[dsp+mvolR]
	Call	InitReg

	Mov	EBX,evolL
	Mov	AL,[dsp+evolL]
	Call	InitReg

	Mov	EBX,evolR
	Mov	AL,[dsp+evolR]
	Call	InitReg

	Mov	EBX,flg
	Mov	AL,[dsp+flg]
	Call	InitReg

	Mov	EBX,efb
	Mov	AL,[dsp+efb]
	Call	InitReg

	Mov	EBX,edl
	Mov	AL,[dsp+edl]
	Call	InitReg

	Mov	ECX,70h
	.NextTap:
		LEA	EBX,[ECX+fc]
		Mov	AL,[ECX+dsp+fc]
		Call	InitReg
	Sub	CL,10h
	JNC	short .NextTap

ENDP


;
;Fix DSP After Seeking

PROC FixSeek, reset
USES ECX,EDI

	Mov	AL,[reset]
	Test	AL,AL
	JZ		.NoReset
		;Turn off all voices ------------------
		Mov	AL,[dsp+kon]							;Mark all playing voices as ended
		Mov	[dsp+endx],AL
		XOr	EAX,EAX
		Mov	[dsp+kon],AL							;Reset key registers
		Mov	[dsp+kof],AL
		Mov	[voiceMix],AL

		Mov	CL,8
		Mov	EDI,mix
		.ResetMix:
			Mov	[EDI+eVal],EAX
			Mov	[EDI+mOut],EAX
			Or 	byte [EDI+mFlg],MFLG_OFF
			Sub	EDI,-80h
		Dec	CL
		JNZ	short .ResetMix

		Mov	CL,8
		Mov	EDI,dsp
		.ResetDSP:
			Mov	[EDI+envx],AL
			Mov	[EDI+outx],AL
			Add	EDI,10h
		Dec	CL
		JNZ	short .ResetDSP
	.NoReset:

	;Erase echo region -----------------------
	XOr	EAX,EAX
	Mov	EDI,echoBuf
	Mov	ECX,ECHOBUF
	Rep	StoSD

	;Erase filter memory ---------------------
	Mov	EDI,firBuf
	Mov	ECX,FIRBUF
	Rep	StoSD

	Mov	EDI,lowBuf
	Mov	ECX,LOWBUF
	Rep	StoSD

	Call	SetFade

ENDP


;
;Save DSP's Current State of Operation

PROC SaveDSP, pStateS
USES ALL

	;DSP Registers ---------------------------
	Mov	EBX,[pStateS]
	Mov	EDI,[EBX+pDSP]
	Test	EDI,EDI
	JZ		.NoRegs
		Mov	ESI,dsp
		Mov	ECX,128/4
		Rep	MovSD
	.NoRegs:

	;Internal Voice Structures ---------------
	Mov	EDI,[EBX+pMix]
	Test	EDI,EDI
	JZ		.NoMix
		Mov	ESI,mix									;Copy structure array
		Mov	ECX,8*128/4
		Rep	MovSD

		Mov	EDI,[EBX+pMix]
		Mov	EDX,[pAPURAM]
		Mov	ECX,80h*7
		.NextVoice:
			Sub	[ECX+EDI+dspv],EDX				;Remove physical address from pointers
			Sub	[ECX+EDI+bCur],EDX

			LEA	EAX,[ECX+mix+sBuf]				;Convert sample pointer into index
			Sub	[ECX+EDI+sIdx],EAX
			ShR	dword [ECX+EDI+sIdx],1

		Add	ECX,-80h
		JNS	.NextVoice
	.NoMix:

	;Echo region -----------------------------
	Mov	EDI,[EBX+pEchoBuf]
	Test	EDI,EDI
	JZ		.NoEcho
		Mov	ESI,[echoCur]
		Add	ESI,echoBuf+8
		Mov	EAX,240*8								;240ms * stereo
		Mov	ECX,1000									;Convert seconds to milliseconds
		Mul	dword [dspRate]
		Div	ECX
		Mov	ECX,EAX
		Sub	ECX,[echoCur]
		Sub	ECX,8
		ShR	ECX,2
		Rep	MovSD

		Mov	ESI,echoBuf
		Mov	ECX,[echoCur]
		Add	ECX,8
		ShR	ECX,2
		Rep	MovSD
	.NoEcho:

	;Volume ----------------------------------
	Mov	EAX,[vMMaxL]
	Mov	[EBX+vMMaxLS],EAX

	Mov	EAX,[vMMaxR]
	Mov	[EBX+vMMaxRS],EAX

	Mov	AL,[voiceMix]
	Mov	[EBX+vActive],AL

	Mov	EAX,[volAmp]
	Mov	[EBX+mAmp],EAX

	XOr	EAX,EAX
	Mov	[EBX+_rd1],EAX
	Mov	[EBX+_rd2],AL
	Mov	[EBX+_rd2+1],AX

ENDP


;
;Restore DSP's State of Operation

PROC RestoreDSP, pStateL
USES ALL

	Mov	EBX,[pStateL]

	;DSP Registers ---------------------------
	Mov	ESI,[EBX+pDSP]
	Test	ESI,ESI
	JZ		.NoRegs
		Mov	EDI,dsp
		Mov	ECX,128/4
		Rep	MovSD
	.NoRegs:

	;Internal Voice Structures ---------------
	Mov	ESI,[EBX+pMix]
	Test	ESI,ESI
	JZ		.NoMix
		Mov	EDI,mix									;Copy structure array
		Mov	ECX,8*128/4
		Rep	MovSD

		Mov	ESI,10h*7
		.ChkVoice:
			Mov	EAX,[pAPURAM]						;Restore physical address to pointers
			Add	[ESI*8+mix+dspv],EAX
			Add	[ESI*8+mix+bCur],EAX

			And	dword [ESI*8+mix+sIdx],0Fh		;Convert sample index into pointer
			ShL	dword [ESI*8+mix+sIdx],1
			LEA	EAX,[ESI*8+mix+sBuf]
			Add	dword [ESI*8+mix+sIdx],EAX

			Mov	EAX,[ESI*8+mix+eRate]
			Mov	EAX,[ESI*8+mix+eCnt]
			Cmp	EAX,-1								;-1 if caller wants values reset
			JNE	.EnvOK
				MovZX	EAX,byte [ESI*8+mix+eRIdx]
				Mov	EAX,[EAX*4+rateTab]
				Mov	[ESI*8+mix+eRate],EAX
				Mov	[ESI*8+mix+eCnt],EAX
				ShL	dword [ESI*8+mix+eVal],E_SHIFT
			.EnvOK:

			Cmp	dword [ESI*8+mix+mRate],-1		;-1 if caller wants values reset
			JNE	.MixOK
				MovSX	EAX,byte [ESI+dsp+outx]
				ShL	EAX,8
				Mov	[ESI*8+mix+mOut],EAX
				Mov	dword [ESI*8+mix+mDec],0

				LEA	EBX,[ESI+pitch]
				Mov	AL,byte [ESI+dsp+pitch]
				Call	InitReg

				LEA	EBX,[ESI+volL]
				Mov	AL,[ESI+dsp+volL]
				Call	InitReg

				LEA	EBX,[ESI+volR]
				Mov	AL,[ESI+dsp+volR]
				Call	InitReg
			.MixOK:

		Sub	ESI,10h
		JNS	.ChkVoice

		Mov	EBX,[pStateL]
	.NoMix:

	;Echo region -----------------------------
	Mov	EAX,240*2									;240ms * stereo
	Mov	ECX,1000										;Convert seconds to milliseconds
	Mul	dword [dspRate]
	Div	ECX
	Mov	ECX,EAX
	ShL	EAX,2
	Sub	EAX,8
	Mov	[echoCur],EAX

	Mov	ESI,[EBX+pEchoBuf]
	Test	ESI,ESI
	JZ		.NoEcho
		Mov	EDI,echoBuf
		Rep	MovSD
		Jmp	.HasEcho
	.NoEcho:
		XOr	EAX,EAX
		Mov	EDI,echoBuf
		Mov	ECX,ECHOBUF
		Rep	StoSD
	.HasEcho:

	;Volume ----------------------------------
	Mov	EAX,[EBX+vMMaxLS]
	Mov	[vMMaxL],EAX

	Mov	EAX,[EBX+vMMaxRS]
	Mov	[vMMaxR],EAX

	Mov	AL,[EBX+vActive]
	Mov	[voiceMix],AL

	Call	SetDSPAmp, dword [EBX+mAmp]			;SetDSPAmp will reset main and echo volumes

	;Setup global paramaters -----------------
	Mov	EBX,flg
	Mov	AL,[dsp+flg]
	Call	InitReg

	Mov	EBX,efb
	Mov	AL,[dsp+efb]
	Call	InitReg

	Mov	EBX,edl
	Mov	AL,[dsp+edl]
	Call	InitReg

	Mov	ECX,70h
	.NextTap:
		LEA	EBX,[ECX+fc]
		Mov	AL,[ECX+dsp+fc]
		Call	InitReg
	Sub	ECX,10h
	JNC	short .NextTap

	Mov	EAX,[t64Cnt]
	ShR	EAX,1
	Mov	[outCnt],EAX

	XOr	EAX,EAX
	Mov	[outDec],EAX
	Mov	[konWr],EAX

	;Restore FIR ring buffer -----------------
	Mov	ESI,[echoDel]
	Add	ESI,echoBuf-FIRBUF*2
	Mov	EDI,firBuf
	Mov	ECX,FIRBUF/2
	Rep	MovSD
	Sub	ESI,FIRBUF*2
	Mov	ECX,FIRBUF/2
	Rep	MovSD
	Mov	dword [firCur],0

ENDP


;
;DSP Pitch Adjustment

PROC SetDSPPitch, base
USES EDX,EBX,ESI

	;Calculate amount to adjust SPC pitch values
	XOr	EDX,EDX
	Mov	EAX,[base]
	Mov	[pitchBas],EAX
	ShLD	EDX,EAX,20
	ShL	EAX,20

	Div	dword [dspRate]
	Mov	[pitchAdj],EAX

	;Adjust voice rates to new pitch ---------
	Mov	ESI,7*16										;Adjust the current rates in each voice incase the
	Mov	EBX,7*80h									;  sample rate is being changed during emulation
	.Voice:
		Mov	EAX,[EBX+mix+mOrgP]					;Set pitch
		Mul	dword [pitchAdj]
		ShRD	EAX,EDX,16
		Mov	[EBX+mix+mRate],EAX

		Sub	ESI,16
	Add	EBX,-80h
	JNS	short .Voice

ENDP


;
;DSP Amplification

PROC SetDSPAmp, amp
USES ECX,EDX,EBX

	Mov	EAX,[amp]									;if amp<0 amp=0
	CDQ
	Not	EDX
	And	EAX,EDX

	Cmp	EAX,256
	JA		short .NewScale
		ShL	EAX,12
	.NewScale:

	Mov	[volAmp],EAX

	;Multiply by volume ----------------------
	Test	byte [dspOpts],DSP_ANALOG
	JZ		short .NoAmp
		ShR	EAX,8
		Neg	EAX
		Mov	[lowAmp],EAX

		Mov	EAX,10000h
	.NoAmp:

	Mul	dword [volAtten]
	ShRD	EAX,EDX,16
	Mov	[volAdj],EAX

	;Update global volumes -------------------
	Mov	EBX,mvolL
	Mov	AL,[dsp+mvolL]
	Call	InitReg

	Mov	EBX,mvolR
	Mov	AL,[dsp+mvolR]
	Call	InitReg

	Mov	EBX,evolL
	Mov	AL,[dsp+evolL]
	Call	InitReg

	Mov	EBX,evolR
	Mov	AL,[dsp+evolR]
	Call	InitReg

ENDP


;
;DSP Fade Volume

PROC SetDSPVol, vol
USES ECX,EDX,EBX

	Mov	EAX,[vol]									;if EAX<0 EAX=0
	CDQ
	Not	EDX
	And	EAX,EDX
	Mov	[volAtten],EAX

	Test	byte [dspOpts],DSP_ANALOG
	JNZ	short .NoAmp
		Mul	dword [volAmp]
		ShRD	EAX,EDX,16
	.NoAmp:

	Mov	[volAdj],EAX

	;Update global volumes -------------------
	Mov	EBX,mvolL
	Mov	AL,[dsp+mvolL]
	Call	InitReg

	Mov	EBX,mvolR
	Mov	AL,[dsp+mvolR]
	Call	InitReg

	Mov	EBX,evolL
	Mov	AL,[dsp+evolL]
	Call	InitReg

	Mov	EBX,evolR
	Mov	AL,[dsp+evolR]
	Call	InitReg

ENDP


;
;Set Song Length

PROC SetDSPLength, song, fade
USES EDX

	Mov	EDX,[fade]
	XOr	EAX,EAX										;if (fadeLen == 0) fadeLen = 1;
	Test	EDX,EDX										;0 will cause a division error
	SetZ	AL
	Or		EDX,EAX
	Mov	[fadeLen],EDX

	Mov	EAX,[song]
	Add	EDX,EAX
	Mov	[songLen],EAX

	Cmp	EAX,[t64Cnt]								;if (t64Cnt > songLen)
	JB		short .SetFade
		Call	SetDSPVol,10000h
		RetN	EDX

	.SetFade:
		Call	SetFade									;If song is in fade mode, set fade volume
		Mov	EAX,EDX

ENDP


;
;Set Fade Volume
;
;Calls SetDSPVol to fade the song out based on t64Cnt, songLen, and fadeLen.

PROC SetFade

	Mov	EDX,[t64Cnt]								;EAX = T64Cnt - songLen;
	Sub	EDX,[songLen]
	JBE	.Done

	XOr	EAX,EAX										;if (EAX > fadeLen) EAX = fadeLen;
	Cmp	EDX,[fadeLen]
	SetA	AL
	Dec	EAX
	And	EDX,EAX
	Not	EAX
	And	EAX,[fadeLen]
	Or		EDX,EAX

	XOr	EAX,EAX										;EDX = 65536 - ((EDX << 16) / fadeLen);
	ShRD	EAX,EDX,16
	ShR	EDX,16
	Div	dword [fadeLen]
	Mov	EDX,65536
	Sub	EDX,EAX

	Call	SetDSPVol,EDX								;SetDSPVol(EDX);

	.Done:

ENDP


;
;Adjust Voice Volume for Stereo Separation
;
;A big nasty function to adjust the left and right channel volumes for stereo separation control
;
;In:
;   EBX = Indexes current voice

%if STEREO
PROC ChnSep
RVolL:
RVolR:
USES ECX,EBX

	ShR	EBX,3
	MovSX	EAX,byte [EBX+dsp+volL]					;If both volumes are 0, an exception will be generated
	MovSX	EDX,byte [EBX+dsp+volR]

	LEA	EBX,[EBX*8+mix]
	Mov	[EBX+mChnL],EAX							;Save the sign of each channel
	Mov	[EBX+mTgtL],EAX
	Mov	[EBX+mChnR],EDX
	Mov	[EBX+mTgtR],EDX

	Cmp	EAX,EDX
	JE		.Done

	Mov	ECX,[volSepar]								;Don't waste time if no separation is applied
	Test	ECX,ECX
	JZ		.Done

	FInit

	SAR	EAX,31
	SAR	EDX,31

	;Convert left/right into vol/pan ---------
	;        _____________
	;       |    2      2
	;       |  L      R
	; V = _ | ---  + ---
	;      \| 128    128
	;
	;        2
	;      R
	; P = ---  - 0+5
	;      V
	;
	; V is 0.0 to 1.414 (-inf to +3 dB)
	; P is -0.5 to 0.5 (left to right)

	FILd	dword [EBX+mChnR]							;V = sqrt((L/128)^2+(R/128)^2)	|chnR
	FMul	dword [fpShR7]								;Change vol to float					|chnR/128=R
	FLd	ST												;											|R R
	FMul	ST,ST											;											|R R^2
	FILd	dword [EBX+mChnL]							;											|R R chnL
	FMul	dword [fpShR7]								;											|R R chnL/128=L
	FMul	ST,ST											;											|R R L^2
	FAddP	ST1,ST										;											|R R+L=V
	FSqrt													;											|R sqrt(V)
	FXch													;P = (R/V)^2 - 0+5					|V R
	FAbs													;											|V |R|
	FDiv	ST,ST1										;											|V R/V
	FMul	ST,ST											;											|V R^2
	FSub	dword [fp0_5]								;											|V R-0+5=P

	;Adjust panning --------------------------
	; S is -1.0 to 0 (pan toward center):
	;
	; P = P + (P * S)
	;
	;
	; S is 0 to 1+0 (pan toward outside):
	;
	; Left
	; P = P + ((0.5 + P) * S)
	;
	; Right
	; P = P + ((0.5 - P) * S)

	FLd	ST												;											|V P D
	Test	byte [3+volSepar],80h					;Is panning adjusted toward center?
	JNZ	short .ToCenter
		FSt	qword [ESP-8]
		FLd	dword [fp0_5]							;											|V P D 0.5
		Test	byte [ESP-1],80h						;Is pan on left side?
		JZ		short .Right
			FChS											;Negate 0.5								|V P D -0.5
		.Right:
		FSubRP	ST1,ST								;Get distance from side				|V P 0.5-D
	.ToCenter:
	FMul	dword [volSepar]							;Get fraction of distance			|V P D*volSepar
	FAddP	ST1,ST										;Add fraction to pan					|V P+D
	FLd	ST												;Copy new panning value				|V P P

	;Convert vol/pan back into left/right ----
	;           _________
	; L = V * \| 0.5 - P  * 128
	;           _________
	; R = V * \| 0.5 + P  * 128

	FAdd	dword [fp0_5]								;R = V*sqrt(+5+P)*128				|V P P+0.5
	FSqrt													;											|V P sqrt(P)
	FMul	ST,ST2										;											|V P P*V=R
	FMul	dword [fp128]								;											|V P R*128
	FIStP	dword [EBX+mChnR]							;											|V P
	XOr	[EBX+mChnR],EDX							;Set the correct sign
	Sub	[EBX+mChnR],EDX

	FSubR	dword [fp0_5]								;L = V*sqrt(+5-P)*128				|V 0.5-P
	FSqrt													;											|V sqrt(P)
	FMulP	ST1,ST										;											|V*P=L
	FMul	dword [fp128]								;											|L*128
	FIStP	dword [EBX+mChnL]							;											|(empty)
	XOr	[EBX+mChnL],EAX
	Sub	[EBX+mChnL],EAX

	Mov	EDX,[EBX+mChnR]
	Mov	EAX,[EBX+mChnL]
	Mov	[EBX+mTgtR],EDX
	Mov	[EBX+mTgtL],EAX

.Done:
	XOr	EAX,EAX

ENDP

;Channel separator for floating-point routines
PROC ChnSepF
RVolLF:
RVolRF:
USES ECX,EBX

	ShR	EBX,3
	MovSX	EAX,byte [EBX+dsp+volL]
	MovSX	EDX,byte [EBX+dsp+volR]

	LEA	EBX,[EBX*8+mix]
	Mov	[EBX+mTgtL],EAX
	Mov	[EBX+mTgtR],EDX

	Cmp	EAX,EDX
	JE		.NoSep

	Mov	ECX,[volSepar]
	Test	ECX,ECX
	JZ		.NoSep

	FInit

	And	AL,80h										;Save sign bit of each volume
	And	DL,80h
	ShL	EAX,24
	ShL	EDX,24

	;Convert left/right into vol/pan ---------
	FILd	dword [EBX+mTgtR]
	FMul	dword [fpShR7]
	FLd	ST
	FMul	ST,ST
	FILd	dword [EBX+mTgtL]
	FMul	dword [fpShR7]
	FMul	ST,ST
	FAddP	ST1,ST
	FSqrt
	FXch
	FAbs
	FDiv	ST,ST1
	FMul	ST,ST
	FSub	dword [fp0_5]

	;Adjust panning --------------------------
	FLd	ST
	Test	byte [3+volSepar],80h
	JNZ	short .ToCenterF
		FSt	qword [ESP-8]
		FLd	dword [fp0_5]
		Test	byte [ESP-1],80h
		JZ		short .RightF
			FChS
		.RightF:
		FSubRP	ST1,ST
	.ToCenterF:
	FMul	dword [volSepar]
	FAddP	ST1,ST
	FLd	ST

	;Convert vol/pan back into left/right ----
	FAdd	dword [fp0_5]
	FSqrt
	FMul	ST,ST2
	FStP	dword [EBX+mTgtR]
	Or		[EBX+mTgtR],EDX

	FSubR	dword [fp0_5]
	FSqrt
	FMulP	ST1,ST
	FStP	dword [EBX+mTgtL]
	Or		[EBX+mTgtL],EAX

	XOr	EAX,EAX
	RetN

.NoSep:
	FILd	dword [EBX+mTgtL]
	FMul	dword [fpShR7]
	FStP	dword [EBX+mTgtL]

	FILd	dword [EBX+mTgtR]
	FMul	dword [fpShR7]
	FStP	dword [EBX+mTgtR]

	XOr	EAX,EAX

ENDP
%endif


;
;Set Stereo Separation

PROC SetDSPStereo, sep
USES EDX,EBX

%if STEREO
	Sub	dword [sep],32768							;Convert fixed point unsigned value to signed float
	FILd	dword [sep]
	FMul	dword [fpShR15]
	FStP	dword [volSepar]

	;Update each voice with new separation ---
	Mov	EBX,7*80h
	Cmp	byte [dspMix],3
	JE		.Float

	.Voice:
		Call	ChnSep
	Add	EBX,-80h
	JNS	short .Voice
	RetN

	.Float:
		Call	ChnSepF
		Mov	EAX,[EBX+mix+mTgtL]
		Mov	EDX,[EBX+mix+mTgtR]
		Mov	[EBX+mix+mChnL],EAX
		Mov	[EBX+mix+mChnR],EDX
	Add	EBX,-80h
	JNS	short .Float
%endif

ENDP


;
;Set Echo Stereo Separation

PROC SetDSPEFBCT, leak
USES EDX,EBX

	Mov	EAX,[leak]
	Add	EAX,32768									;Unsign crosstalk
	Mov	[efbct],EAX

	;Update echo feedback --------------------
	Mov	EBX,efb
	Mov	AL,[dsp+efb]
	Call	InitReg

ENDP


;
;Start Sound Source Decompression
;
;Called when a voice is keyed on to set up the internal data for waveform mixing and decompress the
;first block.
;
;In:
;   EBX-> mix[voice]
;   ESI-> dsp.voice[voice]
;
;Out:
;   nothing
;
;Destroys:
;   EAX,EDX

ALIGN 16
PROC StartSrc

	Push	ESI,EDI,EBP

	MovZX	EAX,byte [ESI+srcn]
	Mov	EDX,[pAPURAM]								;EDX -> Source directory
	Or		DH,[dsp+dir]
	LEA	ESI,[EAX*4+EDX]							;ESI -> Entry in directory

	Mov	SI,[ESI]										;ESI -> First block of waveform

	LEA	EDI,[EBX+sBuf]								;EDI -> Uncompressed sample buffer
	Mov	[EBX+bCur],ESI								;Save physical pointers to wave data
	Mov	[EBX+sIdx],EDI

	;Decompress first block ------------------
	Mov	AL,[ESI]
	Push	EBX
	Mov	[EBX+bHdr],AL								;Save block header
	Mov	EDX,[EBX+sP1]
	Mov	EBX,[EBX+sP2]
	Call	[pDecomp]
	Mov	EAX,EBX
	Pop	EBX
	Mov	[EBX+sP1],EDX
	Mov	[EBX+sP2],EAX

	;Initialize interpolation ----------------
	XOr	EAX,EAX
	Mov	[EBX+sBuf-4],EAX

	Cmp	byte [dspInter],2							;Is interpolation enabled?
	JB		short .NoInter
		Add	byte [EBX+sIdx],6						;Update sample index
	.NoInter:

	Pop	EBP,EDI,ESI

ENDP


;
;Start Envelope
;
;Called when a voice is keyed on to set up the internal data to begin envelope modification based on
;the values in ADSR/Gain.
;
;In:
;   EBX-> mix[voice]
;   ESI-> dsp.voice[voice]
;
;Out:
;   mix.e???       = correct values for envelope routine in mixer
;   dsp.voice.envx = 0
;
;Destroys:
;   EAX,EDX

PROC StartEnv
USES ESI

	XOr	EAX,EAX
	Mov	[EBX+eVal],EAX								;Envelope starts at 0
	Mov	[ESI+envx],AL								;Reset envelope height
	Mov	byte [EBX+eMode],E_ATT << 4			;If envelope gets switched out of gain mode, start ADSR

	Test	byte [ESI+adsr],80h						;Is the envelope in ADSR mode?
	JZ 	ChgGain										;  No, It's in gain mode

ChgAtt:
		Mov	AL,byte [ESI+adsr]
		And	AL,0Fh
		ShL	AL,1										;Adjust AL to index rateTab
		Inc	AL
		Mov	[EBX+eRIdx],AL
		Mov	EDX,[EAX*4+rateTab]					;EDX = Rate of adjustment
		Mov	[EBX+eRate],EDX
		Mov	[EBX+eCnt],EDX
		Mov	byte [EBX+eMode],E_ATT				;Set envelope mode to attack
		Cmp	AL,1Fh									;Is there an attack?
		JE 	.NoAtt

		Mov	dword [EBX+eAdj],A_LIN				;Set adjustment rate to linear
		Mov	dword [EBX+eDest],D_ATTACK			;Set destination to 63/64ths
		RetN												;Exit

	.NoAtt:
		Mov	dword [EBX+eAdj],A_NOATT			;Set adjustment rate to 1/2
		Mov	dword [EBX+eDest],D_MAX				;Set destination to 1.0
		RetN

ALIGN	16
ChgDec:
		Mov	dword [EBX+eAdj],A_EXP				;Set adjustment rate to exponential

		Mov	AL,[ESI+adsr+1]
		ShR	AL,5
		Inc	AL
		IMul	EAX,D_EXP
		Mov	[EBX+eDest],EAX						;Set destination to AL/8

		MovZX	EAX,byte [ESI+adsr]
		And	AL,70h
		ShR	AL,3
		Add	AL,10h									;Adjust AL to index rateTab
		Mov	[EBX+eRIdx],AL
		Mov	EDX,[EAX*4+rateTab]

		Mov	[EBX+eRate],EDX						;Set rate of adjustment
		Mov	[EBX+eCnt],EDX
		Mov	byte [EBX+eMode],E_DECAY			;Set envelope mode to decay
		RetN												;Exit

ALIGN	16
ChgSus:
		Mov	AL,[ESI+adsr+1]
		And	AL,1Fh
		Mov	[EBX+eRIdx],AL
		Mov	EDX,[EAX*4+rateTab]
		SetZ	AL
		ShL	AL,7										;E_IDLE
		Mov	[EBX+eRate],EDX
		Mov	[EBX+eCnt],EDX
		Mov	dword [EBX+eDest],D_MIN				;Set destination to 0
		Or		AL,E_SUST
		Mov	[EBX+eMode],AL							;Set envelope mode to sustain
		RetN												;Exit

ALIGN	16
ChgGain:
	.SetGain:
		Mov	AL,[ESI+gain]
		Test	AL,80h									;Is gain direct?
		JNZ	short .GainMode						;  No, Program envelope
			And	AL,7Fh								;Isolate direct value
			ShL	EAX,E_SHIFT							;Adjust value for internal precision
			Mov	[EBX+eVal],EAX

			Mov	AL,[EBX+eMode]
			And	AL,70h
			Or		AL,E_DIRECT
			Mov	[EBX+eMode],AL						;Set envelope mode to direct
			RetN

ALIGN	16
		.GainMode:
			Mov	DL,AL
			And	AL,1Fh								;Is index zero?
			Mov	[EBX+eRIdx],AL
			Mov	ESI,[EAX*4+rateTab]
			Mov	[EBX+eRate],ESI					;Set rate of change
			Mov	[EBX+eCnt],ESI
			SetZ	AL
			ShL	AL,7									;E_IDLE

			Mov	AH,[EBX+eMode]						;Preserve ADSR mode
			And	AH,70h
			Or		AL,AH

			Test	DL,60h								;Jump to the right mode
			JZ 	.GainDec
			Test	DL,40h
			JZ 	short .GainExp
			Test	DL,20h
			JZ 	short .GainInc

			.GainBent:
				Mov	dword [EBX+eAdj],A_LIN
				Mov	dword [EBX+eDest],D_BENT
				Or		AL,E_BENT						;Set mode to bent line increase
				Mov	[EBX+eMode],AL
				RetN

ALIGN	16
			.GainInc:
				Mov	dword [EBX+eAdj],A_LIN
				Mov	dword [EBX+eDest],D_MAX
				Or		AL,E_INC							;Set mode to linear increase
				Mov	[EBX+eMode],AL
				RetN

ALIGN	16
			.GainExp:
				Mov	dword [EBX+eAdj],A_EXP
				Mov	dword [EBX+eDest],D_MIN
				Or		AL,E_EXP							;Set mode to exponential decrease
				Mov	[EBX+eMode],AL
				RetN

ALIGN	16
			.GainDec:
				Mov	dword [EBX+eAdj],A_LIN
				Mov	dword [EBX+eDest],D_MIN
				Or		AL,E_DEC							;Set mode to linear decrease
				Mov	[EBX+eMode],AL

	EnvDone:

ENDP


;
;Change Envelope
;
;Called when the ADSR registers are written to while an envelope is in progress.  Control is
;transfered to StartEnv where the envelope is updated.
;
;In:
;   EBX = Voice << 4
;
;Destroys:
;   EAX,EDX,EBX

PROC ChgADSR

	Push	ESI											;ESI will get popped on return from StartEnv
	LEA	ESI,[EBX+dsp]
	LEA	EBX,[EBX*8+mix]

	XOr	EAX,EAX
	Mov	DL,[EBX+eMode]
	And	DL,0Fh

	Cmp	DL,E_ATT										;If the envelope isn't in attack, decay, or sustain
	JE		ChgAtt										; mode, changes to the ADSR registers have no effect
	Cmp	DL,E_DECAY
	JE 	ChgDec
	Cmp	DL,E_SUST
	JE 	ChgSus

	Pop	ESI											;No changes were made, pop ESI and return

ENDP


;
;DSP Data Port

;--------------------------------------------
;External procedure for users of SNESAPU.DLL

PROC SetDSPReg, dReg, dVal
USES ECX,EDX,EBX

	MovZX	EBX,byte [dReg]
	MovZX	EAX,byte [dVal]
	Mov	CL,0											;CL = Don't emulate DSP
	Call	DSPInB										;Process register write without calling debug function

ENDP


;--------------------------------------------
;Internal procedure for initialzing DSP registers

PROC InitReg
USES ECX

	Mov	CL,0											;CL = Don't emulate DSP
	Call	DSPInC										;Process register regardless of current register value

ENDP


;--------------------------------------------
;Procedure for writing to the DSP from the SPC700

PROC DSPIn

%if	DEBUG
	Mov	EDX,[pTrace]
	Test	EDX,EDX
	JZ		short .NoDbg

		MovZX	EAX,AL
		Add	EBX,dsp

		Push	ECX,ESI,EDI								;Save these registers
		Push	EAX										;Pass these as parameters
		Push	EBX
		Call	EDX
		Pop	EBX
		Pop	EAX
		Pop	EDI,ESI,ECX

		MovZX	EBX,BL

	.NoDbg:
%endif

	Mov	CL,1											;CL = Emulate DSP to catch up to current state

DSPInB:

	Test	BL,80h										;Writes to 80-FFh have no effect (reads are mirrored
	JNZ	DSPDone										; from lower mem)

	Cmp	BL,kon
	JE		RKOn
	Cmp	BL,kof										;Check for registers that can have duplicate data
	JE		RKOf											; written
	Cmp	BL,endx
	JE 	REndX

	Cmp	AL,[EBX+dsp]								;Is the new data the same as the current data?
	JZ		short DSPDone								;  Yes, Don't bother updating

	Mov	[EBX+dsp],AL								;Update DSP RAM

DSPInC:
	Mov	EDX,[EBX*4+dspRegs]						;Get the pointer to the register handler

	Mov	AH,BL
	And	EBX,70h
	Not	AH
	ShL	EBX,3											;EBX indexes mix (needed by some handlers)
	And	AH,MFLG_OFF									;AH = 08h if the register is in dsp.voice

%if DSPINTEG
	Test	CL,CL											;If write was from SPC700, emulate DSP before
	JZ		short .NoOutput							; processing new register data
		Call	CatchUp
	.NoOutput:
%endif

	Test	[EBX+mix+mFlg],AH							;Is the voice inactive?
	JNZ	short DSPDone								;  Yes, Don't bother updating

	Jmp	EDX

DSPDone:
	XOr	EAX,EAX										;DSP state didn't change

ENDP


;
;DSP Register Handlers

;============================================
;End block decoded

ALIGN	16
REndX:
%if DSPINTEG
	Test	CL,CL											;If write was from SPC700, emulate DSP before
	JZ		short .NoOutput							; processing new register data
		Call	CatchUp
	.NoOutput:
%endif

	Mov	EBX,[pAPURAM]
	MovZX	EAX,byte [dsp+endx]
	Mov	byte [0F3h+EBX],0
	Mov	byte [dsp+endx],0							;Reset the ENDX register
	Test	EAX,EAX
	SetNZ	AL
	Ret

;============================================
;Key On/Off

ALIGN	16
RKOf:
%if DSPINTEG
	Test	CL,CL											;If write was from SPC700, emulate DSP before
	JZ		short .NoOutput							; processing new register data
		Call	CatchUp
	.NoOutput:
%endif

	XOr	EDX,EDX
	Mov	[dsp+kof],AL
	And	AL,[voiceMix]								;Only check voices that are currently playing
	JZ		.Done

	Mov	EDX,[31*4+rateTab]
	Mov	EBX,mix
	.KOffLoop:
		Test	AL,1										;Has the current voice been flagged to be keyed off?
		JZ 	short .NoKOff							;  No
			Mov	byte [EBX+eRIdx],31				;Place envelope in release mode
			Mov	dword [EBX+eRate],EDX
			Mov	dword [EBX+eCnt],EDX
			Mov	dword [EBX+eAdj],A_KOF
			Mov	dword [EBX+eDest],D_MIN
			Mov	byte [EBX+eMode],E_REL
			Or 	byte [EBX+mFlg],MFLG_KOFF		;Flag voice as keying off
		.NoKOff:
		Sub	EBX,-80h

	ShR	AL,1
	JNZ	short .KOffLoop

	XOr	EDX,EDX										;Mark DSP state as having changed
	Inc	EDX

.Done:
	;Problem:
	;The DSP processes the KON register before the KOF register.  Because of this, voices that have
	;their key off flag set can't be keyed on.  However, some games (like Super Turrican 2) write to
	;the KON register then immediately write to the KOF register.  This works on the SNES because the
	;DSP processes the key registers every other sample frame, I think, and the two writes happen
	;within that time frame.
	;
	;Here in the emulation code I process the registers as soon as they're written to, and so games
	;that write to the KOF register after the KON register will be missing notes.
	;
	;Solution:
	;If less than 62.5us have passed since KON was written to, reprocess it.

	Mov	EAX,[t64Cnt]								;Has it been at least 62.5us since KON was written to?
	Sub	EAX,[konWr]
	Cmp	EAX,4
	JB		short .KOn

	Mov	EAX,EDX
	Ret

	.KOn:
	Mov	AL,[dsp+kon]
	Push	EDX
	Call	RKOn2
	Pop	EDX

	Or		EAX,EDX										;If KON or KOF changed DSP, return true
	Ret


ALIGN	16
RKOn:
%if DSPINTEG
	Test	CL,CL
	JZ		short .NoOutput
		Call	CatchUp
	.NoOutput:
%endif

	Mov	[dsp+kon],AL
	Mov	EBX,[t64Cnt]								;Save current time
	Mov	[konWr],EBX

RKOn2:
	Mov	CL,[dsp+kof]
	Not	CL
	And	CL,AL											;Erase any keys that are off
	JZ		short .Done

	Push	ECX,ESI,EDI
	Mov	CH,1
	Mov	EBX,mix
	Mov	ESI,dsp
	.ResetC:
		Test	CL,CH
		JZ 	.NoKey
			And	byte [EBX+mFlg],MFLG_MUTE		;Reset flags

			;Set voice volume ------------------
			Sub	EBX,mix
			Mov	AL,[ESI+volL]
			Call	[dspRegs]
%if STEREO=0
			Mov	AL,[ESI+volR]
			Call	[4+dspRegs]
			Add	EBX,mix
%else
			Add	EBX,mix
			Mov	EAX,[EBX+mTgtL]
			Mov	[EBX+mChnL],EAX

			Mov	EAX,[EBX+mTgtR]
			Mov	[EBX+mChnR],EAX
%endif
			;Set pitch -------------------------
			MovZX	EAX,word [ESI+pitch]
			And	AH,3Fh
			Mov	[EBX+mOrgP],EAX
			Mul	dword [pitchAdj]
			ShRD	EAX,EDX,16
			AdC	EAX,0
			Mov	[EBX+mRate],EAX
			Mov	word [EBX+mDec],0

			Call	StartSrc								;Start waveform decompression
			Call	StartEnv								;Start envelope

			Or		[voiceMix],CH						;Mark voice as being on internally
		.NoKey:
		Add	ESI,10h
		Sub	EBX,-80h

	Add	CH,CH
	JNZ	.ResetC

	Pop	EDI,ESI,ECX
	XOr	EAX,EAX
	Inc	EAX
	Ret

	.Done:
	XOr	EAX,EAX
	Ret

;============================================
;Voice volume

%if STEREO=0
ALIGN	16
RVolL:
	MovSX	EAX,AL										;Sign extend volume to 32-bits
	Mov	[EBX+mix+mChnL],EAX
	Mov	[EBX+mix+mTgtL],EAX

	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
RVolLF:
	MovSX	EAX,AL
	Mov	[ESP-4],EAX
	FILd	dword [ESP-4]
	FMul	dword [fpShR7]								;Convert volume from fixed to floating-point
	FSt	dword [EBX+mix+mChnL]
	FStP	dword [EBX+mix+mTgtL]

	XOr	EAX,EAX
	Inc	EAX
	Ret
%endif

ALIGN	16
RVolLM:
	MovSX	EAX,AL
	CDQ													;Get the absolute value of the volume to keep inverse
	XOr	EAX,EDX										; volumes from cancelling each other out
	Sub	EAX,EDX
	Mov	[EBX+mix+mChnL],EAX
	Mov	[EBX+mix+mTgtL],EAX

	XOr	EAX,EAX
	Inc	EAX
	Ret

%if STEREO=0
ALIGN	16
RVolR:
	MovSX	EAX,AL
	Mov	[EBX+mix+mChnR],EAX
	Mov	[EBX+mix+mTgtR],EAX

	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
RVolRF:
	MovSX	EAX,AL
	Mov	[ESP-4],EAX
	FILd	dword [ESP-4]
	FMul	dword [fpShR7]
	FSt	dword [EBX+mix+mChnR]
	FStP	dword [EBX+mix+mTgtR]

	XOr	EAX,EAX
	Inc	EAX
	Ret
%endif

ALIGN	16
RVolRM:
	MovSX	EAX,AL
	CDQ
	XOr	EAX,EDX
	Sub	EAX,EDX
	Mov	[EBX+mix+mChnR],EAX
	Mov	[EBX+mix+mTgtR],EAX

	XOr	EAX,EAX
	Inc	EAX
	Ret

;============================================
;Pitch

ALIGN	16
RPitch:
	ShR	EBX,3
	And	byte [EBX+dsp+pitch+1],3Fh				;The upper two bits of the pitch don't count
	MovZX	EAX,word [EBX+dsp+pitch]
	ShL	EBX,3
	Mov	[EBX+mix+mOrgP],EAX

	Mul	dword [pitchAdj]							;Convert the pitch into a more meaningful value
	ShRD	EAX,EDX,16									;Remove 16-bit fraction from pitchAdj
	Mov	[EBX+mix+mRate],EAX

	XOr	EAX,EAX
	Inc	EAX
	Ret

;============================================
;Envelope

ALIGN	16
RADSR:
	XOr	EAX,EAX
	Test	byte [EBX+mix+mFlg],MFLG_KOFF			;Is voice in key off mode?
	JNZ	short .NoChg								;  Yes, Envelope setting can't be changed now

		Mov	AL,[EBX+mix+eMode]					;AL = ADSR or Gain mode
		ShR	EBX,3
		And	AL,E_ADSR
		Mov	AH,[EBX+dsp+adsr]
		And	AH,80h
		Or		AL,AH

		Test	AL,80h + E_ADSR
		JZ		short .NoChg							;Envelope is already in gain mode, do nothing
		Test	AL,80h
		JZ		short .SetGain							;Switched from ADSR to Gain
		Test	AL,E_ADSR
		JNZ	.Change									;Envelope is in ADSR mode, update settings

		Mov	AL,[EBX*8+mix+eMode]					;Switched from Gain to ADSR, restore previous ADSR
		ShR	AL,4										; state then update settings
		Or		AL,E_ADSR
		Mov	[EBX*8+mix+eMode],AL

	.Change:
		Call	ChgADSR
		XOr	EAX,EAX
		Inc	EAX

	.NoChg:
	Ret

	.SetGain:
		ShL	EBX,3
		ShL	byte [EBX+mix+eMode],4				;Save ADSR state, ChgGain will set bits 7 and 3-0

;		Mov	AL,[EBX+mix+eMode]					;If envelope switches to gain while in attack mode and
;		Cmp	AL,E_ATT << 4							; and attack time is 0ms, then force envelope to max.
;		JNE	short RGain
;		Cmp	dword [ebx+mix+eAdj],A_NOATT
;		JNE	short RGain
;		Mov	dword [EBX+mix+eVal],D_MAX
		Jmp	short RGain

ALIGN 16
RGain:
	XOr	EAX,EAX
	Test	byte [EBX+mix+mFlg],MFLG_KOFF			;Is voice in key off mode?
	JNZ	short .NoChg								;  Yes, Envelope setting can't be changed now

	ShR	EBX,3
	Test	byte [EBX+dsp+adsr],80h					;Is envelope in gain mode?
	JNZ	short .NoChg								;  No, Setting gain register has no effect

		Push	dword .Return
		Push	ESI										;StartEnv will pop ESI on return
		LEA	ESI,[EBX+dsp]
		LEA	EBX,[EBX*8+mix]
		Jmp	ChgGain									;Begin ADSR envelope

		.Return:
		XOr	EAX,EAX
		Inc	EAX

	.NoChg:
	Ret

;============================================
;Main volumes

ALIGN	16
RMVolL:
	MovSX	EAX,AL
	IMul	EAX,[volAdj]
	SAR	EAX,16										;Remove fraction from multiplication
	Mov	[volMainL],EAX
	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
RMVolLM:
	MovSX	EAX,AL
	CDQ
	XOr	EAX,EDX
	Sub	EAX,EDX
	IMul	EAX,46341									;Multiply by 1/2 sqrt 2 to help average
	IMul	dword [volAdj]
	Mov	[volMainL],EDX								;EDX contains integer result (fraction removed)
	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
RMVolLF:
	MovSX	EAX,AL
	Mov	[ESP-4],EAX
	FILd	dword [ESP-4]
	FIMul	dword [volAdj]
	FMul	dword [fpShR7]								;>> 7 to turn MVOL into a float
	FStP	dword [volMainL]							;Leave the 16-bits added by volAdj so the final
	XOr	EAX,EAX										; output will be 32-bit instead of 16-bit
	Inc	EAX
	Ret

ALIGN	16
RMVolR:
	XOr	AL,[surround]								;Invert right output, if surround is enabled
	Sub	AL,[surround]
	MovSX	EAX,AL
	IMul	EAX,[volAdj]
	SAR	EAX,16
	Mov	[volMainR],EAX
	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
RMVolRM:
	MovSX	EAX,AL
	CDQ
	XOr	EAX,EDX
	Sub	EAX,EDX
	IMul	EAX,46341
	IMul	dword [volAdj]
	Mov	[volMainR],EDX
	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
RMVolRF:
	XOr	AL,[surround]
	Sub	AL,[surround]
	MovSX	EAX,AL
	Mov	[ESP-4],EAX
	FILd	dword [ESP-4]
	FIMul	dword [volAdj]
	FMul	dword [fpShR7]
	FStP	dword [volMainR]
	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
REVolL:
	MovSX	EAX,AL
	IMul	EAX,[volAdj]
	SAR	EAX,16										;Remove fraction
	Mov	[volEchoL],EAX

	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
REVolLM:
	MovSX	EAX,AL
	CDQ
	XOr	EAX,EDX
	Sub	EAX,EDX

	IMul	EAX,46341
	IMul	dword [volAdj]
	Mov	[volEchoL],EDX

	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
REVolLF:
	MovSX	EAX,AL
	Mov	[ESP-4],EAX
	FILd	dword [ESP-4]
	FIMul	dword [volAdj]
	FMul	dword [fpShR7]
	FStP	dword [volEchoL]

	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
REVolR:
	XOr	AL,[surround]
	Sub	AL,[surround]
	MovSX	EAX,AL
	IMul	EAX,[volAdj]
	SAR	EAX,16
	Mov	[volEchoR],EAX

	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
REVolRM:
	MovSX	EAX,AL
	CDQ
	XOr	EAX,EDX
	Sub	EAX,EDX

	IMul	EAX,46341
	IMul	dword [volAdj]
	Mov	[volEchoR],EDX

	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
REVolRF:
	XOr	AL,[surround]
	Sub	AL,[surround]
	MovSX	EAX,AL
	Mov	[ESP-4],EAX
	FILd	dword [ESP-4]
	FIMul	dword [volAdj]
	FMul	dword [fpShR7]
	FStP	dword [volEchoR]

	XOr	EAX,EAX
	Inc	EAX
	Ret

;============================================
;Echo settings

ALIGN	16
REFB:
%if STEREO
	MovSX	EAX,AL
	Mov	EDX,EAX
	IMul	EAX,[efbct]
	SAR	EAX,16
	Mov	[echoFB],EAX
	Mov	[4+echoFB],EAX

	Mov	EAX,10000h									;Calculate crosstalk
	Sub	EAX,[efbct]
	IMul	EAX,EDX
	SAR	EAX,16
	Mov	[echoFBCT],EAX
	Mov	[4+echoFBCT],EAX
%else
	MovSX	EAX,AL
	Mov	[echoFB],EAX
	Mov	[4+echoFB],EAX
%endif

	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
REFBM:
	MovSX	EAX,AL
	Mov	[echoFB],EAX
	Mov	[4+echoFB],EAX
%if STEREO
	XOr	EAX,EAX
	Mov	[echoFBCT],EAX
	Mov	[4+echoFBCT],EAX
%endif

	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
REFBF:
	MovSX	EAX,AL
	Mov	[ESP-4],EAX
	FILd	dword [ESP-4]
%if STEREO
	FLd	ST

	FIMul	dword [efbct]
	FMul	dword [fpShR23]							;Convert from fixed to floating-point
	FSt	dword [echoFB]								;7-bits (efb) + 16-bits (efbct) = 23-bits
	FStP	dword [4+echoFB]

	FLd	dword [fp64k]
	FISub	dword [efbct]
	FMulP	ST1,ST
	FMul	dword [fpShR23]
	FSt	dword [echoFBCT]
	FStP	dword [4+echoFBCT]
%else
	FMul	dword [fpShR7]
	FSt	dword [echoFB]
	FStP	dword [4+echoFB]
%endif

	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
REDl:
	Push	ECX
	ShL	AL,4											;Calculate number of samples to delay
	MovZX	EAX,AL										;EAX = (EFB * 16) * Rate / 1000ms
	Mul	dword [dspRate]
	Mov	ECX,1000
	Div	ECX
	Test	EAX,EAX										;if EAX = 0, EAX = 1
	SetZ	CL
	Or		AL,CL
	ShL	EAX,3											;Multiply by 8, since echo is stored in 32-bit stereo
	Mov	[echoDel],EAX
	Pop	ECX

	XOr	EAX,EAX
	Inc	EAX
	Ret

ALIGN	16
RFCI:
	ShR	EBX,4
	MovSX	EAX,AL										;Sign extend filter coefficient
	Mov	[EBX+firTaps],EAX							;Store for left and right (for parallel instructions)
	Mov	[4+EBX+firTaps],EAX
	Test	EBX,EBX
	JZ		short CheckFC1
	Jmp	short CheckFC								;Check if filter coefficients do anything

ALIGN 16
RFCF:
	ShR	EBX,4
	MovSX	EAX,AL
	Mov	[ESP-4],EAX
	FILd	dword [ESP-4]
	FMul	dword [fpShR7]
	FSt	dword [EBX+firTaps]
	FStP	dword [4+EBX+firTaps]
	Test	EBX,EBX
	JNZ	short CheckFC

CheckFC1:
	Cmp	AL,7Fh										;If first tap is 127, pretend it's 0
	SetE	AH
	Dec	AH
	And	AL,AH

CheckFC:
	Test	AL,AL
	SetNZ	AL
	ShR	EBX,3
	Mov	AH,1
	Mov	EDX,ECX
	Mov	ECX,EBX
	ShL	EAX,CL
	Not	AH
	And	[firEnabl],AH								;Set corresponding bit if tap is non-zero
	Or		[firEnabl],AL
	Mov	ECX,EDX

	;Disable echo if all taps are 0 ----------
	Mov	AL,[firEnabl]								;Mask for all taps
	Or		AL,[dsp+fc]									;Value of first tap (here, 127 doesn't count as 0)
	SetZ	AL												;If all taps are 0, including first one, set AL
	And	byte [disEcho],~1
	Or		[disEcho],AL								;Set first bit to disable echo

	XOr	EAX,EAX										;DSP state changed if echo was enabled
	Inc	EAX
	Ret

;============================================
;Other

ALIGN	16
RPMOn:
	Push	ECX

	And	AL,0FEh										;Disable LSB
	Mov	EDX,[pAPURAM]
	Mov	[dsp+pmon],AL
	Mov	[EDX+0F3h],AL

	;Reset all pitch on all voices -----------
	Mov	EBX,mix
	Mov	CL,8
	.Next:
		Mov	EAX,[EBX+mOrgP]
		Mul	dword [pitchAdj]
		ShRD	EAX,EDX,16
		AdC	EAX,0
		Mov	[EBX+mRate],EAX
		Sub	EBX,-80h
	Dec	CL
	JNZ	short .Next

	XOr	EAX,EAX
	Inc	EAX
	Pop	ECX
	Ret

ALIGN	16
RFlg:
	Test	AL,80h										;Has a soft reset been initialized?
	JZ 	short .NoSRst								;  No
		And	AL,~80h
		Or 	AL,60h									;Turn on mute and disable echo
		Mov	[dsp+flg],AL
		Mov	byte [dsp+endx],0						;Clear end block flags
		Mov	byte [dsp+kon],0
		Mov	byte [dsp+kof],0
		Mov	byte [voiceMix],0

		;Reset internal voice settings --------
		Mov	EBX,mix+mFlg
		Mov	AL,8
		.MFlg:
			And	byte [EBX],MFLG_MUTE				;Preserve mute flag
			Or		byte [EBX],MFLG_OFF				;Set voice to inactive
			Sub	EBX,-80h
		Dec	AL
		JNZ	.MFlg
	.NoSRst:

	;Disable echo ----------------------------
	Mov	DL,AL
	And	DL,20h										;Isolate "disable echo" flag
	And	byte [disEcho],~20h
	Or		[disEcho],DL								;Store it

	;Update noise clock ----------------------
	Mov	dword [nRate],0
	And	EAX,1Fh
	JZ		short .NoNoise
		Mov	EBX,EAX
		Mov	EAX,-1
		Mov	EDX,65535
		Div	dword [EBX*4+rateTab]
		Mov	[nRate],EAX
	.NoNoise:

	XOr	EAX,EAX
	Inc	EAX
	Ret

;============================================
;Null register

ALIGN	16
RNull:
	XOr	EAX,EAX
	Ret


;
;No Interpolation

PROC NoneI
	Mov	ESI,[EBX+sIdx]								;ESI -> Current sample
	MovSX	EAX,word [ESI]								;EAX = Current sample
ENDP

PROC NoneX
	Mov	ESI,[EBX+sIdx]
	MovSX	EAX,word [ESI]
ENDP

PROC NoneF
	Mov	EDX,[EBX+sIdx]
	FILd	word [EDX]
ENDP


;
;Linear Interpolation

PROC LinearI
	Mov	ESI,[EBX+sIdx]								;ESI -> Current sample
	MovSX	EAX,word [ESI]
	MovSX	EDX,word [ESI-2]
	Sub	EAX,EDX
	SAR	EAX,1
	IMul	EAX,dword [EBX+mDec]
	SAR	EAX,15
	Add	EAX,EDX
ENDP

PROC LinearX
	Mov	ESI,[EBX+sIdx]
	MovSX	EAX,word [ESI]
	MovSX	EDX,word [ESI-2]
	Sub	EAX,EDX
	SAR	EAX,1
	IMul	EAX,[EBX+mDec]
	SAR	EAX,15
	Add	EAX,EDX
ENDP

PROC LinearF
	Mov	EDX,[EBX+sIdx]
	FILd	word [EDX-2]
	FILd	word [EDX]
	FSub	ST,ST1										;Difference between samples
	FIMul	dword [EBX+mDec]							;Multiply by delta x from last sample
	FMul	dword [fpShR16]
	FAddP	ST1,ST
ENDP


;
;Cubic Interpolation

PROC CubicI
	Mov	ESI,[EBX+sIdx]								;ESI -> Current sample
	Mov	EAX,[EBX+mDec]								;EAX = Sample delta
	Push	ECX,EBX
	ShR	EAX,8
	LEA	EBX,[cubicTab+EAX*8]						;EBX -> Cubic table value
	Mov	AX,[ESI-6]									;Get first sample
	IMul	word [EBX+d0]								;Multiply by cubic derivative
	Mov	AX,[ESI-4]
	Mov	CX,DX											;DX = Partial sample >> 1
	IMul	word [EBX+d1]
	Mov	AX,[ESI-2]
	Add	CX,DX
	IMul	word [EBX+d2]
	Mov	AX,[ESI-0]
	Add	CX,DX
	IMul	word [EBX+d3]
	Add	CX,DX
	Pop	EBX
	MovSX	EAX,CX										;EAX = New sample
	Pop	ECX
	Add	EAX,EAX										;15-bit to 16-bit
ENDP

PROC CubicX
	Mov		ESI,[EBX+sIdx]
	Mov		EAX,[EBX+mDec]							;EAX = Sample delta decimal
	MovQ		MM7,[ESI-6]								;Load all four samples (isn't MMX nice?)
	ShR		EAX,8										;EAX indexes interpolation table value
	PMAddWD	MM7,[cubicTab+EAX*8]					;Multiply samples by table
	PSRAD		MM7,15
	MovQ		MM6,MM7
	PSLLQ		MM7,32
	PAddD		MM7,MM6									;Add two dwords together and store in upper dword of MM7
	PackSSDW	MM7,MM0									;Pack dwords into words (EmuDSPX requires 16-bit samples)
	MovD		EAX,MM7
	SAR		EAX,16									;Sign extend 16-bit result to 32-bits
ENDP

PROC CubicF
	Mov	EDX,[EBX+sIdx]
	FILd	dword [EBX+mDec]							;delta x = mDec>>16
	FMul	dword [fpShR16]							;								|x

	FLd	ST												;								|x x
	FMul	ST,ST1										;								|x x^2
	FLd	ST												;								|x x2 x2
	FMul	ST,ST2										;								|x x2 x^3
	FLd	ST												;								|x x2 x3 x3
	FMul	dword [fp0_5]								;								|x x2 x3 x3*0.5
	FAdd	ST1,ST										;								|x x2 1.5*x3 .5x3
	FXch	ST3											;								|.5x3 x2 1.5x3 x
	FMul	dword [fp0_5]								;								|.5x3 x2 1.5x3 x*0.5

	;s[-1] = -.5(x^3) + (x^2) - .5x + 0
	FLd	ST2											;								|.5x3 x2 1.5x3 .5x x2
	FSub	ST,ST4										;								|.5x3 x2 1.5x3 .5x x2-.5x3
	FSub	ST,ST1										;								|.5x3 x2 1.5x3 .5x D0-.5x

	;s[1] = -1.5(x^3) + 2(x^2) + .5x + 0
	FXch	ST1											;								|.5x3 x2 1.5x3 D0 .5x
	FAdd	ST,ST3										;								|.5x3 x2 1.5x3 D0 .5x+x2
	FAdd	ST,ST3										;								|.5x3 x2 1.5x3 D0 D2+x2
	FSub	ST,ST2										;								|.5x3 x2 1.5x3 D0 D2-1.5x3

	;s[0] = 1.5(x^3) - 2.5(x^2) + 0x + 1
	FXch	ST2											;								|.5x3 x2 D2 D0 1.5x3
	FLd1
	FAddP	ST1,ST
	FLd	ST3
	FMul	dword [fn2_5]
	FAddP	ST1,ST										;								|.5x3 x2 D2 D0 D1

	;s[2] = .5(x^3) - .5(x^2) + 0x + 0
	FXch	ST3											;								|.5x3 D1 D2 D0 x2
	FMul	dword [fp0_5]								;								|.5x3 D1 D2 D0 x2*0.5
	FSubP	ST4,ST										;								|.5x3-.5x2 D1 D2 D0

	FIMul	word [EDX-6]								;								|D3 D1 D2 D0*S
	FXch	ST3											;								|SO D1 D2 D3
	FIMul	word [EDX-0]								;								|SO D1 D2 D3*S
	FAddP	ST3,ST										;								|S0+S2 D1 D2
	FIMul	word [EDX-2]								;								|S D1 D2*S
	FAddP	ST2,ST										;								|S+S2 D1
	FIMul	word [EDX-4]								;								|S D1*S
	FAddP	ST1,ST										;								|S+S1

;	;Use lookup table ------------------------
;	Mov	EAX,dword [EBX+mDec]						;EAX = Sample delta decimal
;	ShR	EAX,8											;EAX indexes interpolation table value
;	LEA	ESI,[EAX*8+cubicTab]
;	FILd	word [EBX+sIntC]							;Get first sample
;	FIMul	[ESI.d0]
;	FILd	word [2+EBX+sIntC]
;	FIMul	[ESI.d1]
;	FILd	word [4+EBX+sIntC]
;	FIMul	[ESI.d2]
;	FILd	word [6+EBX+sIntC]
;	FIMul	[ESI.d3]
;	FAddP	ST1,ST
;	FAddP	ST1,ST
;	FAddP	ST1,ST
;	FMul	[fpShR15]
ENDP


;
;Gaussian Interpolation

PROC GaussI
	Mov	ESI,[EBX+sIdx]								;ESI -> Current sample
	Mov	EAX,[EBX+mDec]								;EAX = Sample delta
	Push	ECX,EBX
	ShR	EAX,8
	LEA	EBX,[gaussTab+EAX*8]						;EBX -> Gaussian table value
	Mov	AX,[ESI-6]									;Get first sample
	IMul	word [EBX+d0]								;Multiply by cubic derivative
	Mov	AX,[ESI-4]
	Mov	CX,DX											;DX = Partial sample >> 1
	IMul	word [EBX+d1]
	Mov	AX,[ESI-2]
	Add	CX,DX
	IMul	word [EBX+d2]
	Mov	AX,[ESI-0]
	Add	CX,DX
	IMul	word [EBX+d3]
	Add	CX,DX
	Pop	EBX
	MovSX	EAX,CX										;EAX = New sample
	Pop	ECX
	Add	EAX,EAX										;15-bit to 16-bit
ENDP

PROC GaussX
	Mov		ESI,[EBX+sIdx]
	Mov		EAX,[EBX+mDec]							;EAX = Sample delta decimal
	MovQ		MM6,[ESI-6]
	ShR		EAX,8										;EAX indexes interpolation table value
	MovQ		MM7,MM6
	PAnd		MM6,[loSmp]								;MM6 = 00 s2 00 s0
	PAnd		MM7,[hiSmp]								;MM7 = s3 00 s1 00

	PMAddWD	MM6,[gaussTab+EAX*8]					;Multiply samples by table
	PMAddWD	MM7,[gaussTab+EAX*8]

	PSRAD		MM6,15
	PSRAD		MM7,15
	PAddD		MM7,MM6
	MovD		EAX,MM7
	PSRLQ		MM7,32
	MovD		EDX,MM7
	Add		EAX,EDX
	And		EAX,~1
ENDP

PROC GaussF
	Mov	EAX,[EBX+mDec]								;EAX = Sample delta decimal
	Mov	EDX,[EBX+sIdx]
	ShR	EAX,8											;EAX indexes interpolation table value
	LEA	ESI,[EAX*8+gaussTab]
	FILd	word [EDX-6]								;Get first sample
	FIMul	word [ESI+d0]
	FILd	word [EDX-4]
	FIMul	word [ESI+d1]
	FILd	word [EDX-2]
	FIMul	word [ESI+d2]
	FILd	word [EDX]
	FIMul	word [ESI+d3]
	FAddP	ST1,ST
	FAddP	ST1,ST
	FAddP	ST1,ST
	FMul	dword [fpShR15]
ENDP


;
;Noise Generator
;
;Generates white noise samples
;
;Out:
;   nSmp = Random 16-bit sample
;
;Destroys:
;   EAX,EDX

%macro NoiseGen 0
	Mov	EAX,[nRate]
	Add	[nAcc],EAX
	JNC	short %%NoNInc
		IMul	EAX,[nSmp],27865						;X=(AX+C)%M  Where: X<M and 2<=A<M and 0<C<M
		Add	EAX,7263									;Add C
		CWDE												;Modulus M (32768)
		Mov	[nSmp],EAX
	%%NoNInc:
%endmacro


;
;Pitch Modulation
;
;Changes the pitch based on the output of the previous voice:
;
; P' = (P * (OUTX + 32768)) >> 15
;
;Pitch modulation in the SNES uses the full 16-bit sample value, not the 8-bit value in OUTX as
;previously believed.
;
;In:
;   CH  = Bitmask for current voice
;   EBX-> Current voice in 'mix'
;
;Destroys:
;   EAX,EDX

%macro PitchMod 0
	Test	[dsp+pmon],CH								;Is pitch modulation enabled?
	JZ 	short %%NoPMod								;  No, Pitch doesn't need to be adjusted
		;Adjust pitch by sample value ---------
		Mov	EAX,[EBX+mOut-80h]					;EAX = Wave height of last voice (-16.15)
		Add	EAX,32768								;Unsign sample
		IMul	EAX,dword [EBX+mOrgP]				;Apply sample height to pitch
		SAR	EAX,15

		;Clamp pitch to 14-bits ---------------
		Mov	EDX,EAX
		SAR	EDX,14
		JZ		short %%PitchOK
			SetS	AL
			MovZX	EAX,AL
			Dec	EAX
			And	EAX,3FFFh
		%%PitchOK:

		;Convert pitch to sample rate ---------
		Mul	dword [pitchAdj]
		ShRD	EAX,EDX,16
		Mov	[EBX+mRate],EAX
	%%NoPMod:
%endmacro


;
;Process Sound Source
;
;Updates the current sample position and decompresses the next block if necessary
;
;In:
;   CH  = Bitmask for current voice
;   EBX-> Current voice in 'mix'
;
;Destroys:
;   EAX,EDX,CL,ESI

%macro UpdateSrc 0
	;Update sample index ---------------------
	Mov	CL,[EBX+mRate+2]							;CL = Number of whole samples to increase index by
	Mov	EAX,[EBX+mRate]							;AX = Fraction of sample to increase index by
	Add	[EBX+mDec],AX								;Add AX to the decimal counter
	AdC	CL,0											;Add carry, if any, to increase amount
	JZ 	%%NoSInc										;If the amount is zero, index didn't increase

	;Check for end of block ------------------
	Add	CL,CL											;CL <<= 1  (for 16-bit samples)
	Add	[EBX+sIdx],CL								;Increase sample index
	Test	byte [EBX+sIdx],20h						;Have we reached the end of the block?
	JZ		%%NoSInc										;  Nope

		And	byte [EBX+sIdx],~20h					;Adjust sample index for wrap around
		Mov	EAX,[EBX+sBuf+24]						;Copy last four samples of buffer
		Mov	EDX,[EBX+sBuf+28]						; (needed for interpolation)
		Add	word [EBX+bCur],9						;Move to next sample block
		Mov	[EBX+sBuf-8],EAX
		Mov	[EBX+sBuf-4],EDX

		Test	byte [EBX+bHdr],1						;Was this the end block?
		JZ 	short %%NotEndB						;  No, Decompress next block
		Or 	[dsp+endx],CH							;Set flag in ENDX
		Test	byte [EBX+bHdr],2						;Is this source looped?
		JNZ	short %%LoopB							;  Yes, Start over at loop point

		;End voice playback -------------------
		%%EndPlay:
			Not	CH
			And	[voiceMix],CH						;Don't include voice in mixing process
			Not	CH
			Mov	dword [EBX+eVal],0				;Reset envelope and wave height
			Mov	dword [EBX+mOut],0
			Or 	byte [EBX+mFlg],MFLG_OFF		;Mark voice as inactive
			And	byte [EBX+mFlg],~MFLG_KOFF
			Jmp	.VoiceDone

		;Restart loop -------------------------
		%%LoopB:
			Mov	EAX,[EBX+dspv]
			MovZX	EDX,byte [EAX+srcn]				;EDX = Source number
			Mov	EAX,[pAPURAM]
			Mov	AH,[dsp+dir]						;EAX -> Source directory
			Mov	AX,[EDX*4+EAX+2]
			Mov	[EBX+bCur],EAX						;Store loop point in current block pointer

		;Decompress next block ----------------
		%%NotEndB:
			Mov	ESI,[EBX+bCur]						;ESI -> Current sample block
			Push	EDI,EBX
			Mov	AL,[ESI]								;Get block header
			LEA	EDI,[EBX+sBuf]						;EDI -> location to store samples
			Mov	[EBX+bHdr],AL						;Save header byte
			Mov	EDX,[EBX+sP1]						;Load previous two samples
			Mov	EBX,[EBX+sP2]
			Call	[pDecomp]							;Call user selected decompression routine
			Mov	EAX,EBX
			Pop	EBX,EDI
			Mov	[EBX+sP1],EDX						;Save last two samples in 32-bit form
			Mov	[EBX+sP2],EAX

			Mov	AL,[EBX+bHdr]
			And	AL,3
			Cmp	AL,1
			JNE	short %%NoSInc

			XOr	EAX,EAX
			Mov	[16+EBX+sBuf],EAX
			Mov	[20+EBX+sBuf],EAX
			Mov	[24+EBX+sBuf],EAX
			Mov	[28+EBX+sBuf],EAX
	%%NoSInc:
%endmacro


;
;Calculate Envelope Modification
;
;Changes the current height of the volume envelope based on its programming
;
;In:
;   EBX-> Current voice in mix
;   CH  = Current voice bit mask
;
;Destroys:
;   EAX,CL,EDX,ESI

%macro UpdateEnv 0

	Mov	CL,[EBX+eMode]

	Test	CL,E_IDLE									;Is the envelope constant?
	JNZ	%%EnvDone									;  Yes, Quit

	Dec	word [2+EBX+eCnt]							;Decrease sample counter, is it zero?
	JNZ	%%EnvDone									;  No, Quit

	Mov	EAX,[EBX+eRate]							;Restore sample counter
	Add	[EBX+eCnt],EAX

	;Adjust Envelope -------------------------
	Test	CL,E_TYPE									;Is the adjustment an exponential decrease?
	JZ 	short %%AdjLin								;  No, Go to linear
		Mov	EAX,[EBX+eVal]
		Mov	EDX,[EBX+eDest]						;Get destination (SL or 0)

		Neg	EAX
		SAR	EAX,8
		Add	[EBX+eVal],EAX							;Subtract 1/256th of envelope height
		Cmp	EDX,[EBX+eVal]							;Has height reached destination?
		JB 	%%EnvDone								;  No, Quit

		Test	EDX,EDX									;If destination isn't 0, change to sustain mode
		JNZ	%%AdjDone

		Or		byte [EBX+eMode],E_IDLE				;Otherwise mark envelope as no longer changing
		Jmp	%%EnvDone

ALIGN	16
	%%AdjLin:
	Test	CL,E_DIR										;Is the adjustment up or down?
	JZ 	short %%AdjDec
		Mov	EAX,[EBX+eVal]
		Add	EAX,[EBX+eAdj]							;Add adjustment to height
		Mov	[EBX+eVal],EAX
		Cmp	EAX,[EBX+eDest]						;Has height reached destination?
		JB 	%%EnvDone								;  No, Quit

		Cmp	EAX,D_MAX								;Clamp height to D_MAX
		JBE	%%AdjDone
		Mov	dword [EBX+eVal],D_MAX
		Jmp	%%AdjDone

	%%AdjDec:											;Linear decrease
		Mov	EAX,[EBX+eAdj]
		Sub	[EBX+eVal],EAX							;Decrease envelope height
		JA		%%EnvDone								;  Jump if envelope is > 0 (destination is always 0)

		Mov	dword [EBX+eVal],D_MIN

		Mov	AL,[EBX+eMode]							;If the envelope started out in ADSR mode, but was
		And	AL,~70h									; switched to Gain w/ linear decrease, the ADSR state
		Or		AL,E_SUST << 4							; will become sustain if ADSR is re-enabled.
		Mov	[EBX+eMode],AL

		Mov	AL,[EBX+mFlg]							;If the voice was getting keyed off, set MFLG_OFF to
		And	AL,MFLG_KOFF							; mark the voice as now being inactive
		Add	AL,AL
		SetZ	AH
		Or		[EBX+mFlg],AL
		And	byte [EBX+mFlg],~MFLG_KOFF

		Dec	AH
		And	AH,CH
		Not	AH
		And	[voiceMix],AH							;Disable voice mixing if keyed off

		Or		byte [EBX+eMode],E_IDLE				;Envelope is no longer changing
		Jmp	%%EnvDone

ALIGN	16
	%%AdjDone:

	;Change adjustment mode ------------------
	;(see StartEnv)
	Test	CL,E_ADSR									;Is envelope in ADSR mode?
	JZ 	%%EnvGain									;  Nope, Jump to Gain

	Mov	ESI,EBX
	And	CL,0Fh
	Sub	ESI,mix
	XOr	EAX,EAX
	ShR	ESI,3											;ESI indexes current voice in dsp

	Cmp	CL,E_DECAY									;Switch to next mode
	JE 	short %%EnvSust

	%%EnvDecay:
		Mov	dword [EBX+eAdj],A_EXP				;see StartEnv

		Mov	AL,[ESI+dsp+adsr+1]
		ShR	AL,5
		Inc	AL
		IMul	EAX,D_EXP
		Mov	[EBX+eDest],EAX

		MovZX	EAX,byte [ESI+dsp+adsr]
		And	AL,70h
		ShR	AL,3
		Add	AL,10h
		Mov	[EBX+eRIdx],AL
		Mov	ESI,[EAX*4+rateTab]

		Mov	[EBX+eRate],ESI
		Mov	[EBX+eCnt],ESI
		Mov	byte [EBX+eMode],E_DECAY
		Jmp	short %%EnvDone

ALIGN	16
	%%EnvSust:
		Mov	AL,[ESI+dsp+adsr+1]					;see StartEnv
		And	AL,1Fh
		Mov	[EBX+eRIdx],AL
		Mov	ESI,[EAX*4+rateTab]
		SetZ	AL
		Mov	[EBX+eRate],ESI
		Mov	[EBX+eCnt],ESI

		ShL	AL,7
		Mov	dword [EBX+eDest],D_MIN
		Or		AL,E_SUST
		Mov	[EBX+eMode],AL
		Jmp	short %%EnvDone

ALIGN	16
	%%EnvGain:
		Or		byte [EBX+eMode],E_IDLE				;Envelope is now constant

		Test	CL,E_DEST								;If gain is in "bent line" mode and line has reached
		JZ 	short %%EnvDone						; bend point, adjust envelope settings, otherwise
															; envelope is done.
		Cmp	dword [EBX+eDest],D_MAX
		JE		short %%EnvDone

		And	byte [EBX+eMode],~E_IDLE			;Undo idle flag
		Mov	dword [EBX+eAdj],A_BENT				;Slow down increase rate
		Mov	dword [EBX+eDest],D_MAX				;Set destination to max

ALIGN	16
	%%EnvDone:

%endmacro


;
;Finite Impulse Response Echo Filter
;
;Filters the echo using an eight tap FIR filter:
;
;        7
;       ---
;   x = \   c  * s
;       /    n    n
;       ---
;       n=0
;
;   x = output sample
;   c = filter coefficient (-.7)
;   s = unfiltered sample
;   n = 0 is the oldest sample and 7 is the most recent
;
;FIR filters are based on the sample rate.  This was fine in the SNES, because the sample rate was
;always 32kHz, but in the case of an emulator the sample rate can change.  So measures have to be
;taken to ensure the filter will have the same effect, regardless of the output sample rate.
;
;To overcome this problem, I figured each tap of the filter is applied every 31250ns.  So the
;solution is to calculate when 31250ns have gone by, and use the sample at that point.  Of course
;this method really only works if the output rate is a multiple of 32k.  In order to get accurate
;results, some sort of interpolation method needs to be introduced.  I went the cheap route and used
;linear interpolation.

;In:
;   EAX,EDX = Input samples
;
;Out:
;   EAX,EDX = Filtered samples
;   EDI    -> Echo buffer
;
;Destroys:
;   ECX,EBX

%macro FIRFilterI 0
	Sub	byte [firCur],4							;Move index back one sample. (Index will wrap around
	Mov	EDI,[firCur]								; after 64 samples, enough for up to 256kHz output.)
	LEA	EDI,[EDI*2+firBuf]						;EDI -> Current sample in filter buffer

	Mov	[EDI],EAX									;Store new samples in buffer
	Mov	[4+EDI],EDX
	Mov	[FIRBUF*2+EDI],EAX
	Mov	[FIRBUF*2+4+EDI],EDX

	Push	ESI,EBP
	XOr	EBX,EBX
	XOr	ECX,ECX
	Mov	ESI,firTaps+56								;ESI -> Filter taps
	Mov	[firDec],EBX								;Reset decimal overflow, so filtering is consistant
	Mov	EBP,8											;8-tap FIR filter
	%%Tap:
		Mov	EAX,[8+EDI]								;Interpolate left sample
		Sub	EAX,[EDI]
		IMul	dword [firDec]
		ShRD	EAX,EDX,16
		Add	EAX,[EDI]
		IMul	EAX,[ESI]								;Multiply by filter tap
		SAR	EAX,7
		Add	EBX,EAX									;Accumulate result

		Mov	EAX,[12+EDI]							;Right sample
		Sub	EAX,[4+EDI]
		IMul	dword [firDec]
		ShRD	EAX,EDX,16
		Add	EAX,[4+EDI]
		IMul	EAX,[ESI]
		SAR	EAX,7
		Add	ECX,EAX

		Mov	EDX,[firDec]							;Determine next sample to use in filter
		Add	EDX,[firRate]
		Mov	[firDec],DX
		ShR	EDX,16

		LEA	EDI,[EDX*8+EDI]						;EDI -> Sample to use in filter
		Sub	ESI,8										;ESI -> Next filter tap

	Dec	EBP
	JNZ	short %%Tap
	Pop	EBP,ESI

	MovSX	EAX,BX
	MovSX	EDX,CX
	And	EAX,~1
	And	EDX,~1

	Mov	EDI,[echoCur]								;Restore EDI
	Add	EDI,echoBuf
%endmacro


;In:
;   MM2 = Input samples
;
;Out:
;   MM2 = Filtered samples
;   EDI-> Echo buffer
;
;Destroys:
;   EAX,EDX,CL,MM5-7

%macro FIRFilterX 0
	Sub	byte [firCur],4
	Mov	EDI,[firCur]
	LEA	EDI,[EDI*2+firBuf]

	MovQ	[EDI],MM2
	MovQ	[FIRBUF*2+EDI],MM2

	Push	ESI
	PXOr	MM2,MM2
	XOr	EAX,EAX
	XOr	EDX,EDX
	Mov	ESI,firTaps+56
	Mov	CL,8
	%%Tap:
		MovD			MM7,EDX
		MovQ			MM5,[EDI]
		MovQ			MM6,[8+EDI]
		PUnpckLDQ	MM7,MM7
		PSRAW			MM5,1
		PSRAW			MM6,1
		PSRLD			MM7,1
		PSubW			MM6,MM5
		PMAddWD		MM6,MM7
		PAddD			MM6,MM6
		PSRLD			MM6,16
		PAddW			MM6,MM5
		PAddW			MM6,MM6

		PMAddWD		MM6,[ESI]						;Apply filter tap
		PSRAD			MM6,7
		PAddD			MM2,MM6

		Mov	EAX,EDX
		Add	EAX,[firRate]
		Mov	DX,AX
		ShR	EAX,16

		Sub	ESI,8
		LEA	EDI,[EAX*8+EDI]

	Dec	CL
	JNZ	short %%Tap
	Pop	ESI

	PSRAD			MM2,1									;Clamp to 17 bits
	PackSSDW		MM2,MM0
	PUnpckLWD	MM2,MM0
	PAddW			MM2,MM2								;Clear LSB

	Mov	EDI,[echoCur]
	Add	EDI,echoBuf
%endmacro


;In:
;   ST0,1 = Input samples
;
;Out:
;   ST0,1 = Filtered samples
;
;Destroys:
;   EAX,EDX,EBX,CL

%macro FIRFilterF 0
	Sub	byte [firCur],4
	Mov	EBX,[firCur]
	LEA	EBX,[EBX*2+firBuf]

	FSt	dword [EBX]
	FStP	dword [FIRBUF*2+EBX]
	FSt	dword [4+EBX]
	FStP	dword [FIRBUF*2+4+EBX]

	FLdZ
	FLdZ
	Mov	EDX,firTaps+56
	Mov	dword [firDec],0
	Mov	CL,8
	%%Tap:
		FILd	dword [firDec]
		FMul	dword [fpShR16]

		FLd	dword [8+EBX]
		FSub	dword [EBX]
		FMul	ST1
		FAdd	dword [EBX]
		FMul	dword [EDX]
		FAddP	ST2,ST

		FLd	dword [12+EBX]
		FSub	dword [4+EBX]
		FMulP	ST1,ST
		FAdd	dword [4+EBX]
		FMul	dword [EDX]
		FAddP	ST2,ST

		Mov	EAX,[firDec]
		Add	EAX,[firRate]
		Mov	[firDec],AX
		ShR	EAX,16

		Sub	EDX,8
		LEA	EBX,[EAX*8+EBX]

	Dec	CL
	JNZ	short %%Tap
%endmacro


%macro AAFilter 0
	Mov	EAX,[ESI+8]
	Sub	EAX,[ESI]
	IMul	EBP
	ShRD	EAX,EDX,16
	Add	EAX,[ESI]
	Mov	[EBX],EAX
	Mov	[LOWBUF*2+EBX],EAX

	Mov	EAX,[ESI+12]
	Sub	EAX,[ESI+4]
	IMul	EBP
	ShRD	EAX,EDX,16
	Add	EAX,[ESI+4]
	Mov	[EBX+4],EAX
	Mov	[LOWBUF*2+EBX+4],EAX

	Add	EBX,8
	And	EBX,~(LOWBUF*2)

	Push	ECX,ESI,EDI
	Mov	ECX,LOWTAPS-1
	XOr	ESI,ESI
	XOr	EDI,EDI
	%%Filter:
		Mov	EAX,[ECX*8+EBX]
		IMul	dword [ECX*4+lowTaps]
		Add	ESI,EDX

		Mov	EAX,[ECX*8+EBX+4]
		IMul	dword [ECX*4+lowTaps]
		Add	EDI,EDX

	Dec	ECX
	JNS	short %%Filter

	Mov	EAX,ESI
	Mov	EDX,EDI
	Pop	EDI,ESI,ECX
%endmacro


;
;Emulate DSP

PROC CatchUp

	Push	EAX

	Mov	EAX,[t64Cnt]
	ShR	EAX,1
	Sub	EAX,[outCnt]
	JZ		short .Done

	Add	[outCnt],EAX

	Push	EDX
	Mul	dword [outRate]
	Add	EAX,[outDec]
	AdC	EDX,0
	Mov	[outDec],AX
	ShRD	EAX,EDX,16
	Pop	EDX

	Sub	[outLeft],EAX
	JNC	short .Okay
		Add	EAX,[outLeft]
		Mov	dword [outLeft],0
	.Okay:
	Call	EmuDSP,[pOutBuf],EAX
	Mov	[pOutBuf],EAX

	.Done:
	Pop	EAX

ENDP


;--------------------------------------------
;Set Automatic EmuDSP Parameters

PROC SetEmuDSP, pBufD, numD, rateD

	Mov	EAX,[rateD]
	Test	EAX,EAX
	JZ		short .Final

		Push	ECX,EDX
		XOr	EDX,EDX
		ShLD	EDX,EAX,16
		ShL	EAX,16
		Mov	ECX,32000
		Div	ECX
		Mov	[outRate],EAX
		Pop	EDX,ECX

		Mov	EAX,[numD]
		Mov	[outLeft],EAX
		Mov	EAX,[pBufD]
		Mov	[pOutBuf],EAX
		Mov	EAX,[t64Cnt]
		ShR	EAX,1
		Mov	[outCnt],EAX
		Mov	dword [outDec],0
		RetN

	.Final:
		Call	EmuDSP,[pOutBuf],[outLeft]
		Mov	[pOutBuf],EAX
		Mov	dword [outLeft],0

ENDP


PROC EmuDSP, pBuf, num
USES ALL

	Mov	EAX,[pBuf]
	XOr	EBX,EBX										;EBX = 0 if output pointer is null, otherwise it indexes
	Test	EAX,EAX										;      the emulation routine
	SetZ	BL
	Dec	BL
	And	BL,[dspMix]

	.Next:

		;Verify output buffer length ----------
		Mov	EDX,[num]

		Test	EDX,EDX									;Is num > 0?
		JLE	.Quit										;  No, Quit

		Cmp	EDX,MIX_SIZE							;Is num <= size of internal buffer?
		JBE	short .NSmpOK
			Mov	EDX,MIX_SIZE
		.NSmpOK:

		Sub	[num],EDX

%ifdef SPC700_INC
		Test	byte [dbgOpt],SPC_HALT				;Do nothing if APU is suspended
		JNZ	short .Mute
%endif

		;Call emulation routine ---------------
		Test	byte [dspOpts],DSP_ANALOG
		JZ		short .NoLow
			Mov	[realOutS],EDX
			IMul	EDX,[lowRate]
			Add	EDX,[lowDec]
			ShR	EDX,16
		.NoLow:

		Call	[EBX*4+mixRout],EAX,EDX
		JC		short .Next								;Quit, if emulation produced output
	.Mute:
		;Output silence -----------------------
		XOr	EAX,EAX									;EDX = Size of output buffer in bytes
		MovSX	AX,[dspSize]
		XOr	AL,AH
		Sub	AL,AH
		Mul	byte [dspChn]
		IMul	EDX,EAX

		Cmp	byte [dspSize],1						;EAX = 80h if samples are unsigned, 0 otherwise
		SetNE	AL
		Dec	EAX
		And	EAX,80808080h

		Mov	ECX,EDX									;Fill output buffer with baseline samples
		And	EDX,3
		ShR	ECX,2

		test	EAX, EAX	; sexySPC fix
		jz short .NoSD		; of sorts...

		mov	long[0], 0
		Rep	StoSD		; Seg fault

		.NoSD
		Mov	ECX,EDX
		Rep	StoSB
		Mov	EAX,EDI

		Jmp	.Next

	.Quit:

	Push	EAX

	Call	SetFade

	;Update ENVX and OUTX registers ----------
	Mov	EBX,7*10h
	Mov	ESI,mix
	Mov	EDI,dsp
	.XRegs:
		Mov	EAX,[EBX*8+ESI+eVal]
		ShR	EAX,E_SHIFT
		Mov	[EBX+EDI+envx],AL

		Mov	AL,[EBX*8+ESI+mOut+1]
		Mov	[EBX+EDI+outx],AL

	Sub	BL,10h
	JAE	short .XRegs

	;Update DSP data register on SPC700 side -
	Mov	EBX,[pAPURAM]
	MovZX	EDX,byte [0F2h+EBX]
	Mov	DL,[EDX+dsp]
	Mov	[0F3h+EBX],DL

	Pop	EAX

ENDP


;
;Emulate DSP (Integer)
;
;Emulates the DSP of the SNES using standard 80386 instructions.
;Samples can be 8 or 16-bit.
;
;In:
;   pBufI-> Buffer to store output
;   numI  = Number of samples to create (1 - MIX_SIZE)
;
;Out:
;   CF  = Set, samples were created
;   EAX-> End of buffer
;
;   CF  = Clear, DSP is muted
;   EDI = pBufI
;   EDX = numI
;
;Destroys:
;   ECX,EDX,ESI,EDI

PROC EmuDSPI, pBufI, numI
LOCALS countI
USES EBX

	Mov	EAX,[numI]									;Save output size as counter
	Mov	[countI],EAX
	Mov	EDI,mixBuf									;EDI -> Temporary mixing buffer

ALIGN	16
	.NextEmu:

		;Generate Noise =======================
		NoiseGen

		;Voice Loop ===========================
		XOr	ECX,ECX
		Mov	EBX,mix									;EBX -> Internal mixing structure
		Mov	[EDI],ECX								;Erase current samples in buffer
		Mov	[4+EDI],ECX
		Mov	[8+EDI],ECX
		Mov	[12+EDI],ECX

		Mov	CH,1										;CH = Voice bitmask (start with first voice)
ALIGN	16
		.VoiceMix:
			XOr	EAX,EAX
			XOr	EDX,EDX

			Test	[voiceMix],CH						;Is the current voice active?
			JZ 	.VoiceDone							;  No, Goto next voice

			PitchMod										;Apply pitch modulation
			UpdateEnv									;Update envelope

			;Get sample ========================
			Mov	EAX,[nSmp]							;Load noise
			Test	[dsp+non],CH						;Is noise enabled?
			JNZ	short .Noise						;  Yes, Use noise sample
				Call	[pInter]
			.Noise:

			;Mixing ============================
			IMul	EAX,[EBX+eVal]						;Multiply envelope height
			SAR	EAX,E_SHIFT+7						;Remove the effects of envelope multiplication
			Mov	[EBX+mOut],EAX						;Store sample for pitch modulation

			Test	byte [EBX+mFlg],MFLG_MUTE		;Is voice muted by user?
			JNZ	.VoiceOff							;  Yes, Don't add to master
				Mov	EDX,EAX							;Duplicate sample for right side
				IMul	EAX,[EBX+mChnL]				;Apply left and right volumes
				IMul	EDX,[EBX+mChnR]
				SAR	EAX,7								;Remove the effects of volume multiplication
				SAR	EDX,7
				Add	[EDI],EAX						;Add to master samples
				Add	[4+EDI],EDX

				Test	[dsp+eon],CH					;Is echo turned on for this voice?
				JZ 	short .NoChEcho				;  No, Move to next voice
					Add	[8+EDI],EAX					;Add to master samples
					Add	[12+EDI],EDX
				.NoChEcho:
%if VMETERV
				;Save greatest sample output ----
				Mov	ESI,EDX							;Save right sample
				CDQ										;eax = abs(eax)
				XOr	EAX,EDX
				Sub	EAX,EDX

				Sub	EAX,[EBX+vMaxL]				;if (EAX>mix.vMaxL) mix.vMaxL=EAX
				CDQ
				Not	EDX
				And	EDX,EAX
				Add	[EBX+vMaxL],EDX

				Mov	EAX,ESI
				CDQ										;|Right|
				XOr	EAX,EDX
				Sub	EAX,EDX

				Sub	EAX,[EBX+vMaxR]
				CDQ
				Not	EDX
				And	EAX,EDX
				Add	[EBX+vMaxR],EAX
%endif
			.VoiceOff:

			UpdateSrc									;Update sample position

			.VoiceDone:
			Sub	EBX,-80h								;Increase base pointer

		Add	CH,CH										;Move bitmask over
		JNZ	.VoiceMix								;  No, Loop
		Add	EDI,16

	Dec	dword [countI]
	JNZ	.NextEmu


	;Create final output =====================
	Mov	EAX,[numI]
	Mov	[countI],EAX
	Mov	ESI,mixBuf

ALIGN	16
	NextSmp:
		;Multiply samples by main volume ------
		Mov	EAX,[ESI]								;Load main samples
		Mov	EDX,[4+ESI]
		IMul	EAX,[volMainL]							;Multiply by main volume
		IMul	EDX,[volMainR]
		SAR	EAX,7
		SAR	EDX,7
		Mov	[ESI],EAX								;Store samples back into memory
		Mov	[4+ESI],EDX

		;Echo ---------------------------------
		Test	byte [disEcho],-1						;Is echo enabled?
		JNZ	.NoEcho									;  No, Jump past expensive code
			Mov	EDI,[echoCur]
			Add	EDI,echoBuf							;EDI -> Delayed sample

			Mov	EAX,[EDI]							;Get current samples from echo buffer
			Mov	EDX,[4+EDI]

			;Filter echo -----------------------
			Test	byte [firEnabl],-1				;Does song use a filter?
			JZ 	.NoFilter							;  No, Skip more expensive code
				FIRFilterI
			.NoFilter:

			XOr	EBX,EBX
			Sub	dword [echoCur],8					;Move to next sample in echo buffer
			SetNC	BL
			Dec	EBX
			And	EBX,[echoDel]
			Add	[echoCur],EBX						;Start over, if end of buffer was reached

			Mov	ECX,EAX
			Mov	EBX,EDX
			IMul	EAX,[volEchoL]						;Multiply delayed samples by echo volume
			IMul	EDX,[volEchoR]
			SAR	EAX,7
			SAR	EDX,7
			Add	[ESI],EAX							;Add to main output
			Add	[4+ESI],EDX

%if STEREO
			Mov	EDX,EBX
			Mov	EAX,ECX
			IMul	ECX,[echoFB]						;Multiply delayed samples by feedback level
			IMul	EDX,[echoFBCT]						;Multiply delayed samples by crosstalk level
			IMul	EBX,[echoFB]
			IMul	EAX,[echoFBCT]
			Add	ECX,EDX								;Add crosstalk
			Add	EBX,EAX
%else
			IMul	ECX,[echoFB]
			IMul	EBX,[echoFB]
%endif
			SAR	ECX,7
			SAR	EBX,7
			Add	ECX,[8+ESI]							;Add delayed samples to echo samples
			Add	EBX,[12+ESI]
			Mov	[EDI],ECX							;Store feedback back into echo buffer
			Mov	[4+EDI],EBX
		.NoEcho:

%if VMETERM
		Mov	EAX,[ESI]								;See VMETERV above
		CDQ
		XOr	EAX,EDX
		Sub	EAX,EDX

		Sub	EAX,[vMMaxL]
		CDQ
		Not	EDX
		And	EAX,EDX
		Add	[vMMaxL],EAX

		Mov	EAX,[4+ESI]
		CDQ
		XOr	EAX,EDX
		Sub	EAX,EDX

		Sub	EAX,[vMMaxR]
		CDQ
		Not	EDX
		And	EAX,EDX
		Add	[vMMaxR],EAX
%endif

		Add	ESI,16

	Dec	dword [countI]
	JNZ	NextSmp

	;=========================================
	; Store output

	Test	byte [dsp+flg],40h						;Is the DSP muted?
	JNZ	.MuteI										;  Yes, Return silence

	Mov	EBX,32768
	Mov	ECX,[numI]
	Mov	[countI],ECX
	Mov	ESI,mixBuf
	Mov	EDI,[pBufI]

	Mov	AL,[dspSize]
	And	AL,1
	Mov	AH,[dspChn]
	And	AH,2
	Or		AL,AH

	Test	AL,3
	JZ		.R16M
	Test	AL,1
	JZ		.R16S
	Test	AL,2
	JZ		.R8M

	;8-bit Stereo ----------------------------
	.R8S:
		Mov	EAX,[ESI]
		Mov	EDX,[4+ESI]
		Add	EAX,EBX									;Unsign sample
		Add	EDX,EBX

		SHR	EAX,8										;Convert to 8-bit
		Test	AH,AH										;Is sample negative or more than 255?
		JZ 	short .SmpOKL
			SetS	AL
			Dec	AL										;AL = 0 or 255, based on sign
		.SmpOKL:

		SHR	EDX,8
		Test	DH,DH
		JZ 	short .SmpOKR
			SetS	DL
			Dec	DL
		.SmpOKR:

		Mov	AH,DL
		Mov	[EDI],AX
		Add	ESI,16
		Add	EDI,2

	Dec	ECX
	JNZ	short .R8S
	StC
	RetN	EDI

	;16-bit Stereo ---------------------------
	.R16S:
		Mov	EDX,[ESI]
		Mov	EAX,[4+ESI]

		;Clamp samples ------------------------
		LEA	ECX,[EDX+EBX]							;Convert sample to unsigned, for test
		SAR	ECX,16									;Is sample negative or more than 32767?
		JZ		short .SmpL								;  No
			Mov	EDX,EBX								;EDX = -32768
			SetS	DL										;Set DL based on sign
			Dec	EDX									;EDX = -32768 or 32767
		.SmpL:

		LEA	ECX,[EAX+EBX]
		SAR	ECX,16
		JZ		short .SmpR
			Mov	EAX,EBX
			SetS	AL
			Dec	EAX
		.SmpR:

		Mov	[EDI],DX									;Store samples
		Mov	[2+EDI],AX
		Add	ESI,16
		Add	EDI,4

	Dec	dword [countI]
	JNZ	short .R16S
	StC													;Set carry, since output was made
	RetN	EDI											;Return pointer to end of output

	;8-bit Mono ------------------------------
	.R8M:
		Mov	EAX,[ESI]								;Combine left and right
		Add	EAX,[4+ESI]
		IMul	EAX,181
		SAR	EAX,16									;Convert to 8-bit

		Sub	EAX,-128									;Test for clipping
		Test	AH,AH
		JZ 	short .Smp8OK
			SetS	AL
			Dec	AL
		.Smp8OK:

		Mov	[EDI],AL
		Add	ESI,16
		Inc	EDI

	Dec	ECX
	JNZ	short .R8M
	StC
	RetN	EDI

	;16-bit Mono -----------------------------
	.R16M:
		Mov	EAX,[ESI]
		Add	EAX,[4+ESI]								;Combine left and right samples
		IMul	EAX,181									;Reduce output by sqrt(2)/2
		SAR	EAX,8

		LEA	EDX,[EBX+EAX]							;Test for clipping
		SAR	EDX,16
		JZ 	short .Smp16OK
			Mov	EAX,8000h
			SetS	AL
			Dec	EAX
		.Smp16OK:

		Mov	[EDI],AX
		Add	ESI,16
		Add	EDI,2

	Dec	ECX
	JNZ	short .R16M
	StC
	RetN	EDI

.MuteI:
	Mov	EDI,[pBufI]
	Mov	EDX,[numI]
	ClC

ENDP


;
;Emulate DSP (MMX)
;
;Emulates the DSP of the SNES using instructions from the Multi-Media Extensions
;
;In:
;   pBufX-> Buffer to store output
;   numX  = Number of samples to create (1 - MIX_SIZE)
;
;Out:
;   CF  = Set, samples were created
;   EAX-> End of buffer
;
;   CF  = Clear, DSP is muted
;   EDI = pBufX
;   EDX = numX
;
;Destroys:
;   ECX,EDX,ESI,EDI
;   MM0 = 0
;   MM1 = Main
;   MM2 = Echo
;   MM3 = VMax for main output
;   MM4 = temp
;   MM5 = temp
;   MM6 = temp
;   MM7 = temp

PROC EmuDSPX, pBufX, numX
LOCALS countX
USES EBX


	PXOr	MM0,MM0										;MM0 = 0, always

%if VMETERM
	MovQ	MM3,[vMMaxL]
%endif

	;=========================================
	;Mix voices

	Mov	EAX,[numX]
	Test	EAX,EAX
	JZ		.Store
	Mov	[countX],EAX

	XOr	EAX,EAX
	Test	byte [dspOpts],DSP_ANALOG
	SetNZ	AL
	ShL	EAX,6
	LEA	EDI,[EAX+mixBuf]

ALIGN	16
	.NextEmu:

		;Generate noise -----------------------
		NoiseGen

		;Process single voice -----------------
		XOr	ECX,ECX
		Mov	EBX,mix
		PXOr	MM1,MM1
		PXOr	MM2,MM2

		Mov	CH,1
ALIGN	16
		.VoiceMix:
			XOr	EAX,EAX
			XOr	EDX,EDX

			Test	[voiceMix],CH
			JZ 	.VoiceDone

			PitchMod										;Apply pitch modulation
			UpdateEnv									;Update envelope

			;Get sample ------------------------
			Mov	EAX,[nSmp]
			Test	[dsp+non],CH
			JNZ	short .Noise
				Call	[pInter]
			.Noise:

			;Mixing ----------------------------
			IMul	EAX,[EBX+eVal]
			SAR	EAX,E_SHIFT+7+1
			Add	EAX,EAX
			Mov	[EBX+mOut],EAX

			Test	byte [EBX+mFlg],MFLG_MUTE
			JNZ	.VoiceOff
				MovZX			EAX,AX					;Truncate sample to 16-bits and dupilcate in MM5
				MovD			MM4,EAX
				PUnpckLDQ	MM4,MM4

				PMAddWD		MM4,[EBX+mChnL]		;Multiply by channel volume
				PSRAD			MM4,7
%if VMETERV
				;Save greatest sample output ----
				MovQ		MM7,[EBX+vMax]				;Get max voice output so far
				MovQ		MM6,MM4						;Copy current voice output
				MovQ		MM5,MM4

				PSRAD		MM6,31						;Create a mask from negative samples
				PXOr		MM5,MM6						;Invert negative samples
				PSubD		MM5,MM6						;Add 1 to negatives, |MM5|

				MovQ		MM6,MM7
				PCmpGTD	MM7,MM5						;Create a mask from greater samples
				PAnd		MM6,MM7						;Save greater samples
				PAndN		MM7,MM5						;Clear lesser samples
				POr		MM7,MM6						;MM7 = Greatest output so far
				MovQ		[EBX+vMax],MM7
%endif
				PAddSW		MM1,MM4					;Add to main output

				Test	[dsp+eon],CH
				JZ 	short .NoChEcho
					PAddSW		MM2,MM4				;Add to echo
				.NoChEcho:
			.VoiceOff:

			UpdateSrc									;Update sample position

			.VoiceDone:
			Sub	EBX,-80h

		Add	CH,CH
		JNZ	.VoiceMix

		PSRAW	MM1,1
		PAddW	MM1,MM1
		PAnd	MM1,[loSmp]
		PAnd	MM2,[loSmp]
		MovQ	[EDI],MM1
		MovQ	[8+EDI],MM2

		Add	EDI,16

	Dec	dword [countX]
	JNZ	.NextEmu


	;=========================================
	;Apply main volumes and mix in echo

	Mov	EAX,[numX]
	Mov	[countX],EAX

	XOr	EAX,EAX
	Test	byte [dspOpts],DSP_ANALOG
	SetNZ	AL
	ShL	EAX,6
	LEA	ESI,[EAX+mixBuf]

ALIGN	16
	.NextSmp:
		;Multiply samples by main volume ------
		MovQ			MM1,[ESI]
		PMAddWD		MM1,[volMainL]
		PSRAD			MM1,7

		Test	byte [disEcho],-1
		JNZ	.NoEcho
			Mov			EDI,[echoCur]
			Add			EDI,echoBuf

			MovQ			MM2,[EDI]					;Load MM2 with echo samples incase filtering is off

			;Filter echo -----------------------
			FIRFilterX

			XOr			EAX,EAX
			Sub			dword [echoCur],8
			SetNC			AL
			Dec			EAX
			And			EAX,[echoDel]
			Add			[echoCur],EAX

			MovQ			MM5,MM2
			PMAddWD		MM5,[volEchoL]
			PSRAD			MM5,7
			PAddD			MM1,MM5

%if STEREO
			MovQ			MM7,MM2
			PSRLQ			MM7,32						;MM7 = PSwapD MM2
			PUnpckLDQ	MM7,MM2
			PMAddWD		MM2,[echoFB]				;Multiply old samples by feedback
			PMAddWD		MM7,[echoFBCT]
			PAddD			MM2,MM7
%else
			PMAddWD		MM2,[echoFB]
%endif
			PSRAD			MM2,7
			PAddSW		MM2,[8+ESI]					;Add current samples
			PSRAW			MM2,1
			PAddW			MM2,MM2
			PAnd			MM2,[loSmp]
			MovQ			[EDI],MM2
		.NoEcho:

%if VMETERM
		MovQ		MM6,MM1								;See VMETERV above
		MovQ		MM7,MM1

		PSRAD		MM6,31
		PXOr		MM7,MM6
		PSubD		MM7,MM6

		MovQ		MM6,MM3
		PCmpGTD	MM3,MM7
		PAnd		MM6,MM3
		PAndN		MM3,MM7
		POr		MM3,MM6
%endif

		PackSSDW		MM1,MM0
		MovD			[ESI],MM1
		Add			ESI,16

	Dec	dword [countX]
	JNZ	.NextSmp

%if VMETERM
	MovQ	[vMMaxL],MM3
%endif

	;=========================================
	; Store output

.Store:
	Test	byte [dsp+flg],40h
	JNZ	MuteX

	Mov	ESI,mixBuf
	Mov	EDI,[pBufX]
	Mov	ECX,[numX]

	Test	byte [dspOpts],DSP_ANALOG
	JNZ	.UseLow

	Cmp	byte [dspSize],2
	JE		.R16
	Cmp	byte [dspSize],3
	JE		.R24
	Cmp	byte [dspSize],4
	JE		.R32

	.RF:
		FILd	word [ESI]
		FMul	dword [fpShR15]
		FILd	word [2+ESI]
		FMul	dword [fpShR15]
		FStP	dword [4+EDI]
		FStP	dword [EDI]

		Add	ESI,16
		Add	EDI,8

	Dec	ECX
	JNZ	short .RF
	EMMS
	StC
	RetN	EDI

	.R16:
		Mov	EAX,[ESI]
		Mov	[EDI],EAX

		Add	ESI,16
		Add	EDI,4

	Dec	ECX
	JNZ	short .R16
	EMMS
	StC
	RetN	EDI

	.R24:
		MovZX	EAX,word [ESI]
		Mov	DX,[2+ESI]
		ShL	EAX,8
		Mov	[EDI],EAX
		Mov	[4+EDI],DX

		Add	ESI,16
		Add	EDI,6

	Dec	ECX
	JNZ	short .R24
	EMMS
	StC
	RetN	EDI

	.R32:
		MovQ			MM1,MM0
		PUnpckLWD	MM1,[ESI]
		MovQ			[EDI],MM1

		Add	ESI,16
		Add	EDI,8

	Dec	ECX
	JNZ	short .R32
	EMMS
	StC
	RetN	EDI

.UseLow:
	Test	ECX,ECX
	JZ		.NoUnpack

		Mov			EAX,17491						;Initialize white noise generator
		Mov			EDX,1999
		MovD			MM6,EAX
		MovD			MM7,EDX
		PUnpckLDQ	MM6,MM6
		PUnpckLDQ	MM7,MM7
		MovQ			MM5,[lowRFI]

		Mov			EBX,[lowAmp]

		Mov			EAX,07E87A09h					;Clamp to -0.1dB (28-bits signed)
		MovD			MM4,EAX
		PUnpckLDQ	MM4,MM4

%if VMETERM
		MovQ			MM3,[vMMaxL]
		PSLLD			MM3,12
%endif

		Mov	EDI,32
		.Unpack:
			PUnpckLWD	MM1,[EDI*2+ESI]			;Get 16-bit samples

			PSRAD			MM1,3							;Add -72dB of white noise
			PMulLW		MM5,MM6
			PAddW			MM5,MM7
			PAddD			MM1,MM5
			PSRAD			MM1,9							;Reduce to 20 bits

			MovD			EAX,MM1						;Multiply by volAmp
			PSRLQ			MM1,32
			MovD			EDX,MM1
			IMul			EAX,EBX
			IMul			EDX,EBX
			MovD			MM1,EAX
			MovD			MM2,EDX
			PUnpckLDQ	MM1,MM2

%if VMETERM
			MovQ		MM2,MM1
			PSRAD		MM2,31
			PXOr		MM2,MM1

			MovQ		MM0,MM3
			PCmpGTD	MM3,MM2
			PAnd		MM0,MM3
			PAndN		MM3,MM2
			POr		MM3,MM0
%endif

			;Clamp samples ---------------------
			PAddD		MM1,MM4							;Unsign samples
			PSRAD		MM1,1

			MovQ		MM2,MM1							;Clamp < 0
			PSRAD		MM1,31
			PAndN		MM1,MM2

			MovQ		MM2,MM4							;Clamp > MM4
			PCmpGTD	MM2,MM1
			PAnd		MM1,MM2
			PAndN		MM2,MM4
			POr		MM1,MM2

			PAddD		MM1,MM1							;Sign samples
			PSubD		MM1,MM4
			PSLLD		MM1,3

			MovQ		[EDI+ESI],MM1

			Add	EDI,8

		Dec	ECX
		JNZ	short .Unpack

		Mov	EDI,[pBufX]
		MovQ	[lowRFI],MM5

%if VMETERM
		PSRAD	MM3,12
		MovQ	[vMMaxL],MM3

		PXOr	MM0,MM0
%endif

	.NoUnpack:

	FInit

	Push	EBP
	Mov	ECX,[realOutS]
	Mov	EBP,[lowDec]
	Mov	EBX,[lowCur]

	Cmp	byte [dspSize],2
	JE		.RL16
	Cmp	byte [dspSize],3
	JE		.RL24
	Cmp	byte [dspSize],4
	JE		.RL32

	.RLF:
		AAFilter

		Mov	[ESP-4],EAX
		Mov	[ESP-8],EDX
		FILd	dword [ESP-4]
		FILd	dword [ESP-8]
		FMul	dword [fpShR30]
		FStP	dword [4+EDI]
		FMul	dword [fpShR30]
		FStP	dword [EDI]
		Mov	[EDI],EAX
		Mov	[4+EDI],EDX

		XOr	EAX,EAX
		Add	BP,[lowRate]
		AdC	AL,[2+lowRate]
		LEA	ESI,[EAX*8+ESI]
		Add	EDI,8

	Dec	ECX
	JNZ	.RLF
	Jmp	.DoneL

	.RL16:
		AAFilter

		ShR	EAX,14
		ShR	EDX,14
		Mov	[EDI],AX
		Mov	[2+EDI],DX

		XOr	EAX,EAX
		Add	BP,[lowRate]
		AdC	AL,[2+lowRate]
		LEA	ESI,[EAX*8+ESI]
		Add	EDI,4

	Dec	ECX
	JNZ	.RL16
	Jmp	.DoneL

	.RL24:
		AAFilter

		ShL	EAX,2
		ShR	EDX,6
		ShRD	EAX,EDX,8
		ShR	EDX,8
		Mov	[EDI],EAX
		Mov	[4+EDI],DX

		XOr	EAX,EAX
		Add	BP,[lowRate]
		AdC	AL,[2+lowRate]
		LEA	ESI,[EAX*8+ESI]
		Add	EDI,6

	Dec	ECX
	JNZ	.RL24
	Jmp	.DoneL

	.RL32:
		AAFilter

		ShL	EAX,2
		ShL	EDX,2
		Mov	[EDI],EAX
		Mov	[4+EDI],EDX

		XOr	EAX,EAX
		Add	BP,[lowRate]
		AdC	AL,[2+lowRate]
		LEA	ESI,[EAX*8+ESI]
		Add	EDI,8

	Dec	ECX
	JNZ	.RL32

.DoneL:
	Mov	[lowCur],EBX
	Mov	[lowDec],EBP
	Pop	EBP

	Mov	EDX,[numX]
	ShL	EDX,3
	JZ		short .Blah
		MovQ	MM1,[EDX+mixBuf]
		MovQ	MM2,[8+EDX+mixBuf]
		MovQ	MM3,[16+EDX+mixBuf]
		MovQ	MM4,[24+EDX+mixBuf]
		MovQ	[mixBuf],MM1
		MovQ	[8+mixBuf],MM2
		MovQ	[16+mixBuf],MM3
		MovQ	[24+mixBuf],MM4
	.Blah:

	EMMS													;Empty MMX regs to keep FP operations from breaking
	StC
	RetN	EDI

MuteX:
	Mov	EDI,[pBufX]
	Mov	EDX,[numX]
	EMMS
	ClC

ENDP


;
;Emulate DSP (Float)
;
;Emulates the DSP of the SNES using floating-point instructions
;
;In:
;   pBufF-> Buffer to store output
;   numF  = Number of samples to create (1 - MIX_SIZE)
;
;Out:
;   CF  = Set, samples were created
;   EAX-> End of buffer
;
;   CF  = Clear, DSP is muted
;   EDI = pBufF
;   EDX = numF
;
;Destroys:
;   ECX,EDX,ESI,EDI,ST0-ST7

PROC EmuDSPF, pBufF, numF
LOCALS countF
USES EBX

	FInit

	FILd	dword [vMMaxL]
	FMul	dword [fp64k]								;Convert to a 32-bit sample (<< 16)
	FStP	dword [vMMaxL]								;Save as a float
	FILd	dword [vMMaxR]
	FMul	dword [fp64k]
	FStP	dword [vMMaxR]

	;=========================================
	; Mix voices

	Mov	EAX,[numF]
	Mov	[countF],EAX
	Mov	EDI,mixBuf
ALIGN	16
	.NextEmu:

		;Generate Noise =======================
		NoiseGen

		;Voice Loop ===========================
		XOr	ECX,ECX
		XOr	EAX,EAX
		Mov	EBX,mix
		Mov	[EDI],EAX
		Mov	[4+EDI],EAX
		Mov	[8+EDI],EAX
		Mov	[12+EDI],EAX

		Mov	CH,1
ALIGN	16
		.VoiceMix:
			XOr	EAX,EAX
			XOr	EDX,EDX

			Test	[voiceMix],CH
			JZ 	.VoiceDone

			PitchMod										;Apply pitch modulation
			UpdateEnv									;Update envelope

			;Get sample ========================
			Call	[pInter]

			Test	[dsp+non],CH
			JZ		short .NoNoise
				FStP	ST
				FILd	dword [nSmp]
			.NoNoise:

			;Mixing ============================
			FIMul	dword [EBX+eVal]
			FMul	dword [fpEShR]
			FISt	dword [EBX+mOut]

			Test	byte [EBX+mFlg],MFLG_MUTE
			JNZ	.VoiceOff
%if STEREO
				Mov	EAX,[EBX+mTgtL]
				Cmp	EAX,[EBX+mChnL]
				JZ		short .NoRampL
					SetG	DL
					Test	EAX,[EBX+mChnL]
					SetS	DH
					XOr	DL,DH
					JZ		short .SubL

						FLd	dword [EBX+mChnL]
						FAdd	dword [volRamp]
						FStP	dword [EBX+mChnL]
						Cmp	EAX,[EBX+mChnL]
						SetL	DL
						XOr	DL,DH
						JZ		short .NoRampL

						Mov	[EBX+mChnL],EAX
						Jmp	short .NoRampL

					.SubL:
						FLd	dword [EBX+mChnL]
						FSub	dword [volRamp]
						FStP	dword [EBX+mChnL]
						Cmp	EAX,[EBX+mChnL]
						SetG	DL
						XOr	DL,DH
						JZ		short .NoRampL

						Mov	[EBX+mChnL],EAX

				.NoRampL:

				Mov	EAX,[EBX+mTgtR]
				Cmp	EAX,[EBX+mChnR]
				JZ		short .NoRampR
					SetG	DL
					Test	EAX,[EBX+mChnR]
					SetS	DH
					XOr	DL,DH
					JZ		short .SubR

						FLd	dword [EBX+mChnR]
						FAdd	dword [volRamp]
						FStP	dword [EBX+mChnR]
						Cmp	EAX,[EBX+mChnR]
						SetL	DL
						XOr	DL,DH
						JZ		short .NoRampR

						Mov	[EBX+mChnR],EAX
						Jmp	short .NoRampR

					.SubR:
						FLd	dword [EBX+mChnR]
						FSub	dword [volRamp]
						FStP	dword [EBX+mChnR]
						Cmp	EAX,[EBX+mChnR]
						SetG	DL
						XOr	DL,DH
						JZ		short .NoRampR

						Mov	[EBX+mChnR],EAX

				.NoRampR:
%endif

%if VMETERV
				Sub	ESP,8								;Create a temporary stack space for samples
%endif
				FLd	ST
				Test	[dsp+eon],CH
				JNZ 	short .VoiceEcho

					FMul	dword [EBX+mChnL]
%if VMETERV
					FISt	dword [ESP]					;Store sample as an integer
%endif
					FAdd	dword [EDI]
					FStP	dword [EDI]

					FMul	dword [EBX+mChnR]
%if VMETERV
					FISt	dword [4+ESP]
%endif
					FAdd	dword [4+EDI]
					FSt	dword [4+EDI]
					Jmp	short .NoVoiceEcho

				.VoiceEcho:
					FMul	dword [EBX+mChnL]
%if VMETERV
					FISt	dword [ESP]
%endif
					FLd	ST
					FAdd	dword [EDI]
					FStP	dword [EDI]
					FAdd	dword [8+EDI]
					FStP	dword [8+EDI]

					FMul	dword [EBX+mChnR]
%if VMETERV
					FISt	dword [4+ESP]
%endif
					FLd	ST
					FAdd	dword [4+EDI]
					FStP	dword [4+EDI]
					FAdd	dword [12+EDI]
					FSt	dword [12+EDI]

				.NoVoiceEcho:
%if VMETERV
				;Save greatest sample output ----
				Pop	EAX								;Pop left sample off stack
				CDQ
				XOr	EAX,EDX
				Sub	EAX,EDX

				Sub	EAX,[EBX+vMaxL]
				CDQ
				Not	EDX
				And	EAX,EDX
				Add	[EBX+vMaxL],EAX

				Pop	EAX								;Pop right sample off stack
				CDQ
				XOr	EAX,EDX
				Sub	EAX,EDX

				Sub	EAX,[EBX+vMaxR]
				CDQ
				Not	EDX
				And	EAX,EDX
				Add	[EBX+vMaxR],EAX
%endif
			.VoiceOff:
			FStP	ST

			UpdateSrc									;Update sample position

			.VoiceDone:
			Sub	EBX,-80h

		ShL	CH,1
		JNC	.VoiceMix

		Add	EDI,16

	Dec	dword [countF]
	JNZ	.NextEmu


	;=========================================
	; Apply main volumes and mix in echo

	Mov	EAX,[numF]
	Mov	[countF],EAX
	Mov	ESI,mixBuf
ALIGN	16
	.NextSmp:
		;Multiply samples by main volume ------
		FLd	dword [ESI]
		FMul	dword [volMainL]
		FStP	dword [ESI]
		FLd	dword [4+ESI]
		FMul	dword [volMainR]
		FStP	dword [4+ESI]

		Test	byte [disEcho],-1
		JNZ	.NoEcho
			Mov	EDI,[echoCur]
			Add	EDI,echoBuf

			FLd	dword [4+EDI]						;FBR
			FLd	dword [EDI]							;FBR FBL

			;Filter echo -----------------------
			FIRFilterF
			FLd	ST1									;FBR FBL FBR
			FLd	ST1									;FBR FBL FBR FBL

			;Advance echo sample pointer -------
			XOr	EAX,EAX
			Sub	dword [echoCur],8
			SetNC	AL
			Dec	EAX
			And	EAX,[echoDel]
			Add	[echoCur],EAX

			;Add echo to main output -----------
			FMul	dword [volEchoL]					;FBR FBL FBR FBL*EchoL
			FAdd	dword [ESI]							;FBR FBL FBR EchoL+ML
			FStP	dword [ESI]							;FBR FBL FBR

			FMul	dword [volEchoR]					;FBR FBL FBR*EchoR
			FAdd	dword [4+ESI]						;FBR FBL FBR+MR
			FStP	dword [4+ESI]						;FBR FBL

			;Calculate echo feedback -----------
%if STEREO
			FLd	ST										;FBR FBL FBL
			FMul	dword [echoFB]						;FBR FBL FBL*EchoFB
			FLd	ST2									;FBR FBL EFBL FBR
			FMul	dword [echoFBCT]					;FBR FBL EFBL FBR*EchoFBCT
			FAddP	ST1,ST								;FBR FBL EFBL+EFBCR
			FAdd	dword [8+ESI]						;FBR FBL EFBL+EL
			FStP	dword [EDI]							;FBR FBL

			FMul	dword [echoFBCT]					;FBR FBL*EchoFBCT
			FXCh	ST1									;EFBCL FBR
			FMul	dword [echoFB]						;EFBCL FBR*EchoFB
			FAddP	ST1,ST								;EFBCL+EFBR
			FAdd	dword [12+ESI]						;EFBR+ER
			FStP	dword [4+EDI]						;
%else
			FMul	dword [echoFB]
			FAdd	dword [8+ESI]
			FStP	dword [EDI]

			FMul	dword [echoFB]
			FAdd	dword [12+ESI]
			FStP	dword [4+EDI]
%endif
		.NoEcho:

%if VMETERM
		Mov	EAX,[ESI]								;EAX = |Left|
		And	EAX,7FFFFFFFh

		Sub	EAX,[vMMaxL]							;*** Positive floats can be operated on as integers ***
		CDQ
		Not	EDX
		And	EAX,EDX
		Add	[vMMaxL],EAX

		Mov	EAX,[4+ESI]								;EAX = |Right|
		And	EAX,7FFFFFFFh

		Sub	EAX,[vMMaxR]
		CDQ
		Not	EDX
		And	EAX,EDX
		Add	[vMMaxR],EAX
%endif

		Add	ESI,16

	Dec	dword [countF]
	JNZ	.NextSmp


	;=========================================
	; Store output

	Test	byte [dsp+flg],40h
	JNZ	MuteF

	Mov	ESI,mixBuf
	Mov	EDI,[pBufF]
	Mov	ECX,[numF]

	Cmp	byte [dspSize],-4
	JE		.OutFloat

	Mov	[countF],ECX
	Mov	ECX,4EFFFE00h								;ECX = 2147418112.0 (32767 << 16)
	.NextInt:
		;Clamp samples ------------------------
		Mov	EAX,[ESI]								;EAX = Sample
		XOr	EDX,EDX
		XOr	EBX,EBX
		BTR	EAX,31									;EAX = Absolute value
		RCR	EDX,1										;EDX = Sign of sample
		Sub	EAX,ECX
		SetA	BL											;EBX = -1 if EAX < ECX
		Dec	EBX
		And	EAX,EBX									;Clamp EAX
		Add	EAX,ECX
		Or		EAX,EDX									;Restore sign
		Mov	[ESI],EAX
		FLd	dword [ESI]

		Mov	EAX,[4+ESI]
		XOr	EDX,EDX
		XOr	EBX,EBX
		BTR	EAX,31
		RCR	EDX,1
		Sub	EAX,ECX
		SetA	BL
		Dec	EBX
		And	EAX,EBX
		Add	EAX,ECX
		Or		EAX,EDX
		Mov	[ESI],EAX
		FLd	dword [ESI]

		;Reduce to integer form ---------------
		Cmp	byte [dspSize],4
		JE		short .Out32
		Cmp	byte [dspSize],3
		JE		short .Out24

			FMul	dword [fpShR16]
			FIStP	word [2+EDI]
			FMul	dword [fpShR16]
			FIStP	word [EDI]
			Add	EDI,4
			Jmp	short .Out16

		.Out32:
			FIStP	dword [4+EDI]
			FIStP	dword [EDI]
			Add	EDI,8
			Jmp	short .Out16

		.Out24:
			FIStP	dword [ESP-4]
			FIStP	dword [ESP-7]
			Mov	DX,[ESP-6]
			Mov	EAX,[ESP-4]
			Mov	[0+EDI],DX
			Mov	[2+EDI],EAX
			Add	EDI,6

		.Out16:

;		;*** Slower floating-point method for clamping ***
;		FLd		[fp2M]								;Load clamping range      |Fp2M
;
;		FLd		dword [ESI]							;Load left sample         |Fp2M Left
;		FLd		ST										;                         |Fp2M Left Left
;		FAbs												;Get absolute value       |Fp2M Left |Left|
;		FComP		ST2									;Cmp |Left|,Fp2M and pop  |Fp2M Left
;		FNStSW	AX										;Store FPU flags in AX
;		SAHF												;Store AH in x86 flags
;		JBE		short .SmpL
;			FStP	[fTemp]								;                         |Fp2M
;			FLd	ST										;                         |Fp2M Fp2M
;			Test	byte [3+ESI],80h					;Is sample negative?
;			JZ		short .SmpL							;  No, Leave ST positive
;				FChS										;  Yes, Negate ST         |Fp2M -Fp2M
;		.SmpL:
;		FIStP		dword [ESI]							;Save sample as an int    |Fp2M
;		Mov		EDX,[ESI]
;
;		FLd		dword [4+ESI]
;		FLd		ST
;		FAbs
;		FComP		ST2
;		FNStSW	AX
;		SAHF
;		JBE		short .SmpR
;			FStP	[fTemp]
;			FLd	ST
;			Test	byte [7+ESI],80h
;			JZ		short .SmpR
;				FChS
;		.SmpR:
;		FIStP		dword [4+ESI]
;		Mov		EAX,[4+ESI]
;		FStP		[fTemp]								;Pop clipping range
;
;		;Store output -------------------------
;		Mov	[EDI],EDX
;		Mov	[4+EDI],EAX

		Add	ESI,16

	Dec	dword [countF]
	JNZ	.NextInt
	Jmp	.Done

	;32-bit floating-point -------------------
	.OutFloat:
		Mov	EAX,[ESI]
		Mov	EDX,[4+ESI]
		Mov	[EDI],EAX
		Mov	[4+EDI],EDX

		Add	ESI,16
		Add	EDI,8

	Dec	ECX
	JNZ	.OutFloat

.Done:
	FLd	dword [vMMaxL]								;Convert _vMMax back to integer form
	FMul	dword [fpShR16]
	FIStP	dword [vMMaxL]
	FLd	dword [vMMaxR]
	FMul	dword [fpShR16]
	FIStP	dword [vMMaxR]
	StC
	RetN	EDI

MuteF:
	FLd	dword [vMMaxL]
	FMul	dword [fpShR16]
	FIStP	dword [vMMaxL]
	FLd	dword [vMMaxR]
	FMul	dword [fpShR16]
	FIStP	dword [vMMaxR]
	Mov	EDI,[pBufF]
	Mov	EDX,[numF]
	ClC

ENDP


;
;Emulate DSP (no mixing)
;
;Emulates the DSP of the SNES, except for pitch modulation, noise generator, and mixing.
;
;Notes:
;   OutBuf can be null, in which case no output is created.  If OutBuf is not null, silence will be
;     returned.
;
;In:
;   outBuf -> Buffer that would store output
;   outSize = Number of samples to not create (1-3072)
;
;Out:
;   EAX = outBuf
;   CF  = True, if EAX is null
;
;Destroys:
;   ECX,EDX,ESI,EDI

ALIGN 16
PROC EmuDSPN, pBufN, numN
LOCALS countN
USES EBX

	Mov	EAX,[numN]
	Mov	[countN],EAX

	Mov	EBX,mix
	Mov	CH,1											;CH = Voice bitmask
ALIGN	16
	.NextVoice:
		;Voice Loop ---------------------------
		Test	[voiceMix],CH							;Is the current voice on?
		JZ 	.VoiceDone								;  No

			;Figure new sample position ========
			Mov	EAX,[EBX+mRate]
			Mul	dword [countN]
			Add	[EBX+mDec],AX
			SetC	CL
			ShRD	EAX,EDX,16
			MovZX	EDX,CL
			Add	EAX,EDX
			Mov	EDX,EAX
			ShR	EDX,4									;EDX = Number of sample blocks to increase
			And	EAX,0Fh
			Add	EAX,EAX								;EAX = Position in sample RAM

			Add	[EBX+sIdx],AL						;Increase current index
			Test	byte [EBX+sIdx],20h				;Have we moved into the next block?
			JZ		short .NoSInc						;  No
				And	byte [EBX+sIdx],~20h			;Correct index in
				Inc	EDX
			.NoSInc:

			Test	EDX,EDX								;Do any blocks need to be decompressed?
			JZ		.NoBInc								;  No

			Mov	ESI,[EBX+bCur]						;ESI -> Current block
			.NextBlk:
				Mov	AL,[ESI]

				Test	AL,1								;Is this the end block?
				JZ		short .NotEnd					;  Not yet
				Or 	[dsp+endx],CH					;Set end flag
				Test	AL,2
				JNZ	short .LoopB

					Not	CH								;Turn off voice
					And	[voiceMix],CH
					Not	CH
					Mov	dword [EBX+eVal],0
					Mov	dword [EBX+mOut],0
					Or 	byte [EBX+mFlg],MFLG_OFF
					And	byte [EBX+mFlg],~MFLG_KOFF
					Jmp	.VoiceDone

				.LoopB:
					Mov	EDI,[EBX+dspv]
					Mov	EAX,[pAPURAM]
					MovZX	ESI,byte [EDI+srcn]
					Mov	AH,[dsp+dir]
					Mov	AX,[2+ESI*4+EAX]
					Mov	ESI,EAX

				.NotEnd:

				Push	EDX,EBX							;Decompress block
				LEA	EDI,[EBX+sBuf]
				Mov	EDX,[EBX+sP1]
				Mov	EBX,[EBX+sP2]
				Call	[pDecomp]
				Mov	EAX,EBX
				Pop	EBX
				Mov	[EBX+sP1],EDX
				Pop	EDX
				Mov	[EBX+sP2],EAX
			Dec	EDX
			JNZ	short .NextBlk

			Sub	SI,9									;Save block header and pointer for last block decomp
			Mov	AL,[ESI]
			Mov	[EBX+bCur],ESI
			Mov	[EBX+bHdr],AL
			.NoBInc:

			;Adjust envelope height ============
			Mov	EDI,[countN]
ALIGN	16
			.VoiceEnv:
				Mov	CL,[EBX+eMode]
				Test	CL,E_IDLE						;Is envelope constant?
				JNZ	.EnvDone							;  Yes, Skip height adjustment

				Mov	EAX,EDI
				Cmp	DI,[2+EBX+eCnt]				;EAX = min(eCnt, countN)
				JB		short .EmuS
					Mov	AX,[2+EBX+eCnt]
				.EmuS:
				Sub	EDI,EAX

				Dec	EAX
				Sub	[2+EBX+eCnt],AX
				UpdateEnv								;Calculate envelope change

			Test	EDI,EDI
			JNZ	.VoiceEnv
			.EnvDone:

         ;Update ENVX and OUTX registers ====
			Mov	ESI,dword [EBX+sIdx]
			Mov	EDX,[EBX+eVal]						;EDX = Envelope height
			MovSX	EAX,word [ESI]						;EAX = Current sample
			IMul	EAX,EDX
			ShR	EDX,E_SHIFT							;Reduce height to 7-bit number
			SAR	EAX,E_SHIFT+7						;Reduce sample to 16-bit number
			Mov	[EBX+mOut],EAX

			Mov	ESI,EBX
			Sub	ESI,mix
			ShR	ESI,3									;ESI indexes dsp.voice
			SAR	EAX,8
			Mov	[ESI+dsp+envx],DL
			Mov	[ESI+dsp+outx],AL
		.VoiceDone:
		Sub	EBX,-80h									;Increase base pointer

	ShL	CH,1											;Move bitmask over
	JNC	.NextVoice

	Mov	EAX,[pBufN]                       ;Set carry if OutBuf is null, so EmuDSP doesn't crash
	Cmp	EAX,1

ENDP


;
;Play Waveform

PROC PlayWave, pSrcw, loopw, ratew
USES ALL

	Mov	EAX,[ratew]
	Test	EAX,EAX
	JNZ	.New

	Mov	EAX,[tBlk]
	Test	EAX,EAX
	JZ		.Done
	FInit

	.Next:

		;=====================================
		;Fill mixBuf

		Mov	ECX,2*2*MIX_SIZE
		Mov	EDI,mixBuf

		Sub	[loopw],ECX
		JGE	short .SizeOK
			Add	ECX,[loopw]
		.SizeOK:
		Mov	[ratew],ECX

		.NextSmp:
			Mov	DL,[2+tRate]
			Mov	EAX,[tRate]
			Add	[tDec],AX
			AdC	DL,0
			JZ 	.NoSInc

			Add	DL,DL
			Add	[tIdx],DL
			Test	byte [tIdx],20h
			JZ		.NoSInc

			Mov	AX,[2+pAPURAM]						;Move to next BRR block
			Cmp	AX,[2+tBlk]							;If decompressing BRR blocks in APU RAM, make sure
			JNE	short .NotInRAM					; pointer wraps around at 64k boundary
				Add	word [tBlk],9
				Jmp	short .InRAM
			.NotInRAM:
				Add	dword [tBlk],9
			.InRAM:

			Sub	byte [tIdx],20h
			Mov	EAX,[tBuf+24]
			Mov	EDX,[tBuf+28]
			Mov	[tBuf-8],EAX
			Mov	[tBuf-4],EDX
			Mov	ESI,[tBlk]

			Test	byte [ESI-9],1
			JZ 	short .NotEndB
			Test	byte [ESI-9],2
			JNZ	short .LoopB

				Sub	[ratew],ECX
				JZ		.Done
				Mov	dword [tBlk],0
				Mov	dword [loopw],0
				Jmp	.Reduce

			;Restart sample loop ---------------
			.LoopB:
				Mov	ESI,[tLoop]
				Mov	[tBlk],ESI

			;Decompress sample block -----------
			.NotEndB:
				Mov	EAX,[ESI]						;Copy BRR block into temp buffer
				Mov	EDX,[4+ESI]
				Mov	BL,[8+ESI]
				Mov	[tBRR],EAX
				Mov	[4+tBRR],EDX
				Mov	[8+tBRR],BL

				Push	EDI
				Mov	ESI,tBRR
				Mov	EDI,tBuf
				Mov	EDX,[tP1]
				Mov	EBX,[tP2]
				Call	[pDecomp]
				Mov	[tP1],EDX
				Mov	[tP2],EBX
				Pop	EDI

			.NoSInc:

			;Get sample ========================
			Mov	EDX,[tIdx]
			Test	byte [dspInter],3
			JZ 	.NoInter
			Test	byte [dspInter],2
			JZ		.LinInter
			Test	byte [dspInter],1
			JZ		short .CubInter

			;Gaussian Interpolation ------------
			.GaussInter:
				Mov	EAX,[tDec]
				ShR	EAX,8
				LEA	ESI,[EAX*8+gaussTab]
				FILd	word [EDX-6]
				FIMul	word [ESI+d0]
				FILd	word [EDX-4]
				FIMul	word [ESI+d1]
				FILd	word [EDX-2]
				FIMul	word [ESI+d2]
				FILd	word [EDX]
				FIMul	word [ESI+d3]
				FAddP	ST1,ST
				FAddP	ST1,ST
				FAddP	ST1,ST
				FAdd	ST,ST
				FIStP	dword [EDI]
				Add	EDI,4

		Dec	ECX
		JNZ	.NextSmp
		Jmp	.Reduce
ALIGN 16
			;Cubic Interpolation ---------------
			.CubInter:
				FILd	dword [tDec]
				FMul	dword [fpShR16]

				FLd	ST
				FMul	ST,ST1
				FLd	ST
				FMul	ST,ST2
				FLd	ST
				FMul	dword [fp0_5]
				FAdd	ST1,ST
				FXch	ST3
				FMul	dword [fp0_5]

				FLd	ST2
				FSub	ST,ST4
				FSub	ST,ST1

				FXch	ST1
				FAdd	ST,ST3
				FAdd	ST,ST3
				FSub	ST,ST2

				FXch	ST2
				FLd1
				FAddP	ST1,ST
				FLd	ST3
				FMul	dword [fn2_5]
				FAddP	ST1,ST

				FXch	ST3
				FMul	dword [fp0_5]
				FSubP	ST4,ST

				FIMul	word [EDX-6]
				FXch	ST3
				FIMul	word [EDX-0]
				FAddP	ST3,ST
				FIMul	word [EDX-2]
				FAddP	ST2,ST
				FIMul	word [EDX-4]
				FAddP	ST1,ST
				FMul	dword [fp64k]
				FSt	dword [EDI]
				MovZX	EAX,byte [3+EDI]				;Get mantissa and sign of float
				FIStP	dword [EDI]

				;Check for clipping -------------
				ShL	EAX,1								;Save sign bit in AH
				Cmp	AL,9Dh							;Is |sample| less than 2^31?
				JB		short .SmpOK
					ShR	EAX,8							;EAX = sign
					Add	EAX,7FFFFFFFh				;Add (2^31)-1
					Mov	[EDI],EAX
				.SmpOK:
				Add	EDI,4

		Dec	ECX
		JNZ	.NextSmp
		Jmp	.Reduce
ALIGN	16
			;Linear Interpolation -----------
			.LinInter:
				FILd	word [EDX-2]
				FILd	word [EDX]
				FSub	ST,ST1
				FIMul	dword [tDec]
				FMul	dword [fpShR16]
				FAddP	ST1,ST
				FMul	dword [fp64k]
				FIStP	dword [EDI]
				Add	EDI,4

		Dec	ECX
		JNZ	.NextSmp
		Jmp	.Reduce
ALIGN	16
			;No Interpolation ---------------
			.NoInter:
				Mov	AX,[EDX]
				ShL	EAX,16
				Mov	[EDI],EAX
				Add	EDI,4

		Dec	ECX
		JNZ	.NextSmp

		;=====================================
		;Reduce samples

		.Reduce:

		Mov	ESI,mixBuf
		Mov	EDI,[pSrcw]
		Mov	ECX,[ratew]

		Cmp	byte [dspChn],1
		JE		.Mono

		Cmp	byte [dspSize],1
		JE		.S8
		Cmp	byte [dspSize],2
		JE		.S16
		Cmp	byte [dspSize],-2
		JE		.S16
		Cmp	byte [dspSize],3
		JE		.S24

		.S32:
			Mov	EAX,[ESI]
			Mov	[0+EDI],EAX
			Mov	[4+EDI],EAX
			Add	ESI,4
			Add	EDI,8

		Dec	ECX
		JNZ	short .S32
		Jmp	.Converted

		.S24:
			Mov	AX,[2+ESI]
			Mov	DL,[1+ESI]
			Mov	[0+EDI],DL
			Mov	[1+EDI],AX
			Mov	[3+EDI],DL
			Mov	[4+EDI],AX

		Dec	ECX
		JNZ	short .S24
		Jmp	.Converted


		.S16:
			Mov	AX,[2+ESI]
			Mov	[0+EDI],AX
			Mov	[2+EDI],AX
			Add	ESI,4
			Add	EDI,4

		Dec	ECX
		JNZ	short .S16
		Jmp	short .Converted

		.S8:
			Mov	AL,[3+ESI]
			Mov	[0+EDI],AL
			Mov	[1+EDI],AL
			Add	ESI,4
			Add	EDI,2

		Dec	ECX
		JNZ	short .S8
		Jmp	short .Converted

		.Mono

		Cmp	byte [dspSize],1
		JE		.M8

		.M16:
			Mov	AX,[2+ESI]
			Mov	[EDI],AX
			Add	ESI,4
			Add	EDI,2

		Dec	ECX
		JNZ	short .M16
		Jmp	short .Converted

		.M8:
			Mov	AL,[3+ESI]
			Mov	[EDI],AL
			Add	ESI,4
			Inc	EDI

		Dec	ECX
		JNZ	short .M8
		Jmp	short .Converted

		.Converted:
		Mov	[pSrcw],EDI

	Mov	EAX,[loopw]
	Test	EAX,EAX
	JG		.Next

	.Done:
	RetN	[pSrcw]

	;=========================================
	;Initialize source playback

	.New:
		XOr	EDX,EDX
		ShLD	EDX,EAX,16
		ShL	EAX,16
		Div	dword [dspRate]
		Cmp	EAX,100000h
		JBE	short .RateOK
			Mov	EAX,100000h
		.RateOK:
		Mov	[tRate],EAX
		Mov	dword [tDec],0

		Mov	ESI,[pSrcw]
		Test	word [2+pSrcw],-1
		JNZ	short .HavePtr
			Or		ESI,[pAPURAM]
		.HavePtr
		Mov	[tBlk],ESI

		Mov	EAX,[loopw]
		LEA	EAX,[EAX*8+EAX]
		Add	EAX,ESI
		Mov	[tLoop],EAX

		Mov	EDI,tBuf
		Mov	[tIdx],EDI

		XOr	EDX,EDX
		XOr	EBX,EBX
		Mov	[00h+EDI],EDX

		Mov	AL,[ESI]
		Call	[pDecomp]
		Mov	[tP1],EDX
		Mov	[tP2],EBX

		Mov	EAX,[dspRate]
		Mul	dword [tRate]
		ShRD	EAX,EDX,16

ENDP


;
;Decompress Sound Source
;
;Decompresses a 9-byte bit-rate reduced block into 16 16-bit samples
;
;In:
;   AL  = Block header
;   ESI-> Sample Block
;   EDI-> Output buffer
;   EDX = Last sample of previous block
;   EBX = Next to last sample
;
;Out:
;   ESI-> Next Block
;   EDI-> After last sample
;   EDX = Last sample
;   EBX = Next to last sample
;
;Destroys:
;   EAX

ALIGN	16
UnpckSrc:

	Push	ECX,EBP

	Inc	SI												;Inc SI so pointer will wrap around a 16-bit value
	XOr	ECX,ECX
	Mov	CH,AL
	ShR	CH,4
	Add	ECX,brrTab									;ECX -> Row in brrTab

	Mov	EBP,8											;Decompress 8 bytes (16 nybbles)
	Test	AL,0Ch										;Does block use ADPCM compression?
	JZ 	short .Filter0								;  No

	Test	AL,08h										;Does block use filter 1?
	JZ 	short .Filter1								;  Yes

	Test	AL,04h										;Does block use filter 2?
	JZ 	.Filter2										;  Yes

	Jmp	.Filter3										;Then it must use filter 3

ALIGN	16
	;[Smp] -----------------------------------
	.Filter0:
	XOr	EAX,EAX
	XOr	EBX,EBX

	Mov	EDX,[ESI]
	Or		EDX,[ESI+4]
	JNZ	short .Next0								;Check if block contains any data

	Mov	ECX,8											;Fill output buffer with 0's
	Add	SI,byte 8
	Rep	StoSD
	Pop	EBP,ECX
	Ret

ALIGN	16
	.Next0:
		Mov	AL,[ESI]									;Get byte
		Inc	SI

		Mov	BL,AL										;EAX and EBX index sample value
		And	AL,0Fh
		ShR	BL,4

		Mov	EDX,[ECX+EAX*4]
		ShL	EDX,16
		Mov	DX,[ECX+EBX*4]

		Mov	[EDI],EDX								;Store samples
		Add	EDI,byte 4

	Dec	EBP
	JNZ	short .Next0
	MovSX	EBX,DX
	SAR	EDX,16
	Pop	EBP,ECX
	Ret

ALIGN	16
	;[Delta]+[Smp-1](15/16) ------------------
	.Filter1:
		Mov	CL,[ESI]									;CL indexes delta value
		And	CL,0F0h									;ECX -> value
		ShR	CL,2

		Mov	EBX,EDX
		Neg	EBX
		SAR	EBX,5
		LEA	EAX,[EBX*2+EDX]						;EAX = ((-p1 >> 4) & ~1) + p1
		Add	EAX,[ECX]								;EAX += delta
		MovSX	EBX,AX

		Mov	[EDI],EBX

		Mov	CL,[ESI]
		Inc	SI
		And	CL,0Fh
		ShL	CL,2

		Mov	EDX,EBX
		Neg	EDX
		SAR	EDX,5
		LEA	EAX,[EDX*2+EBX]
		Add	EAX,[ECX]
		MovSX	EDX,AX

		Mov	[2+EDI],DX
		Add	EDI,4

	Dec	EBP
	JNZ	short .Filter1
	Pop	EBP,ECX
	Ret

ALIGN	16
	;[Delta]+[Smp-1](61/32)-[Smp-2](15/16) ---
	.Filter2:
		Mov	CL,[ESI]
		And	CL,0F0h
		ShR	CL,2

		;Subtract 15/16 of second sample ------
		Mov	EAX,EBX
		Neg	EBX
		SAR	EAX,5
		LEA	EAX,[EAX*2+EBX]						;s = ((p2 >> 4) & ~1) + -p2
		Mov	EBX,EDX

		;Add 61/32 of last sample -------------
		LEA	EAX,[EDX*2+EAX]						;s += 2 * p1
		LEA	EDX,[EDX*2+EDX]
		Neg	EDX
		Add	EAX,[ECX]								;s += delta
		SAR	EDX,6
		LEA	EAX,[EDX*2+EAX]						;s += ((-3 * p1) >> 5) & ~1
		MovSX	EDX,AX

		;Clamp 16-bit sample to a 17-bit value
		Add	EAX,65536								;if s >= -65536 && s <= 65534
		SAR	EAX,17
		JZ		short .F2AOK
			SetS	DL										;if s > 65534 then s = -2
			MovZX	EDX,DL								;if s < -65536 then s = 0
			Dec	EDX
			Add	EDX,EDX
		.F2AOK:

		Mov	[EDI],EDX

		Mov	CL,[ESI]
		Inc	SI
		And	CL,0Fh
		ShL	CL,2

		Mov	EAX,EBX
		Neg	EBX
		SAR	EAX,5
		LEA	EAX,[EAX*2+EBX]
		Mov	EBX,EDX

		LEA	EAX,[EDX*2+EAX]
		LEA	EDX,[EDX*2+EDX]
		Neg	EDX
		Add	EAX,[ECX]
		SAR	EDX,6
		LEA	EAX,[EDX*2+EAX]
		MovSX	EDX,AX

		Add	EAX,65536
		SAR	EAX,17
		JZ		short .F2BOK
			SetS	DL
			MovZX	EDX,DL
			Dec	EDX
			Add	EDX,EDX
		.F2BOK:

		Mov	[2+EDI],DX
		Add	EDI,4

	Dec	EBP
	JNZ	.Filter2
	Pop	EBP,ECX
	Ret

ALIGN	16
	;[Delta]+[Smp-1](115/64)-[Smp-2](13/16) --
	.Filter3:
		Mov	CL,[ESI]
		And	CL,0F0h
		ShR	CL,2

		;Subtract 52/64 of second sample ------
		Mov	EAX,EBX
		LEA	EBX,[EBX*2+EBX]
		SAR	EBX,5
		Neg	EAX
		LEA	EAX,[EBX*2+EAX]						;s = -p2 + (((p2 * 3) >> 4) & ~1)
		Mov	EBX,EDX

		;Add 115/64 of last sample ------------
		LEA	EAX,[EDX*2+EAX]						;s += p1 * 2
		LEA	EDX,[EBX*4+EBX]
		LEA	EDX,[EBX*8+EDX]
		Add	EAX,[ECX]
		Neg	EDX
		SAR	EDX,7
		LEA	EAX,[EDX*2+EAX]						;s += ((-13 * p1) >> 6) & ~1
		MovSX	EDX,AX

		Add	EAX,65536
		SAR	EAX,17
		JZ		short .F3AOK
			SetS	DL
			MovZX	EDX,DL
			Dec	EDX
			Add	EDX,EDX
		.F3AOK:

		Mov	[EDI],EDX

		Mov	CL,[ESI]
		Inc	SI
		And	CL,0Fh
		ShL	CL,2

		Mov	EAX,EBX
		LEA	EBX,[EBX*2+EBX]
		SAR	EBX,5
		Neg	EAX
		LEA	EAX,[EBX*2+EAX]
		Mov	EBX,EDX

		LEA	EAX,[EDX*2+EAX]
		LEA	EDX,[EBX*4+EBX]
		LEA	EDX,[EBX*8+EDX]
		Add	EAX,[ECX]
		Neg	EDX
		SAR	EDX,7
		LEA	EAX,[EDX*2+EAX]
		MovSX	EDX,AX

		Add	EAX,65536
		SAR	EAX,17
		JZ		short .F3BOK
			SetS	DL
			MovZX	EDX,DL
			Dec	EDX
			Add	EDX,EDX
		.F3BOK:

		Mov	[2+EDI],DX
		Add	EDI,4

	Dec	EBP
	JNZ	.Filter3
	Pop	EBP,ECX
	Ret

;	;Expand nybbles into byte form -----------
;	MovQ			MM4,[ESI]
;	MovQ			MM6,[hiMask]
;	PAnd			MM6,MM4								;MM6 = high nybbles
;	PSLLQ			MM4,4
;	PAnd			MM4,[hiMask]						;MM4 = low nybbles
;	MovQ			MM7,MM6
;	PUnpckLBW	MM6,MM4								;MM7:MM6 = Delta values shifted out to byte form
;	PUnpckHBW	MM7,MM4
;
;	;Expand into dwords and shift by range ---
;	MovQ			MM1,MM6
;	MovQ			MM2,MM6
;	MovQ			MM3,MM6
;	MovQ			MM4,MM6
;
;	PUnpckLBW	MM1,MM0
;	PUnpckLBW	MM2,MM0
;	PUnpckHBW	MM3,MM0
;	PUnpckHBW	MM4,MM0
;
;	PUnpckLWD	MM1,MM0
;	PUnpckHWD	MM2,MM0
;	PUnpckLWD	MM3,MM0
;	PUnpckHWD	MM4,MM0
;
;	PSLLD			MM1,24
;	PSLLD			MM2,24
;	PSLLD			MM3,24
;	PSLLD			MM4,24
;
;	PSRAD			MM1,MM5
;	PSRAD			MM2,MM5
;	PSRAD			MM3,MM5
;	PSRAD			MM4,MM5
;
;	MovQ			[00+EDI],MM1
;	MovQ			[08+EDI],MM2
;	MovQ			[16+EDI],MM3
;	MovQ			[24+EDI],MM4


;
;Decompress Sound Source (Old school method)

ALIGN	16
UnpckSrcOld:

	Push	ECX

	;Get range -------------------------------
	Mov	CL,0CFh
	Inc	SI
	Sub	CL,AL											;CL = 12 - Range (change range from << to >>)
	SetNC	AH												;If result is negative (invalid range) add 3
	Dec	AH
	And	AH,30h
	Add	CL,AH
	ShR	CL,4

	Mov	CH,8
	Test	AL,0Ch
	JZ 	short .Filter0

   Add	CL,10											;Values will be shifted right from 32-bit values
	Test	AL,8
	JZ 	short .Filter1

	Test	AL,4
	JZ 	.Filter2

	Jmp	.Filter3

ALIGN	16
	;[Smp] -----------------------------------
	.Filter0:
		XOr	EAX,EAX
		XOr	EDX,EDX
		Mov	AH,[ESI]
		Mov	DH,AH
		And	AH,0F0h
		ShL	DH,4

		SAR	AX,CL
		SAR	DX,CL
		Mov	[EDI],AX
		Mov	[2+EDI],DX
		Add	EDI,4

		Inc	SI

	Dec	CH
	JNZ	short .Filter0
	MovSX	EDX,DX
	MovSX	EBX,AX
	Pop	ECX
	Ret

ALIGN	16
	;[Delta]+[Smp-1](15/16) ------------------
	.Filter1:
		Mov	EBX,[ESI]
		And	BL,0F0h
		ShL	EBX,24
		SAR	EBX,CL

		Mov	EAX,EDX
		IMul	EAX,60
		Add	EBX,EAX
		SAR	EBX,6

		Mov	[EDI],EBX

		Mov	EDX,[ESI]
		ShL	EDX,28
		SAR	EDX,CL

		Mov	EAX,EBX
		IMul	EAX,60
		Add	EDX,EAX
		SAR	EDX,6

		Mov	[2+EDI],DX
		Add	EDI,4

		Inc	SI

	Dec	CH
	JNZ	short .Filter1
	Pop	ECX
	Ret

ALIGN	16
	;[Delta]+[Smp-1](61/32)-[Smp-2](30/32) ---
	.Filter2:
		Mov	EAX,[ESI]
		And	AL,0F0h
		ShL	EAX,24
		SAR	EAX,CL

		;Subtract 15/16 of second sample ------
		IMul	EBX,60
		Sub	EAX,EBX
		Mov	EBX,EDX

		;Add 61/32 of last sample -------------
		IMul	EDX,122
		Add	EAX,EDX
		SAR	EAX,6

		Mov	[EDI],EAX

		Mov	EDX,[ESI]
		ShL	EDX,28
		SAR	EDX,CL

		IMul	EBX,60
		Sub	EDX,EBX
		Mov	EBX,EAX

		IMul	EAX,122
		Add	EDX,EAX
		SAR	EDX,6

		Mov	[2+EDI],DX
		Add	EDI,4

		Inc	SI

	Dec	CH
	JNZ	.Filter2
	Pop	ECX
	Ret

ALIGN	16
	;[Delta]+[Smp-1](115/64)-[Smp-2](52/64) --
	.Filter3:
		Mov	EAX,[ESI]
		And	AL,0F0h
		ShL	EAX,24
		SAR	EAX,CL

		;Subtract 13/16 of second sample ------
		IMul	EBX,52
		Sub	EAX,EBX
		Mov	EBX,EDX

		;Add 115/64 of last sample ------------
		IMul	EDX,115
		Add	EAX,EDX
		SAR	EAX,6

		Mov	[EDI],EAX

		Mov	EDX,[ESI]
		ShL	EDX,28
		SAR	EDX,CL

		IMul	EBX,52
		Sub	EDX,EBX
		Mov	EBX,EAX

		IMul	EAX,115
		Add	EDX,EAX
		SAR	EDX,6

		Mov	[2+EDI],DX
		Add	EDI,4

		Inc	SI

	Dec	CH
	JNZ	.Filter3
	Pop	ECX
	Ret


;
;Unpack Waveform

PROC UnpackWave, pBufu, pSrcu, numu, optsu, prev1u, prev2u
LOCALS method, blocks, temp
USES ALL

	;Initialize local variables --------------
	Mov	EDX,[optsu]									;Select unpacking method based on option
	XOr	EAX,EAX
	Test	DL,DSP_OLDSMP
	SetZ	AL
	Dec	EAX
	And	EAX,UnpckSrcOld-UnpckSrc
	Add	EAX,UnpckSrc
	Mov	[method],EAX

	XOr	EAX,EAX
	Mov	[blocks],EAX
	Mov	[temp],EAX

	;Initialize previous samples -------------
	LEA	EBX,[temp]
	XOr	EAX,EAX										;if (prev2==NULL) prev2=&temp
	Cmp	[prev2u],EAX
	SetNE	AL
	Dec	EAX
	And	EAX,EBX
	Or		[prev2u],EAX

	XOr	EAX,EAX										;if (prev1==NULL) prev1=&temp
	Cmp	[prev1u],EAX
	SetNE	AL
	Dec	EAX
	And	EAX,EBX
	Or		[prev1u],EAX

	Mov	EBX,[prev1u]								;EDX = *prev1
	Mov	EDX,[EBX]
	Mov	EBX,[prev2u]								;EBX = *prev2
	Mov	EBX,[EBX]

	Mov	ECX,[numu]
	Mov	EDI,[pBufu]

	Mov	AX,[2+pSrcu]
	Cmp	AX,[2+pAPURAM]
	JE		short .InRAM
	Test	AX,AX
	JZ		short .InRAM

	;Decompress samples not in APU RAM -------
	.NextC:
		Mov	ESI,[pSrcu]
		Add	dword [pSrcu],9

		Mov	AL,[8+ESI]								;Copy BRR block into 'mixBuf'
		Mov	[8+mixBuf],AL
		Mov	EAX,[4+ESI]
		Mov	[4+mixBuf],EAX
		Mov	EAX,[ESI]
		Mov	[mixBuf],EAX

		Mov	ESI,mixBuf								;Decompress block
		Call	[method]
		Inc	dword [blocks]
		Test	byte [ESI-9],1
		JNZ	short .Done

	Dec	ECX
	JNZ	short .NextC
	Jmp	short .Done

	;Decompress samples in APU RAM -----------
	.InRAM:
	Mov	ESI,[pAPURAM]
	Mov	SI,[pSrcu]

	.Next:
		Mov	AL,[ESI]
		Call	[method]
		Inc	dword [blocks]
		Test	byte [ESI-9],1							;Have we reached the end block?
		JNZ	.Done
		Cmp	SI,[pSrcu]								;Have we wrapped around in RAM?
		JBE	.Done
	Dec	ECX
	JNZ	short .Next
	.Done:

	;Return previous samples -----------------
	Mov	EAX,[prev2u]
	Mov	[EAX],EBX
	Mov	EBX,[prev1u]
	Mov	[EBX],EDX

	Mov	EAX,[blocks]

ENDP


;
;Compress Waveform
;
;In:
;   pIn  -> Samples
;   pOut -> Output buffer
;   prev1 = Last sample of previous block
;   prev2 = Next to last sample
;
;Out:
;   EDX = Last sample
;   EBX = Next to last sample
;
;Destroys:
;   EAX

STRUC Blk
	brr	resb	16	; [0]
	range	resb	1	;[10]
			resb	3
	diff	resd	1	;[14]
	p1		resd	1	;[18]
	p2		resd	1	;[1C]
ENDSTRUC

PROC PackSrc, pOut, pIn, prev1, prev2, optpw
USES ECX,ESI,EDI

	LEA	EDI,[ESP-128]

	Mov	dword [diff+32+EDI],-1
	Mov	dword [diff+64+EDI],-1
	Mov	dword [diff+96+EDI],-1


	;=========================================
	;Linear Compression (Filter 0)

	;Find maximum sample value ---------------
	Mov	ESI,[pIn]
	Mov	CL,16
	XOr	EBX,EBX
	.Next0:
		MovSX	EAX,word [ESI]
		CDQ
		XOr	EAX,EDX
		Add	ESI,2
		BSR	EAX,EAX
		Cmp	BL,AL
		CMovB	EBX,EAX

	Dec	CL
	JNZ	short .Next0
	Sub	ESI,32

	Sub	BL,2
	SetS	CL
	Dec	CL
	And	CL,BL
	Mov	[range+EDI],CL

	;Reduce samples to 4-bits ----------------
	XOr	EAX,EAX
	XOr	EBX,EBX
	Mov	[diff+EDI],EAX
	Mov	CH,16
	.Test0:
		Mov	AX,[ESI]
		And	AL,~1
		SAR	AX,CL
		Mov	[EBX+EDI],AL
		Inc	EBX
		ShL	AX,CL
		Sub	AX,[ESI]
		Add	ESI,2
		CWD
		XOr	AX,DX
		Sub	AX,DX
		Add	[diff+EDI],EAX

	Dec	CH
	JNZ	short .Test0

	MovSX	EAX,byte [14+EDI]							;Create previous samples
	ShL	EAX,CL
	Mov	[p2+EDI],EAX
	MovSX	EAX,byte [15+EDI]
	ShL	EAX,CL
	Mov	[p1+EDI],EAX

	Add	EDI,32

	Test	byte [optpw],BRR_LINEAR
	JNZ	.Done

	;=========================================
	;Filter 1

	Mov	CL,0
	.Range1:
		Mov	ESI,[pIn]

		Mov	EAX,[prev1]
		Mov	EDX,[prev2]
		Mov	[p1+EDI],EAX
		Mov	[p2+EDI],EDX

		XOr	EBX,EBX
		Mov	[diff+EDI],EBX

		Mov	CH,16
		.Next1:
			Mov	EDX,[p1+EDI]
			Mov	EAX,EDX
			Neg	EDX
			SAR	EDX,5
			LEA	EDX,[EDX*2+EAX]

			MovSX	EAX,word [ESI]
			And	AL,~1
			Sub	EAX,EDX
			SAR	EAX,CL

			Cmp	EAX,7
			JG		.Broken1
			Cmp	EAX,-8
			JL		.Broken1

			Mov	[EBX+EDI],AL
			Inc	EBX

			ShL	EAX,CL
			Add	EAX,EDX

			Cmp	EAX,32767
			JG		.Broken1
			Cmp	EAX,-32768
			JL		.Broken1

			Mov	EDX,[p1+EDI]
			Mov	[p2+EDI],EDX
			Mov	[p1+EDI],EAX

			MovSX	EDX,word [ESI]
			Add	ESI,2
			Sub	EAX,EDX
			CDQ
			XOr	EAX,EDX
			Sub	EAX,EDX
			Add	[diff+EDI],EAX

		Dec	CH
		JNZ	short .Next1

		Mov	[range+EDI],CL
		Jmp	short .Good1

		.Broken1:

	Inc	CL
	Cmp	CL,12
	JBE	.Range1
	Mov	dword [diff+EDI], -1

	.Good1:
	Add	EDI,32

	;=========================================
	;Filter 2

	Mov	CL,0
	.Range2:
		Mov	ESI,[pIn]

		Mov	EAX,[prev1]
		Mov	EDX,[prev2]
		Mov	[p1+EDI],EAX
		Mov	[p2+EDI],EDX

		XOr	EBX,EBX
		Mov	[diff+EDI],EBX

		Mov	CH,16
		.Next2:
			Mov	EAX,[p2+EDI]
			Mov	EDX,EAX
			Neg	EAX
			SAR	EDX,5
			LEA	EDX,[EDX*2+EAX]
			Mov	[p2+EDI],EDX

			Mov	EAX,[p1+EDI]
			Mov	EDX,EAX
			Add	EAX,EAX
			Neg	EDX
			LEA	EDX,[EDX*2+EDX]
			SAR	EDX,6
			LEA	EDX,[EDX*2+EAX]

			MovSX	EAX,word [ESI]
			And	AL,~1
			Sub	EAX,EDX
			Sub	EAX,[p2+EDI]
			SAR	EAX,CL

			Cmp	EAX,7
			JG		.Broken2
			Cmp	EAX,-8
			JL		.Broken2

			Mov	[EBX+EDI],AL
			Inc	EBX

			ShL	EAX,CL
			Add	EAX,EDX
			Add	EAX,[p2+EDI]

			Cmp	EAX,32767
			JG		.Broken2
			Cmp	EAX,-32768
			JL		.Broken2

			Mov	EDX,[p1+EDI]
			Mov	[p2+EDI],EDX
			Mov	[p1+EDI],EAX

			MovSX	EDX,word [ESI]
			Add	ESI,2
			Sub	EAX,EDX
			CDQ
			XOr	EAX,EDX
			Sub	EAX,EDX
			Add	[diff+EDI],EAX

		Dec	CH
		JNZ	.Next2

		Mov	[range+EDI],CL
		Jmp	short .Good2

		.Broken2:

	Inc	CL
	Cmp	CL,12
	JBE	.Range2
	Mov	dword [diff+EDI],-1

	.Good2:
	Add	EDI,32

	;=========================================
	;Filter 3

	Mov	CL,0
	.Range3:
		Mov	ESI,[pIn]

		Mov	EAX,[prev1]
		Mov	EDX,[prev2]
		Mov	[p1+EDI],EAX
		Mov	[p2+EDI],EDX

		XOr	EBX,EBX
		Mov	[diff+EDI],EBX

		Mov	CH,16
		.Next3:
			Mov	EDX,[p2+EDI]
			Mov	EAX,[p2+EDI]
			LEA	EDX,[EDX*2+EDX]
			SAR	EDX,5
			Neg	EAX
			LEA	EAX,[EDX*2+EAX]
			Mov	[p2+EDI],EAX

			Mov	EDX,[p1+EDI]
			LEA	EAX,[EDX*4+EDX]
			LEA	EAX,[EDX*8+EAX]
			Add	EDX,EDX
			Neg	EAX
			SAR	EAX,7
			LEA	EDX,[EAX*2+EDX]

			MovSX	EAX,word [ESI]
			And	AL,~1
			Sub	EAX,EDX
			Sub	EAX,[p2+EDI]
			SAR	EAX,CL

			Cmp	EAX,7
			JG		.Broken3
			Cmp	EAX,-8
			JL		.Broken3

			Mov	[EBX+EDI],AL
			Inc	EBX

			ShL	EAX,CL
			Add	EAX,EDX
			Add	EAX,[p2+EDI]

			Cmp	EAX,32767
			JG		.Broken3
			Cmp	EAX,-32768
			JL		.Broken3

			Mov	EDX,[p1+EDI]
			Mov	[p2+EDI],EDX
			Mov	[p1+EDI],EAX

			MovSX	EDX,word [ESI]
			Add	ESI,2
			Sub	EAX,EDX
			CDQ
			XOr	EAX,EDX
			Sub	EAX,EDX
			Add	[diff+EDI],EAX

		Dec	CH
		JNZ	.Next3

		Mov	[range+EDI],CL
		Jmp	short .Good3

		.Broken3:

	Inc	CL
	Cmp	CL,12
	JBE	.Range3
	Mov	dword [diff+EDI],-1

	.Good3:
	.Done:

	;=========================================
	;Store Final Block

	;Find best filter ------------------------
	XOr	EDX,EDX										;EDX = Filter
	LEA	ESI,[ESP-128]								;ESI -> First filter block
	Mov	EAX,[diff+ESI]								;EAX = Total difference for first filter
	Mov	EBX,ESI										;EBX -> Array of block stuctures

	Mov	CL,1
	.NextB:
		Add	EBX,32
		Cmp	EAX,[diff+EBX]							;Is this a better filter?
		JBE	short .Better							;  Nope, keep what we have
			Mov	EAX,[diff+EBX]
			Mov	ESI,EBX
			Mov	DL,CL
		.Better:

	Inc	CL
	Cmp	CL,4
	JB		short .NextB

	;Create header byte ----------------------
	Mov	EDI,[pOut]

	Mov	AL,[range+ESI]								;Get BRR range
	ShL	AL,4

	ShL	DL,2											;Get filter number
	Or		AL,DL

	Mov	DL,[optpw]									;Set looping flag
	And	DL,2
	Or		AL,DL

	Mov	[EDI],AL										;Store header byte
	Inc	EDI

	;Pack bytes into nybbles -----------------
	Mov	ECX,7
	.Copy:
		Mov	AL,[ECX*2+ESI]
		Mov	DL,[1+ECX*2+ESI]
		ShL	AL,4
		And	DL,0Fh									;Remove sign bits
		Or		AL,DL
		Mov	[ECX+EDI],AL

	Dec	ECX
	JNS	short .Copy

	;Return previous samples -----------------
	Mov	EDX,[p1+ESI]
	Mov	EBX,[p2+ESI]

ENDP


;
;Pack Waveform

PROC PackWave, pBufp, pSmpp, nump, optsp, prev1p, prev2p
USES ALL

	;Initialize local variables --------------
	Mov	EDX,[optsp]

	;Initialize previous samples -------------
	Mov	EBX,[prev2p]								;if (prev2p) EBX = *prev2p
	Test	EBX,EBX
	JZ		short .ZeroP2
		Mov	EBX,[EBX]
	.ZeroP2:

	Mov	EDX,[prev1p]								;if (prev1p) EDX = *prev1p
	Test	EDX,EDX
	JZ		short .ZeroP1
		Mov	EDX,[EDX]
	.ZeroP1:

	;Compress samples ------------------------
	Mov	ESI,[pSmpp]
	Mov	EDI,[pBufp]
	Mov	ECX,[nump]
	.Next:
		Call	PackSrc,EDI,ESI,EDX,EBX,[optsp]
		Add	ESI,32
		Add	EDI,9
	Dec	ECX
	JNZ	short .Next

	;Return previous samples -----------------
	Mov	ECX,[prev2p]
	Test	ECX,ECX
	JZ		short .NoP2
		Mov	[ECX],EBX
	.NoP2:

	Mov	ECX,[prev1p]
	Test	ECX,ECX
	JZ		short .NoP1
		Mov	[ECX],EDX
	.NoP1:

ENDP


;
;VMax to Decible

PROC VMax2dB, pList, f
USES ALL

	FInit

	Mov	EDI,[pList]									;EDI -> Output array
	Mov	EBX,mix+vMaxL								;EBX -> vMax values
	XOr	EDX,EDX
	Mov	CL,8											;Eight voices
	FLd	dword [fp6]

	Mov	AL,[f]
	Test	AL,AL											;Does caller want floating-point values?
	JNZ	.Float										;  Yes
		;Integer format =======================
		;
		; dB = (6.0 (Log  vMax) + 6.0) * 65536
		;               2

		;Voice output -------------------------
		.IVoice:
			Mov	[EDI],EDX							;Default to 0 (negative infinity)
			Mov	EAX,[EBX]							;Get vMaxL
			Test	EAX,EAX								;Is there a value?
			JZ		short .INoRangeL					;  No, Move onto vMaxR
				FLd	ST
				FILd	dword [EBX]						;Load vMaxL into FPU
				Mov	[EBX],EDX						;Reset vMaxL
				FYL2X										;Compute Log2*6.0
				FAdd	ST,ST1							;Add 6.0
				FMul	dword [fp64k]					;Multiply by 65536
				FIStP	dword [EDI]						;Store dB as 16.16 integer
			.INoRangeL:
			Add	EDI,4
			Add	EBX,4

			Mov	[EDI],EDX							;Do the same thing for vMaxR
			Mov	EAX,[EBX]
			Test	EAX,EAX
			JZ		short .INoRangeR
				FLd	ST
				FILd	dword [EBX]
				Mov	[EBX],EDX
				FYL2X
				FAdd	ST,ST1
				FMul	dword [fp64k]
				FIStP	dword [EDI]
			.INoRangeR:
			Add	EDI,4
			Add	EBX,124

		Dec	CL
		JNZ	short .IVoice

		;Main output (vMMax) ------------------
		Mov	[EDI],EDX
		Mov	EAX,[vMMaxL]
		Test	EAX,EAX
		JZ		short .INoRangeML
			FLd	ST
			FILd	dword [vMMaxL]
			Mov	[vMMaxL],EDX
			FYL2X
			FAdd	ST,ST1
			FMul	dword [fp64k]
			FIStP	dword [EDI]
		.INoRangeML:
		Add	EDI,4

		Mov	[EDI],EDX
		Mov	EAX,[vMMaxR]
		Test	EAX,EAX
		JZ		.Done
			FLd	ST
			FILd	dword [vMMaxR]
			Mov	[vMMaxR],EDX
			FYL2X
			FAdd	ST,ST1
			FMul	dword [fp64k]
			FIStP	dword [EDI]
		Jmp	.Done

	.Float:
		;Floating-point Format ================
		;
		; dB = 6.0 (Log  vMax) - 90.0
		;              2

		;Voice output -------------------------
		.FVoice:
			Mov	dword [EDI],42C00000h			;Default to -96.0 (negative infinity)
			Mov	EAX,[EBX]							;Get vMaxL
			Test	EAX,EAX								;Is there a value?
			JZ		short .FNoRangeL					;  No, Move onto vMaxR
				FLd	ST
				FILd	dword [EBX]						;Load vMaxL into FPU
				Mov	[EBX],EDX						;Reset vMaxL
				FYL2X										;Compute Log2*6.0
				FSub	dword [fp90]					;Subtract result from 90.0
				FStP	dword [EDI]						;Store dB as floating-point
			.FNoRangeL:
			Add	EDI,4
			Add	EBX,4

			Mov	dword [EDI],42C00000h			;Do the same thing for vMaxR
			Mov	EAX,[EBX]
			Test	EAX,EAX
			JZ		short .FNoRangeR
				FLd	ST
				FILd	dword [EBX]
				Mov	[EBX],EDX
				FYL2X
				FSub	dword [fp90]
				FStP	dword [EDI]
			.FNoRangeR:
			Add	EDI,4
			Add	EBX,124

		Dec	CL
		JNZ	short .FVoice

		;Main output (vMMax) ------------------
		Mov	dword [EDI],42C00000h
		Mov	EAX,[vMMaxL]
		Test	EAX,EAX
		JZ		short .FNoRangeML
			FLd	ST
			FILd	dword [vMMaxL]
			Mov	[vMMaxL],EDX
			FYL2X
			FSub	dword [fp90]
			FStP	dword [EDI]
		.FNoRangeML:
		Add	EDI,4

		Mov	dword [EDI],42C00000h
		Mov	EAX,[vMMaxR]
		Test	EAX,EAX
		JZ		short .Done
			FLd	ST
			FILd	dword [vMMaxR]
			Mov	[vMMaxR],EDX
			FYL2X
			FSub	dword [fp90]
			FStP	dword [EDI]

	.Done:
	FStP	ST
ENDP
